/*

  Sun Public License Notice

 The contents of this file are subject to the Sun Public License Version
 1.0 (the "License"). You may not use this file except in compliance with
 the License. A copy of the License is available at http://www.sun.com/

 The Original Code is the Sunlabs Mobile IP.
 The Initial Developer of the Original Code is Vipul Gupta.
 Portions created by Vipul Gupta are Copyright (C) 2003.
 All Rights Reserved.

 Contributor(s): Niko A. Fikouras (niko@comnets.uni-bremen.de)
		 Asanga D. Udugama (adu@comnets.uni-bremen.de)

*/

/*
 * An implementation of Mobile-IP as described in RFC 2002.
 * Author: Vipul Gupta (vgupta@eng.sun.com)
 * agent.c : Main program for a Mobile IP agent.
 */

#pragma ident   "@(#)agent.c 1.16 98/11/30"

#include <stdio.h>
/* Modified by Uni. of Sheffield */
#include <sys/time.h>
#include "mip.h"
#include "agent.h"

/* Controls verbosity of debug messages when compiled w/ "-D MIP_DEBUG" */
int mipdebug = 3;
int IDfreshnessSlack = DEFAULT_FRESHNESS_SLACK;
int regLifetime = DEFAULT_MAX_REG_TIME;
int advLifetime = DEFAULT_MAX_ADV_TIME;
int periodicInterval = DEFAULT_PERIODIC_INTERVAL;
/* Modified by Uni. of Sheffield */
int granularity = DEFAULT_GRANULARITY;
int advertiseOnBcast = 0;
int timerWentOff = 1;
/* Modified by Uni. of Sheffield */
int totalTime = 0;

boolean faBusy = FALSE;

/* used for debug messages */
char msg[256], currentTime[32], relativeTime[32];
char addr1[20], addr2[20], addr3[20];

/* ----------------- Common to all mobility agents ------------------- */
/*
 * This table stores configuration information about agent
 * advertisements. There's one entry for each mobility
 * supporting interface.
 */
MaAdvConfigEntry maAdvConfigTable[DEFAULT_MAX_INTERFACES];
int maAdvConfigTableSize = 0;

/* Counters common to all Mobility Agents */
u32  maAdvSentCnt = 0;
u32  maAdvSentForSolicitationsCnt = 0;
u32  maSolicitationsRecvdCnt = 0;

/* Sockets used for Mobile IP messages */
int  ICMPsock;      /* for sending and receiving ICMP messages */
int  BcastSock;     /* for registrations received on 255.255.255.255 */

/* TODO: Allow a mobility agent to support security associations with
 *        nodes other than those it supports as a Home Agent. MSA shared
 *        with those nodes can be kept in a table of "MipSecAssocEntry"s.
 */
/* ------------------ Specific to foreign agents -------------------- */
/*
 * This table stores information about visitors for which this
 * mobility agent is a foreign agent. Some of the entries may
 * correspond to unfulfilled registration requests.
 */
FaVisitorEntry faVisitorTable[DEFAULT_MAX_VISITORS];
int faVisitorTableSize = 0;

/*
 * This table stores addresses for which ICMP unreachables were
 * recently received.
 */
FaUnreachableEntry faUnreachableTable[DEFAULT_MAX_UNREACHABLES];
int faUnreachableTableSize;

/* Counters maintained by Foreign Agents */
u32  faRegReqRecvdCnt = 0;           /* total valid reg requests received */
u32  faRegReqRelayedCnt = 0;         /* total valid reg reqs relayed to HA */
u32  faReasonUnspecifiedCnt = 0;     /* rejected with code 64 */
u32  faAdmProhibitedCnt = 0;         /* rejected with code 65 */
u32  faInsufficientResourceCnt = 0;  /* rejected with code 66 */
u32  faMNAuthFailureCnt = 0;         /* rejected with code 67 */
u32  faHAAuthFailureCnt = 0;         /* rejected with code 68 */
u32  faRegLifetimeTooLongCnt = 0;    /* rejected with code 69 */
u32  faPoorlyFormedRequestsCnt = 0;  /* rejected with code 70 */
u32  faPoorlyFormedRepliesCnt = 0;   /* rejected with code 71 */
u32  faEncapUnavailableCnt = 0;      /* rejected with code 72 */
u32  faVJCompUnavailableCnt = 0;     /* rejected with code 73 */
u32  faHAUnreachableCnt = 0;         /* rejected with codes 80-95 */
u32  faRegRepliesRecvdCnt = 0;       /* well-formed reg replies received */

/* ------------------ Specific to home agents -------------------- */
/*
 * This table has one entry for each mobile node for which a mobility
 * agent offers Home Agent services.
 */
HaMobileNodeEntry haMobileNodeTable[DEFAULT_MAX_MOBILE_NODES];
int haMobileNodeTableSize = 0;

/*
 * This table has one entry for each mobility binding currently
 * supported this Home Agent. A mobile node may have multiple
 * binding entries at a given time.
 */
HaBindingEntry haBindingTable[DEFAULT_MAX_BINDINGS];
int haBindingTableSize = 0;

/* Counters maintained by Home Agents */
u32  haRegAccepted0Cnt = 0;           /* reg requests accepted with code 0 */
u32  haRegAccepted1Cnt = 0;           /* reg requests accepted with code 1 */
u32  haReasonUnspecifiedCnt = 0;      /* denied with code 128 */
u32  haAdmProhibitedCnt = 0;          /* denied with code 129 */
u32  haInsufficientResourceCnt = 0;   /* denied with code 130 */
u32  haMNAuthFailureCnt = 0;          /* denied with code 131 */
u32  haFAAuthFailureCnt = 0;          /* denied with code 132 */
u32  haIDMismatchCnt = 0;             /* denied with code 133 */
u32  haPoorlyFormedRequestsCnt = 0;   /* denied with code 134 */
u32  haTooManyBindingsCnt = 0;        /* denied with code 135 */
u32  haUnknownHACnt = 0;              /* denied with code 136 */
u32  haGratuitousARPsSentCnt = 0;
u32  haProxyARPsSentCnt = 0;
u32  haRegReqRecvdCnt = 0;
u32  haDeRegReqRecvdCnt = 0;
u32  haRegRepliesSentCnt = 0;
u32  haDeRegRepliesSentCnt = 0;

/* Security related stuff */
DomainInfo domainInfo;

/* to store the filters that have been defined for each connected mobile node */
BoundMN boundMNTable[DEFAULT_MAX_BINDINGS];			   // one for each MN bound
int boundMNTableSize = 0;

/* ----------------------------------------------------------------- */
extern void Initialize(FILE *);
extern void getAndDispatchMessages();
extern void printBuffer(unsigned char *, int);
extern int  sendUDPmessage(int, unsigned char *, int, u32, unsigned short);
extern int appendAuthExt(unsigned char *, int, int type, MipSecAssocEntry *);
extern boolean isAuthExtOk(unsigned char *, int, MipSecAssocEntry *);
extern HaMobileNodeEntry * findHAMNE(HaMobileNodeEntry *, int, u32);
extern boolean HAisIDok(u32, u32, u32, u32, int);
extern void HAnewID(u32 *, u32 *, u32, u32, int, boolean);
extern void HAstoreID(u32 *, u32 *, u32, u32, int, boolean);
extern char * hwAddrWrite(unsigned char *, char *);
extern char * ntoa(u32, char *);
extern char * sprintTime(char *, int);
extern char * sprintRelativeTime(char *, int);
extern char *err2str(int);

/* OS dependent function prototypes */
extern int encapadd(u32, u32, u32, unsigned short);
extern int encaprem(u32, u32 );
extern int decapadd(u32, u32, u32, char *);
extern int decaprem(u32, u32, u32);
extern int routeadd(u32, u32, int);
extern int routedel(u32, u32);
extern int arpadd(u32, unsigned char *, char *);
extern int arpdel(u32);
extern int arprefresh(u32);
extern int mkroutethroughtunnel(u32, u32, u32); // MSB
extern int rmroutethroughtunnel(u32, u32, u32); // MSB

extern int extractFilters(int, unsigned char [], unsigned char *[], BoundMN [], int *, u32 , u32);
extern void dispFilters(BoundMN [], u32);
extern int haveFiltersChanged(BoundMN [], u32);
extern void FAapplyFilters(BoundMN [], u32);
extern int HAapplyFilters(BoundMN [], HaBindingEntry [], u32);
extern int addFilterTableEntry(BoundMN [], u32);

#define MAX_EXPECTED_EXTENSIONS   32



/* TODO : Function to be moved in an appropriate file */
/* Check each interval to see if the given addr is outside the protected domain */
boolean isInsideProtectedDomain(u32 addr)
{
    int i;

    if (domainInfo.addrIntervalCnt == 0)
        return(TRUE);

    for (i = 0; i < domainInfo.addrIntervalCnt; i++) {
        if (((addr ^ domainInfo.addr[i]) & domainInfo.netmask[i]) == 0)
            return(TRUE);
    }

    return(FALSE);
}

boolean isOutsideProtectedDomain(u32 addr)
{
    return (isInsideProtectedDomain(addr) ? FALSE: TRUE);
}


/*
 * Examine buflen bytes in buffer and place all the Mobile IP
 * extensions encountered in extType[] and the start points
 * in extIdx[]. The maximum size of these two arrays is given
 * in cnt. At termination, cnt contains the numebr of extensions
 * found in the buffer. Buffer starts at an extension.
 */
int mkRegExtList(unsigned char buffer[], int buflen, int *cnt,
		 unsigned char extType[], unsigned char *extIdx[])
{
    unsigned char *currentPos = buffer;
    int bytesSeen = 0;
    int extSeen = 0;
    int skipOver;

    while ((bytesSeen < (buflen - 1)) && (extSeen < *cnt)) {
	extType[extSeen] = *currentPos;
        extIdx[extSeen] = currentPos;
	extSeen++;
	skipOver = *(currentPos+1) + 2;
	bytesSeen += skipOver;
	currentPos += skipOver;
    }

    if ((bytesSeen < (buflen - 1)) && (extSeen >= *cnt)) {
        mip1dbg(("Too many extensions.\n"));
	return(-1);
    } else if (bytesSeen != buflen) {
        mip1dbg(("Extensions buffer too small.\n"));
	return(-1);
    } else {
        *cnt = extSeen;
        return(0);
    }
}


/*
 * Check if the extension list is correct for a registration request
 * received at a Foreign Agent or a reply received at a Mobile Node.
 */
boolean MnFaOkExtList(int cnt, unsigned char extType[])
{
    int foundMHauth = 0;
    int foundMFauth = 0;
    int i, ctlExtRequired = FALSE;

    for (i = 0; i < cnt; i++) {
        if (extType[i] == REG_MH_AUTH_EXT_TYPE) {
	    /* we should not have seen any other auth extensions */
	    if (foundMHauth) {
	        mip1dbg(("Multiple MHauth\n"));
		return (FALSE);
	    } else {
	        foundMHauth++;
	    }
	} else if (extType[i] == REG_MF_AUTH_EXT_TYPE) {
	    /* we should have seen MHauth but no MFauth */
	    if (!foundMHauth) {
	        mip1dbg(("No MHauth before MFauth\n"));
		return (FALSE);
	    }
	    if (foundMFauth) {
	        mip1dbg(("Multiple MFauth\n"));
		return (FALSE);
            } else {
	        foundMFauth++;
	    }
	} else if(extType[i] >= FLT_DEL_EXT && extType[i] <= FLT_DSTPORTRNG_EXT) {
		// if filter extensions, then accept them
	} else if (extType[i] <= 127) {
	    /* we should never see an unrecognized ext in range 0-127 */
	    mip1dbg(("Unrecognized ext in range 0-127.\n"));
	    return(FALSE);
	} else {
	    /*
	     * Extensions in the range 128-255 should not appear
	     * after MFauth
	     */
	    if (foundMFauth) {
	        mip1dbg(("Ext found after MFauth\n"));
		return(FALSE);
	    }
	}
    }

    /* make sure we have one MHauth */
    if (foundMHauth) {
        return(TRUE);
    } else {
        mip1dbg(("MHauth missing\n"));
	return(FALSE);
    }
}

/*
 * Check if the extension list is correct for a registration request
 * recived at a Home Agent or a reply received at a Foreign Agent.
 */
boolean FaHaOkExtList(int cnt, unsigned char extType[])
{
    int foundMHauth = 0;
    int foundFHauth = 0;
    int i, ctlExtRequired = FALSE;

    for (i = 0; i < cnt; i++) {
        if (extType[i] == REG_MH_AUTH_EXT_TYPE) {
	    /* we should not have seen any other auth extensions */
	    if (foundMHauth) {
	        mip1dbg(("Multiple MHauth\n"));
	        return (FALSE);
	    } else {
	        foundMHauth++;
	    }
	} else if (extType[i] == REG_FH_AUTH_EXT_TYPE) {
	    /* we should have seen MHauth but no FHauth */
	    if (!foundMHauth) {
	        mip1dbg(("No MHauth before FHauth\n"));
	        return (FALSE);
	    } else if (foundFHauth) {
	        mip1dbg(("Multiple FHauth\n"));
	        return (FALSE);
	    } else {
	        foundFHauth++;
	    }

	} else if(extType[i] >= FLT_DEL_EXT && extType[i] <= FLT_DSTPORTRNG_EXT) {
		// if filter extensions, then accept them
	} else if (extType[i] <= 127) {
	    /* we should never see an unrecognized ext in range 0-127 */
	    mip1dbg(("Unrecognized ext in range 0-127.\n"));
	    return(FALSE);
	} else {
	    /*
	     * Extensions in the range 128-255 should not appear
	     * after FHauth
	     */
	    if (foundFHauth) {
	        mip1dbg(("Ext found after MFauth\n"));
		return(FALSE);
	    }
	}
    }

    if (foundMHauth) {    /* make sure we have one MHauth */
        return(TRUE);
    } else {
        mip1dbg(("MHauth missing\n"));
        return(FALSE);
    }
}

/* find a faUnreachableEntry in faueTable with addr equal to haAddr */
FaUnreachableEntry * findFAUE(FaUnreachableEntry faueTable[],
			      int faueTableSize, u32 haAddr)  {
    int toSee = faueTableSize;
    int seen = 0;			       
    int i;

    for (i = 0; (i < DEFAULT_MAX_UNREACHABLES) && (seen < toSee); i++) {
        if (faueTable[i].faUnreachableTimeRemaining) {
	    /* Entry in use. */
	    seen++;
	    if (faueTable[i].faUnreachableAddr == haAddr)
	        return(&faueTable[i]);
	}
    }
			       
    return(NULL);			       
}

/* 
 * Create a pending visitor entry in faveTable based on a registration
 * request pointed to by reqPtr. The visitor is reachable through the
 * interface with address localAddr and the request was sent from
 * visitor's port vPort and source address vSrc.
 */
int mkPendingFAVE(FaVisitorEntry faveTable[], int *faveTableSize, 
		  regRequest *reqPtr, u32 localAddr, u32 vSrc,
		  unsigned short vPort)
{
    int found;
    int empty;
    int i;

    found = FALSE;
    empty = (-1);
   
    /* TODO: make this more efficient */
    /* Find a place for this entry, perhaps overwriting an existing one */
    for (i = 0; (i < DEFAULT_MAX_VISITORS) && !found; i++) {
        if (faveTable[i].faVisitorTimeRemaining) {
	    /* Entry in use. */
	    if ((faveTable[i].faVisitorHomeAddr == reqPtr->homeAddr) &&
		(faveTable[i].faVisitorRegIsAccepted == FALSE)) {
	        found = TRUE;
		break;
	    }
	} else {
	    if (empty == -1)
	        empty = i;      /* remember an empty slot */
	}
    }

    if (!found) {
        if (empty == -1) {
	    /* did not find existing or empty entry  */
	    return(-1);
	} else {
	    *faveTableSize += 1;
	    i = empty;
	}
    }

    faveTable[i].faVisitorAddr = vSrc;
    faveTable[i].faVisitorIfaceAddr = localAddr;
    faveTable[i].faVisitorRegIsAccepted = FALSE;
    faveTable[i].faVisitorRegFlags = reqPtr->regFlags;
    faveTable[i].faVisitorPort = vPort;
    faveTable[i].faVisitorHomeAddr = reqPtr->homeAddr;
    faveTable[i].faVisitorHomeAgentAddr = reqPtr->haAddr;
    faveTable[i].faVisitorCOAddr = reqPtr->COAddr;
    faveTable[i].faVisitorTimeRemaining = DEFAULT_VISITOR_EXPIRY;
    faveTable[i].faVisitorRegIDHigh = ntohl(reqPtr->IDHigh);
    faveTable[i].faVisitorRegIDLow = ntohl(reqPtr->IDLow);

    mip2dbg(("Adding pending visitor entry for %s.%d at pos'n %d.\n",
	    ntoa(reqPtr->homeAddr, addr1), vPort, i));
    return 0;
}

/*
 * Check if mnAddr has an accepted visitor entry with VisitorIfaceAddr
 * equal to IfaceAddr.
 */
boolean isAcceptedVisitor(FaVisitorEntry faveTable[], int faveTableSize,
			  u32 mnAddr, u32 ifaceAddr)
{
    int i;
    int found = FALSE;

    /* TODO: minimize the search by using faveTableSize */
    for (i = 0; (i < DEFAULT_MAX_VISITORS) && !found; i++) {
        /*
	 * It is important to compare mnAddr to faVisitorHomeAddr
	 * rather than faVisitorAddr (the latter is the src of
	 * reg request and may be different from the mobile node's
	 * home address.
	 */
        if ((faveTable[i].faVisitorTimeRemaining > 0) &&
	    (faveTable[i].faVisitorHomeAddr == mnAddr) &&
	    (faveTable[i].faVisitorIfaceAddr == ifaceAddr) &&
	    (faveTable[i].faVisitorRegIsAccepted == TRUE))
	    found = TRUE;
    }

    return (found);
}


/* Process a registration request received at a foreign agent. */
void FAprocessRegRequest(unsigned char *pkt, int pktLen, u32 src,
			 unsigned short srcPort, int ifidx)
{
    int i;
    int code;
    int extCnt;
    int val;
    boolean routeExists;
    unsigned char *extp;
    unsigned char extType[MAX_EXPECTED_EXTENSIONS];
    unsigned char *extIdx[MAX_EXPECTED_EXTENSIONS];
    regRequest *reqPtr = (regRequest *) pkt;
    regReply *repPtr;
    int fltSeq, fltCnt;

    mip2dbg(("\n---- %s (%s) ----\n", sprintTime(currentTime, 30),
	     sprintRelativeTime(relativeTime, 30)));
    mip2dbg(("FA got reg req [MN %s, HA %s, COA %s]\n",
	     ntoa(reqPtr->homeAddr, addr1), ntoa(reqPtr->haAddr, addr2),
	     ntoa(reqPtr->COAddr, addr3)));
    mip2dbg(("               [Lifetime %d sec, ID %0#10x : %0#10x]\n",
             (u32) ntohs(reqPtr->regLifetime), ntohl(reqPtr->IDHigh),
	     ntohl(reqPtr->IDLow)));

    mip3dbg(("FAprocessRegRequest called with packet:\n"));
    if (mipdebug > 2) printBuffer(pkt, pktLen);
    mip3dbg(("\n"));

    /* If we get here, pkt size is atleast equal to sizeof(regRequest) */
    if ((reqPtr->regFlags & 0x03) ||
	((reqPtr->homeAddr == reqPtr->COAddr) && (reqPtr->regLifetime != 0))) {
        code = FA_POORLY_FORMED_REQUEST;
	faPoorlyFormedRequestsCnt++;
	goto reject;
    };

    /* Are extensions ok and in the right order? */
    extp = (pkt + sizeof(regRequest));

    extCnt = MAX_EXPECTED_EXTENSIONS;
    if ((mkRegExtList(extp, (pktLen - sizeof(regRequest)),
		      &extCnt, extType, extIdx) < 0) ||
	(MnFaOkExtList(extCnt, extType) == FALSE)) {
        code = FA_POORLY_FORMED_REQUEST;
	faPoorlyFormedRequestsCnt++;
        goto reject;
    };

    mip3dbg(("Found %d extensions:", extCnt));
    for (i = 0; i < extCnt; i++) {
        mip3dbg((" <%d, %d, ...>", extType[i], *(extIdx[i]+1)));
    }
    mip3dbg(("\n"));

    /* What about MN-FA authentication? */
    /*
     * If a Mobile node Foreign agent authentication extension exists
     * check it. For now, we don't expect one and if we find one
     * we consider it an authentication failure.
     */
    for (i = 0; i < extCnt; i++) {
        if (extType[i] == REG_MF_AUTH_EXT_TYPE) {
	    code = FA_MN_AUTH_FAILURE;
	    faMNAuthFailureCnt++;
	    goto reject;
	}
    }

    /* Is lifetime too long? */
    if (ntohs(reqPtr->regLifetime) > (unsigned short) regLifetime) {
        code = FA_REG_LIFETIME_TOO_LONG;
	faRegLifetimeTooLongCnt++;
	goto reject;
    }

    /* Do we offer the requested encapsulation services? (Minimal Encap) */
    if ((reqPtr->regFlags & REG_MIN_ENCAP) &&
	((maAdvConfigTable[ifidx].maAdvServiceFlags & ADV_MIN_ENCAP)==0)){
        code = FA_ENCAP_UNAVAILABLE;
	faEncapUnavailableCnt++;
	goto reject;
    }

    /* Do we offer the requested encapsulation services? (GRE) */
    if ((reqPtr->regFlags & REG_GRE_ENCAP) &&
	((maAdvConfigTable[ifidx].maAdvServiceFlags & ADV_GRE_ENCAP)==0)){
        code = FA_ENCAP_UNAVAILABLE;
	faEncapUnavailableCnt++;
	goto reject;
    }

    /* Do we offer the requested compression service? */
    if ((reqPtr->regFlags & REG_VJ_COMPRESSION) &&
	((maAdvConfigTable[ifidx].maAdvServiceFlags & ADV_VJ_COMPRESSION)==0)){
        code = FA_VJ_UNAVAILABLE;
	faVJCompUnavailableCnt++;
	goto reject;
    }

    /* Is HA reachable? */
    if (findFAUE(faUnreachableTable, faUnreachableTableSize,
		 reqPtr->haAddr) != NULL)  {
	code = FA_HA_UNREACHABLE;
        faHAUnreachableCnt++;
        goto reject;
    }

    /*
     * TODO: If the HA address specified is the same as the arriving
     *       interface, we MUST not forward this packet. What about
     *       the case when this address belongs to another interface
     *       on this FA, one on which HA services are being offered.
     */
    if (reqPtr->haAddr == reqPtr->COAddr) {
        mip0dbg(("Warning: FA dropping nasty reg req with HAaddr same "\
                 "as COA.\a\n"));
        return;
    }

    /* Try creating a visitor list entry ... */
    if (mkPendingFAVE(faVisitorTable, &faVisitorTableSize, reqPtr,
		      maAdvConfigTable[ifidx].maIfaceAddr, src, srcPort) < 0) {
        code = FA_INSUFFICIENT_RESOURCES;
	goto reject;
    }

    // extract the filter data from the packet
    if(!extractFilters(extCnt, extType, extIdx, boundMNTable, &boundMNTableSize, src, reqPtr->COAddr)) {
        code = FA_INSUFFICIENT_RESOURCES;
	goto reject;
    }

    /*
     * TODO: Remove FA-specific extensions, add any new ones and relay the
     * request. For now, this means stripping everything after MN-HA auth
     * extension. Need to send it out of a socket where address is picked
     * up from outgoing iface.
     */
    faRegReqRecvdCnt++;
    for (i = 0; i < extCnt; i++) {
        if (extType[i] == REG_MH_AUTH_EXT_TYPE) break;
    }

    if (sendUDPmessage(maAdvConfigTable[ifidx].maIfaceUnicastSock, pkt,
		       ((extIdx[i] - pkt) + 2 + *(extIdx[i]+1)),
                       reqPtr->haAddr, MIP_PORT) == 0) {

        mip2dbg(("\n---- %s (%s) ----\n", sprintTime(currentTime, 30),
		 sprintRelativeTime(relativeTime, 30)));
	mip2dbg(("FA relayed reg req from %s.%d to HA %s.%d\n",
		 ntoa(src, addr1), srcPort, ntoa(reqPtr->haAddr, addr2),
		 MIP_PORT));

	mip3dbg(("Registration relayed by FA is:\n"));
	if (mipdebug > 2) printBuffer(pkt, pktLen);
	mip3dbg(("\n"));

	// display and apply the filters, if they have changed
	if(haveFiltersChanged(boundMNTable, src)) {
		dispFilters(boundMNTable, src);
		FAapplyFilters(boundMNTable, src);
	}

        faRegReqRelayedCnt++;
    } else {
        mip1dbg(("sendto failed at FA while relaying registration.\n"));
    }

    printf("\n\n------ FA Registration Request Received & Processed -----\n");
    printf("HA Binding Table Size %d\n", haBindingTableSize);
    printf("HA Mobile Node Table Size %d\n", haMobileNodeTableSize);
    printf("Filter Table Size %d\n", boundMNTableSize);
    printf("FA Visitor Table Size %d\n", faVisitorTableSize);
    printf("FA Unreachable Table Size %d\n", faUnreachableTableSize);
    printf("---------------------------------------------------------\n\n");

    mip2dbg(("\n\n"));

    sendInfoToAllGUIs();

    return;

 reject:
    /* Send a denial. For now, we'll create the new packet in place. */
    repPtr = (regReply *) pkt;
    repPtr->type = REG_REPLY_TYPE;
    repPtr->code = code;
    if (code == FA_REG_LIFETIME_TOO_LONG)
       repPtr->regLifetime=htons(maAdvConfigTable[ifidx].maAdvMaxRegLifetime);

    /* Pull up ID */
    bcopy(pkt+16, pkt+12, 8);

    /*
     * TODO: handle extensions? keep in mind that the request is
     * poorly formed so we may not be able to do the right thing.
     */

    routeExists = isAcceptedVisitor(faVisitorTable, faVisitorTableSize,
				    repPtr->homeAddr,
				    maAdvConfigTable[ifidx].maIfaceAddr);
    if (routeExists == FALSE) {
        mip3dbg(("Adding temporary route for visitor %s through %s\n",
		 ntoa(repPtr->homeAddr, addr1),
		 ntoa(maAdvConfigTable[ifidx].maIfaceAddr, addr2)));
	if ((val = routeadd(repPtr->homeAddr,
		     maAdvConfigTable[ifidx].maIfaceAddr, 0)) < 0) {
	    mip0dbg(("routeadd failed ... %s\n", err2str(val)));
	}
    }

    if (sendUDPmessage(maAdvConfigTable[ifidx].maIfaceUnicastSock, pkt,
                       sizeof(regReply), src, srcPort) == 0) {
        mip2dbg(("\n---- %s (%s) ----\n", sprintTime(currentTime, 30),
		 sprintRelativeTime(relativeTime, 30)));
	mip2dbg(("FA denied reg req from %s.%d (Code %d)\n",
		 ntoa(src, addr1), srcPort, repPtr->code));

	mip3dbg(("Registration denial sent by FA is:\n"));
	if (mipdebug > 2) printBuffer(pkt, pktLen);
	mip3dbg(("\n"));

    } else {
        mip2dbg(("sendto failed at FA while sending denial.\n"));
    }

    if (routeExists == FALSE) {
        mip3dbg(("Deleting temporary route for visitor %s through %s\n",
		 ntoa(repPtr->homeAddr, addr1),
		 ntoa(maAdvConfigTable[ifidx].maIfaceAddr, addr2)));
        if ((val = routedel(repPtr->homeAddr,
			    maAdvConfigTable[ifidx].maIfaceAddr)) < 0) {
	    mip0dbg(("routedel failed ... %s\n", err2str(val)));
        }
    }
    mip2dbg(("\n\n"));

    sendInfoToAllGUIs();

}

/*
 * Delete a binding entry for the specified mnAddr in habTable. If the
 * care-of address COAddr equals mnAddr, all bindings for the mobile node
 * are deleted, empty is set to the index of an empty binding entry
 */
int delHABE(HaBindingEntry habTable[], int *habTableSize,
	    u32 mnAddr, u32 COAddr, int *empty)
{
    int i;

    mip3dbg(("delHABE called for mnAddr %s COAddr %s\n",
	     ntoa(mnAddr, addr1), ntoa(COAddr, addr2)));

    *empty = -1;
    for (i = 0; i < DEFAULT_MAX_BINDINGS; i++) {
        if (habTable[i].haBindingTimeRemaining) {
	    /* Entry in use. */
	    if ((habTable[i].haBindingMN == mnAddr) &&
		((COAddr==mnAddr) || (habTable[i].haBindingCOA == COAddr)))  {
		 /* Found a match, delete it */
	         delHABEidx(habTable, habTableSize, i);
		 if (*empty == -1)
		     *empty = i;
	    }
	} else {
	    if (*empty == -1)
	        *empty = i;
	}
    }
    return 0;
}

/*
 * Find a binding entry for the specified mnAddr and coAddr pair in
 * habTable and return its index in idx. If no match is found, return
 * the index of an empty entry in Idx. If the table is full and no
 * match is found, set idx to -1. If a match is found findHABE returns
 * TRUE else FALSE.
 */
int findHABE(HaBindingEntry habTable[], int *habTableSize,
	    u32 mnAddr, u32 COAddr, int *idx)
{
    int i;
    boolean found = FALSE;
    int empty = (-1);

    /* TODO: make this more efficient */
    for (i = 0; i < DEFAULT_MAX_BINDINGS; i++) {
        if (habTable[i].haBindingTimeRemaining) {
	    /* Entry in use. */
	    if ((habTable[i].haBindingMN == mnAddr) &&
		(habTable[i].haBindingCOA == COAddr))  {
		 /* Found an existing entry, use it  */
		 found = TRUE;
		 break;
	    }
	} else  {
	    if (empty == -1)
	        empty = i;      /* remember an empty slot */
	}
    }

    if (!found) {
        *idx = empty;
    } else {
        *idx = i;
    }

    return (found);
}

/*
 * Add a binding for the specified pair of mobile node and care-of
 * address. By the time we get here, we know that we are processing
 * a registration and coAddr is not the same as mnAddr. Note that ifidx
 * is the index for the mobility interface on the mobile node's
 * home network.
 */
int addHABE(HaBindingEntry habTable[], int *habTableSize, regRequest *reqPtr,
	    u32 src, unsigned short srcPort, int ifidx)
{
    boolean found;          /* true if a previous binding already exists */
    int  usable = -1;       /* marks a usable slot for the binding */
    int val;
    int i;
    unsigned char regFlags = reqPtr->regFlags;
    HaMobileNodeEntry *hamnePtr = NULL;


    found = FALSE;

#ifdef NO_SIMULTANEOUS
    /* this ensures that all previous bindings will be cleared */
    regFlags &= ~REG_SIMULTANEOUS_BINDINGS;
#endif
    regFlags &= REG_SIMULTANEOUS_BINDINGS;

    for (i = 0; i < DEFAULT_MAX_BINDINGS; i++) {
        if (habTable[i].haBindingTimeRemaining == 0) {
	    /* Empty entry */
	    if (usable == -1) usable = i;
	    continue;
	}
	/* Entry in use. */
	if (habTable[i].haBindingMN == reqPtr->homeAddr) {
	    if (habTable[i].haBindingCOA == reqPtr->COAddr)  {
	        /* Found a match, update the lifetime */
	        found = TRUE;
		habTable[i].haBindingSrcAddr = src;
		/*
		 * TODO: we should use the maAdvMaxAdvLifetime field for
		 * this interface instead of regLifetime once we allow
		 * each interface to advertise a different value.
		 */
		habTable[i].haBindingTimeGranted =
		    (ntohs(reqPtr->regLifetime)>(unsigned short) regLifetime)?
		    (u32) regLifetime : (u32) ntohs(reqPtr->regLifetime);
		habTable[i].haBindingTimeRemaining =
		    habTable[i].haBindingTimeGranted;
		habTable[i].haBindingSrcPort = srcPort;
		habTable[i].haBindingRegFlags = reqPtr->regFlags;

		if ((hamnePtr = findHAMNE(haMobileNodeTable,
					  haMobileNodeTableSize,
					  reqPtr->homeAddr)) == NULL) {
		    mip0dbg(("Error1: no mobile node entry when doing " \
			     "addHABE().\n"));
		    continue;
		}

		mip2dbg(("HA renewed binding %s@%s at pos'n %d " \
			 "for %d sec (Binding cnt %d).\n",
			 ntoa(reqPtr->homeAddr, addr1),
			 ntoa(reqPtr->COAddr, addr2), i,
			 habTable[i].haBindingTimeGranted,
			 hamnePtr->haMnBindingCnt));
	    } else {
	        /* found an entry for this MN with another COA */
	        if ((regFlags & REG_SIMULTANEOUS_BINDINGS) == 0) {
			fprintf(stderr, "No Simul\n");
		    delHABEidx(habTable, habTableSize, i);
		    if (usable == -1) usable = i;
		}
	    }
	}
    }  /* end of for loop */

    if (found) goto applyflt; //return(0);

    if (usable == -1) {
        mip0dbg(("Binding table full at HA.\n"));
        return (-1);     /* sorry, table is full */
    }

    if ((hamnePtr = findHAMNE(haMobileNodeTable, haMobileNodeTableSize,
			      reqPtr->homeAddr)) == NULL) {
        mip0dbg(("Error2: no mobile node entry when doing addHABE().\n"));
	return(-1);
    }

    /*
     * We need to create a new entry ...
     * update binding count for this mobile node and total bindings
     */
    hamnePtr->haMnBindingCnt += 1;
    *habTableSize += 1;

    mip3dbg(("At %s, creating tunnel for %s through %s %s\n",
	     ntoa(reqPtr->haAddr,addr1), ntoa(reqPtr->homeAddr, addr2),
	     ntoa(reqPtr->COAddr, addr3),
	     (regFlags & REG_SIMULTANEOUS_BINDINGS) ? "(simultaneous)":"" ));

    if ((val = encapadd(reqPtr->homeAddr, reqPtr->haAddr, reqPtr->COAddr,
		 (regFlags & REG_SIMULTANEOUS_BINDINGS))) < 0) {
        mip0dbg(("encapadd failed ... %s\n", err2str(val)));
    }

    //if (hamnePtr->haMnBindingCnt == 1) {   // add route entry for all bindings in MSB
        mip3dbg(("Enabling tunneling for %s.\n",
		 ntoa(reqPtr->homeAddr, addr1)));

	// set routing entries based on filters, standard route setting
	// has been stopped
	// if ((val = mkroutethroughtunnel(reqPtr->homeAddr, reqPtr->haAddr, reqPtr->COAddr)) < 0) { // MSB
	//    mip0dbg(("mkroutethroughtunnel failed ... %s\n", err2str(val)));
	// }

	mip3dbg(("Setting proxy arp for %s at %s\n",
		 ntoa(reqPtr->homeAddr, addr1),
		 hwAddrWrite(maAdvConfigTable[ifidx].maIfaceHWaddr, addr2)));

	if ((val = arpadd(reqPtr->homeAddr,
			  maAdvConfigTable[ifidx].maIfaceHWaddr,
			  "pub")) < 0) {
	    mip0dbg(("arpadd (proxy) failed ... %s\n", err2str(val)));
	}

	/*
	 * Solaris 2.6 automatically sends out a gratuitous ARP
	 * when a publishable ARP entry is created and in another
	 * OS, we expect arpadd to do so explicitly.
	 */
	haGratuitousARPsSentCnt += 1;

	mip3dbg(("Enabling decapsulation of pkts with inner src %s\n",
		 ntoa(reqPtr->homeAddr, addr1)));
	/*
	 * TODO: decapadd should take an addr and flags rather than
	 * a string
	 */
	if ((val = decapadd(reqPtr->COAddr, reqPtr->haAddr, reqPtr->homeAddr, "src")) < 0) {
	    mip0dbg(("decapadd failed ... %s\n", err2str(val)));
	}
    //}

    /*
     * If the care-of-address is outside the protected domain then
     * encapsulate and redirect those packets to the firewall's internal
     * address.
     * TODO: This will likely change since we should not delete the tunneling
     *       of a COA through FW unless *ALL* mobile nodes using that COA
     *       have their bindings expire, i.e. we need to keep a count of
     *       the number of MNs using that COA. We should be OK as long as an
     *       external COA is not shared by multiple mobile nodes.
     */
    if (isOutsideProtectedDomain(reqPtr->COAddr)) {
        mip3dbg(("At %s, creating tunnel for %s through firewall %s 0\n",
	          ntoa(reqPtr->haAddr,addr1), ntoa(reqPtr->COAddr, addr2),
	          ntoa(domainInfo.fwAddr[0], addr3)));
        if ((val = encapadd(reqPtr->COAddr,
		   reqPtr->haAddr, domainInfo.fwAddr[0], 0)) < 0) {
            mip0dbg(("encapadd failed ... %s\n", err2str(val)));
        }

	mip3dbg(("Enabling tunneling for %s.\n",
		 ntoa(reqPtr->COAddr, addr1)));
	if ((val = mkroutethroughtunnel(reqPtr->COAddr, reqPtr->haAddr, domainInfo.fwAddr[0])) < 0) { // MSB
	    mip0dbg(("mkroutethroughtunnel failed ... %s\n", err2str(val)));
	}
    }

    habTable[usable].haBindingMN = reqPtr->homeAddr;
    habTable[usable].haBindingCOA = reqPtr->COAddr;
    habTable[usable].haBindingSrcAddr = src;

    /*
     * TODO: we should use the maAdvMaxAdvLifetime field for this
     * interface instead of regLifetime once we allow each interface
     * to advertise a different value.
     */
    habTable[usable].haBindingTimeGranted =
           (ntohs(reqPtr->regLifetime) > (unsigned short) regLifetime) ?
           (u32) regLifetime : (u32) ntohs(reqPtr->regLifetime);
    habTable[usable].haBindingTimeRemaining =
                          habTable[usable].haBindingTimeGranted;
    habTable[usable].haBindingSrcPort = srcPort;
    habTable[usable].haBindingRegFlags = reqPtr->regFlags;

    mip2dbg(("HA added binding %s@%s at pos'n %d " \
	     "for %d sec (Binding cnt %d).\n",
	     ntoa(reqPtr->homeAddr, addr1), ntoa(reqPtr->COAddr, addr2),
	     usable, habTable[usable].haBindingTimeGranted,
	     hamnePtr->haMnBindingCnt));

    // when a new binding is registered, always add a MN entry to
    // the filter list if not present
    addFilterTableEntry(boundMNTable, reqPtr->homeAddr);

applyflt:

    // apply filters as an HA if changed
    if(haveFiltersChanged(boundMNTable, reqPtr->homeAddr)) {
    	if(!HAapplyFilters(boundMNTable, haBindingTable, reqPtr->homeAddr)) {
		mip0dbg(("mkroutethroughtunnels failed ... %s\n", err2str(val)));
    	}
    }

    return(0);
}

/* Process a registration request received at a home agent. */
void HAprocessRegRequest(unsigned char *pkt, int pktLen, u32 src,
			 unsigned short srcPort, int ifidx)
{
    int i;
    int code;
    int repLen;
    int okID = FALSE;
    int extCnt;
    int val;
    int empty;
    u32 IDhi;
    u32 IDlo;
    boolean ignoreTunnel;
    boolean encapToFw;
    int habeTblidx;
    u32 COAddr;
    u32 haAddr;
    u32 homeAddr;
    unsigned char *extp;
    unsigned char extType[MAX_EXPECTED_EXTENSIONS];
    unsigned char *extIdx[MAX_EXPECTED_EXTENSIONS];
    regRequest *reqPtr = (regRequest *) pkt;
    regReply *repPtr;
    HaMobileNodeEntry *hamnePtr = NULL;
    int fltSeq, fltCnt;

    mip2dbg(("\n---- %s (%s) ----\n", sprintTime(currentTime, 30),
	     sprintRelativeTime(relativeTime, 30)));
    mip2dbg(("HA got reg req [MN %s, HA %s, COA %s]\n",
	     ntoa(reqPtr->homeAddr, addr1), ntoa(reqPtr->haAddr, addr2),
	     ntoa(reqPtr->COAddr, addr3)));
    mip2dbg(("               [Lifetime %d sec, ID %0#10x : %0#10x]\n",
             (u32) ntohs(reqPtr->regLifetime), ntohl(reqPtr->IDHigh),
	     ntohl(reqPtr->IDLow)));

    mip3dbg(("HAprocessRegRequest called for pkt:\n"));
    if (mipdebug > 2) printBuffer(pkt, pktLen);
    mip3dbg(("\n"));


    /*
     * If we get here, pkt size is atleast equal to sizeof(regRequest)
     * NOTE: For now, only our HAs support reverse tunneling and having
     *       the T bit set in incoming registrations at the HA should
     *       not be a reason for complaint.
     */
    if ((reqPtr->regFlags & 0x01) ||
	((reqPtr->homeAddr == reqPtr->COAddr) && (reqPtr->regLifetime != 0))) {
        code = HA_POORLY_FORMED_REQUEST;
	haPoorlyFormedRequestsCnt++;
	goto reply;
    };

    /* Are extensions ok and in the right order? */
    extp = (pkt + sizeof(regRequest));
    extCnt = MAX_EXPECTED_EXTENSIONS;
    if ((mkRegExtList(extp, (pktLen - sizeof(regRequest)),
		      &extCnt, extType, extIdx) < 0) ||
	(FaHaOkExtList(extCnt, extType) == FALSE)) {
        code = HA_POORLY_FORMED_REQUEST;
	haPoorlyFormedRequestsCnt++;
        goto reply;
    };

    mip3dbg(("Found %d extensions:", extCnt));
    for (i = 0; i < extCnt; i++) {
        mip3dbg((" <%d, %d, ...>", extType[i], *(extIdx[i]+1)));
    }
    mip3dbg(("\n"));

    /* Was HA a broadcast address? */
    if (reqPtr->haAddr != maAdvConfigTable[ifidx].maIfaceAddr) {
        if ((hamnePtr = findHAMNE(haMobileNodeTable, haMobileNodeTableSize,
				  reqPtr->homeAddr)) == NULL) {
	    mip1dbg(("Dropping discovery request from unsupported MN %s.\n",
		     ntoa(reqPtr->homeAddr, addr1)));
	    return;
	} else {
	    code = HA_UNKNOWN_HOME_AGENT;
	    haUnknownHACnt++;
	    goto reply;
	}
    }

    /* Is MN supported? */
    if ((hamnePtr = findHAMNE(haMobileNodeTable, haMobileNodeTableSize,
			      reqPtr->homeAddr)) == NULL) {
        code = HA_ADM_PROHIBITED;
	haAdmProhibitedCnt++;
	goto reply;
    }

    /* What about MN-HA authentication? */
    for (i = 0; i < extCnt; i++) {
        if (extType[i] == REG_MH_AUTH_EXT_TYPE) break;
    }

    if (!isAuthExtOk(pkt, (extIdx[i]-pkt), &(hamnePtr->haMnMSA))) {
        code = HA_MN_AUTH_FAILURE;
	haMNAuthFailureCnt++;
	goto reply;
    }

    /* TODO:
     * If a Foreign agent Home Agent authentication extension exists
     * check it. For now, we don't expect one and if we find one
     * we consider it an authentication failure.
     */
    for (i = 0; i < extCnt; i++) {
        if (extType[i] == REG_FH_AUTH_EXT_TYPE) {
	    code = HA_FA_AUTH_FAILURE;
	    haFAAuthFailureCnt++;
	    goto reply;
	}
    }

    /* Check ID used for replay protection */
    okID = HAisIDok(hamnePtr->haMnRegIDHigh, hamnePtr->haMnRegIDLow,
		    ntohl(reqPtr->IDHigh), ntohl(reqPtr->IDLow),
		    hamnePtr->haMnMSA.mipSecReplayMethod);

    if (okID == FALSE) {
        code = HA_ID_MISMATCH;
	haIDMismatchCnt++;
	goto reply;
    }

    if ((reqPtr->regFlags & REG_SIMULTANEOUS_BINDINGS) &&
	(hamnePtr->haMnBindingCnt == MAX_SIMULTANEOUS_BINDINGS)) {
        code = HA_TOO_MANY_SIMULTANEOUS;
	haTooManyBindingsCnt++;
	goto reply;
    }

    // extract the filter data from the packet
    if(!extractFilters(extCnt, extType, extIdx, boundMNTable, &boundMNTableSize, reqPtr->homeAddr, reqPtr->COAddr)) {
        code = FA_INSUFFICIENT_RESOURCES;
	goto reply;
    }

    // display the filters, if they have changed for this HA
    if(haveFiltersChanged(boundMNTable, reqPtr->homeAddr)) {
	dispFilters(boundMNTable, reqPtr->homeAddr);
    }

	//fprintf(stderr,"FooBar\n");
    if (reqPtr->regLifetime == 0) {
        haDeRegReqRecvdCnt++;
	delHABE(haBindingTable, &haBindingTableSize, reqPtr->homeAddr,
		reqPtr->COAddr, &empty);
	//delFilter(bndFltTable, &bndFltCount, reqPtr->COAddr);
    } else {
        haRegReqRecvdCnt++;
	if (addHABE(haBindingTable, &haBindingTableSize, reqPtr,
		    src, srcPort, ifidx) < 0) {
	    code = HA_INSUFFICIENT_RESOURCES;
	    goto reply;
	}
		//fprintf(stderr,"FooBar3\n");
    }


	//fprintf(stderr,"FooBar3\n");
    /*
     * Choose the appropriate acceptance code depending on whether or
     * not simultaneous bindings are supported.
     */

    code = 0;

#ifdef NO_SIMULTANEOUS
    if (reqPtr->regFlags & REG_SIMULTANEOUS_BINDINGS)
        code = 1;
#endif

 reply:
    /*
     * For now, we'll create the new packet in place. But not before we
     * save some of the fields that could get overwritten.
     */
    COAddr = reqPtr->COAddr;  /* this is the only one we risk losing */
    haAddr = reqPtr->haAddr;
    homeAddr = reqPtr->homeAddr;

    repPtr = (regReply *) pkt;
    repPtr->type = REG_REPLY_TYPE;
    repPtr->code = code;
    if (ntohs(reqPtr->regLifetime) >
	maAdvConfigTable[ifidx].maAdvMaxRegLifetime) {
        repPtr->regLifetime =
	  htons(maAdvConfigTable[ifidx].maAdvMaxRegLifetime);
    } else {
      repPtr->regLifetime = reqPtr->regLifetime;
    }
    repPtr->homeAddr = reqPtr->homeAddr;
    repPtr->haAddr = maAdvConfigTable[ifidx].maIfaceAddr;
    HAnewID(&IDhi, &IDlo, ntohl(reqPtr->IDHigh), ntohl(reqPtr->IDLow),
	    (hamnePtr == NULL) ? NONE : hamnePtr->haMnMSA.mipSecReplayMethod,
	    okID);
    if (hamnePtr != NULL) {
        HAstoreID(&(hamnePtr->haMnRegIDHigh), &(hamnePtr->haMnRegIDLow),
		  IDhi, IDlo, hamnePtr->haMnMSA.mipSecReplayMethod, okID);
    }
    repPtr->IDHigh = htonl(IDhi);
    repPtr->IDLow = htonl(IDlo);
    repLen = sizeof(regReply);

    repLen += appendAuthExt(pkt, sizeof(regReply), REG_MH_AUTH_EXT_TYPE,
			    (hamnePtr==NULL) ? NULL : &(hamnePtr->haMnMSA));

    /*
     * First assume nothing special is needed to ensure that denials reach
     * the mobile node.
     */
    encapToFw = FALSE;
    ignoreTunnel = FALSE;

    /* Is this really the case? */
    if ((repPtr->code != 0) && (repPtr->code != 1)) {
        /*
	 * For "deregister all" requests, ignore the tunnel to ensure
	 * denial is sent on the local (home) network.
	 */
         if ((repPtr->regLifetime == 0) &&
	     (COAddr == homeAddr) &&
	     hamnePtr && (hamnePtr->haMnBindingCnt)) {
	     ignoreTunnel = TRUE;
	 }

	 /*
	  * If MN is using an external COA and there isn't a tunnel to
	  * FW already, create one temporarily.
	  */
	 if (isOutsideProtectedDomain(COAddr) &&
	     ((findHABE(haBindingTable, &haBindingTableSize, homeAddr,
			COAddr, &habeTblidx) == FALSE))) {
	     encapToFw = TRUE;
	 }
    }

    /*
     * If needed, set up temporary tunnel to redirect packets for external COAs
     * at FW.
     * TODO: Note that this works only when the destination of the
     *       registration request is the COAddr. This is the case
     *       for an MN operating in colocated state so we are fine
     *       for now.
     */
    if (encapToFw) {
        mip3dbg(("At %s, creating a temporary tunnel for %s through " \
		 "firewall %s 0\n", ntoa(haAddr,addr1), ntoa(COAddr, addr2),
	          ntoa(domainInfo.fwAddr[0], addr3)));
        if ((val = encapadd(COAddr, haAddr, domainInfo.fwAddr[0], 0)) < 0) {
	    mip0dbg(("encapadd failed ... %s\n", err2str(val)));
        }
	if ((val = mkroutethroughtunnel(COAddr, haAddr, domainInfo.fwAddr[0])) < 0) { // MSB
	    mip0dbg(("mkroutethroughtunnel failed ... %s\n", err2str(val)));
	}
    }

    /* If needed, ignore existing tunnel for MN to COA */
    if (ignoreTunnel) {
        /* make sure response will be sent locally */
        mip3dbg(("Temporarily deleting ARP entry for %s\n",
		 ntoa(repPtr->homeAddr, addr1)));
        if ((val = arpdel(repPtr->homeAddr)) < 0) {
	    mip0dbg(("arpdel failed ... %s\n", err2str(val)));
        }

        mip3dbg(("Temporarily suspending tunneling for %s\n",
		 ntoa(repPtr->homeAddr, addr1)));
        if ((val = rmroutethroughtunnel(repPtr->homeAddr, haAddr, COAddr)) < 0) {
	    mip0dbg(("rmroutethroughtunnel failed ... %s\n", err2str(val)));
        }
    }

    if (sendUDPmessage(maAdvConfigTable[ifidx].maIfaceUnicastSock, pkt,
		       repLen, src, srcPort) == 0) {

        mip2dbg(("\n---- %s (%s) ----\n", sprintTime(currentTime, 30),
		 sprintRelativeTime(relativeTime, 30)));
	mip2dbg(("HA sent reg reply to %s.%d (Code %d)\n",
		 ntoa(src, addr1), srcPort, repPtr->code));
	mip2dbg(("               [Lifetime %d sec, ID %0#10x : %0#10x]\n",
		 (u32) ntohs(repPtr->regLifetime),
		 ntohl(repPtr->IDHigh), ntohl(repPtr->IDLow)));

	mip3dbg(("HA's reply is:\n"));
	if (mipdebug > 2) printBuffer((unsigned char *) pkt, repLen);
	mip3dbg(("\n"));

	if (repPtr->regLifetime != 0)
	    haRegRepliesSentCnt++;
	else
	    haDeRegRepliesSentCnt++;

    } else {
        mip1dbg(("sendto failed at HA while replying.\n"));
    }

    /* Undo any temporary changes related to "deregistration all" messages. */
    if (ignoreTunnel) {
        /* restore prior state */
	mip3dbg(("Re-enabling tunneling for %s\n",
		 ntoa(repPtr->homeAddr, addr1)));
	if ((val = mkroutethroughtunnel(repPtr->homeAddr, haAddr, COAddr)) < 0) { // MSB
	    mip0dbg(("mkroutethroughtunnel failed ... %s\n", err2str(val)));
	}

	mip3dbg(("Restoring proxy ARP for %s at %s\n",
		 ntoa(repPtr->homeAddr, addr1),
		 hwAddrWrite(maAdvConfigTable[ifidx].maIfaceHWaddr, addr2)));
	if ((val = arpadd(repPtr->homeAddr,
			  maAdvConfigTable[ifidx].maIfaceHWaddr, "pub")) < 0) {
	    mip0dbg(("arpadd (proxy) failed ... %s\n", err2str(val)));
	}
    }

    /* Undo any temporary changes related to external COAs */
    if (encapToFw) {
        mip3dbg(("Removing temporary tunnel for %s through FW %s\n",
		 ntoa(COAddr, addr1), ntoa(domainInfo.fwAddr[0], addr2)));
        if ((val = encaprem(COAddr, domainInfo.fwAddr[0])) < 0) {
           mip0dbg(("encaprem failed ... %s\n", err2str(val)));
        }

	/*
	 * TODO: This would work only if we assume that COA is not shared
	 *       between multiple mobile nodes registered at this HA.
	 */
	if ((val = rmroutethroughtunnel(COAddr, haAddr, domainInfo.fwAddr[0])) < 0) {
	    mip0dbg(("rmroutethroughtunnel failed ... %s\n",
		     err2str(val)));
	}
    }

    /*
     * If we successfully deregistered all bindings for a MN
     * we MUST also send out a gratuitous ARP with the MN's correct
     * mapping taken from our ARP cache.
     */

    if (((repPtr->code == 0) || (repPtr->code == 1)) &&
	(repPtr->regLifetime == 0) &&
	(hamnePtr != NULL) && (hamnePtr->haMnBindingCnt == 0)) {
        if ((val = arprefresh(repPtr->homeAddr)) < 0)
	    mip0dbg(("arprefresh failed ... %s\n", err2str(val)));
    }

    printf("\n\n------ FA Registration Request Received & Processed -----\n");
    printf("HA Binding Table Size %d\n", haBindingTableSize);
    printf("HA Mobile Node Table Size %d\n", haMobileNodeTableSize);
    printf("Filter Table Size %d\n", boundMNTableSize);
    printf("FA Visitor Table Size %d\n", faVisitorTableSize);
    printf("FA Unreachable Table Size %d\n", faUnreachableTableSize);
    printf("---------------------------------------------------------\n\n");

    mip2dbg(("\n\n"));

    sendInfoToAllGUIs();
}

/*
 * Find a visitor entry with faVisitorHomeAddr equal to mnAddr,
 * faVisitorRegIDLow equal to IDlo and faVisitorRegIsAccepted equal
 * to FALSE.
 */
FaVisitorEntry * findPendingFAVE(FaVisitorEntry faveTable[], int faveTableSize,
				 u32 mnAddr, u32 IDlo)
{
    int toSee = faveTableSize;
    int seen = 0;
    int i;

    for (i = 0; (i < DEFAULT_MAX_VISITORS) && (seen < toSee); i++) {
        if (faveTable[i].faVisitorTimeRemaining) {
	    /* Entry in use. */
	    seen++;
	    if ((faveTable[i].faVisitorHomeAddr == mnAddr) &&
		(faveTable[i].faVisitorRegIDLow == IDlo) &&
		(faveTable[i].faVisitorRegIsAccepted == FALSE))
	        return(&faveTable[i]);
	}
    }

    return(NULL);
}

/*
 * Delete a visitor entry matching the specified mnAddr and coAddr in
 * faveTable. If the coAddr equals mnAddr, all entries for that
 * mobile node are deleted.
 */
int delFAVE(FaVisitorEntry faveTable[], int *faveTableSize, u32 mnAddr,
	    u32 COAddr)
{
    int toSee = *faveTableSize;
    int seen = 0;
    int i;

    /*
     * TODO: make this efficient, e.g. we could minimize our stay
     * in the loop if we figure out that COAddr != mnAddr and we
     * already deleted one entry.
     */
    for (i = 0; (i < DEFAULT_MAX_VISITORS) && (seen < toSee); i++) {
        if (faveTable[i].faVisitorTimeRemaining) {
	    /* Entry in use. */
	    seen++;
	    if ((faveTable[i].faVisitorHomeAddr == mnAddr) &&
		((COAddr == mnAddr) ||
		 (faveTable[i].faVisitorCOAddr == COAddr))) {
		 /* Found a match, delete it */
	         delFAVEptr(faveTable, faveTableSize, &faveTable[i]);
		 sendInfoToAllGUIs();
	    }
	}
    }
    return 0;
}

/*
 * If we find an ACCEPTED visitor entry for this MN-COA pair, update
 * that entry and delete the pending entry. Otherwise, change the
 * "pending" entry pointed to by favep to "accepted" with specified
 * lifetime.
 *
 *  NOTE: COA is available in the faVisitorCOAddr field.
 */
int acceptFAVE(FaVisitorEntry faveTable[], int *faveTableSize, u32 mnAddr,
	       FaVisitorEntry *favep, unsigned short lifetime)
{
    int toSee = *faveTableSize;
    int seen = 0;
    int i, val;

    for (i = 0; (i < DEFAULT_MAX_VISITORS) && (seen < toSee); i++) {
        if (faveTable[i].faVisitorTimeRemaining) {
	    /* Entry in use. */
	    seen++;
	    if ((faveTable[i].faVisitorHomeAddr == mnAddr) &&
		(faveTable[i].faVisitorCOAddr == favep->faVisitorCOAddr) &&
		(faveTable[i].faVisitorRegIsAccepted == TRUE)) {
	        /* Found a match, update it */
	        faveTable[i].faVisitorAddr = favep->faVisitorAddr;
	        faveTable[i].faVisitorIfaceAddr = favep->faVisitorIfaceAddr;
	        faveTable[i].faVisitorRegFlags = favep->faVisitorRegFlags;
	        faveTable[i].faVisitorPort = favep->faVisitorPort;
	        faveTable[i].faVisitorHomeAgentAddr =
		           favep->faVisitorHomeAgentAddr;
	        faveTable[i].faVisitorTimeGranted = lifetime;
		faveTable[i].faVisitorTimeRemaining = lifetime;
	        faveTable[i].faVisitorRegIDHigh = favep->faVisitorRegIDHigh;
	        faveTable[i].faVisitorRegIDLow = favep->faVisitorRegIDLow;

		mip2dbg(("FA renewed visitor %s on iface %s (%d sec).\n",
			 ntoa(faveTable[i].faVisitorHomeAddr, addr1),
			 ntoa(faveTable[i].faVisitorIfaceAddr, addr2),
			 faveTable[i].faVisitorTimeRemaining));

		/* remove the pending entry */
		delFAVEptr(faveTable, faveTableSize, favep);
		sendInfoToAllGUIs();

		return(0);
	    }
	}
    }

    /* We did not find an existing ACCEPTED entry. */
    favep->faVisitorTimeGranted = lifetime;
    favep->faVisitorTimeRemaining = lifetime;
    favep->faVisitorRegIsAccepted = TRUE;
    mip2dbg(("FA accepted visitor %s on iface %s (%d sec).\n",
	     ntoa(favep->faVisitorHomeAddr, addr1),
	     ntoa(favep->faVisitorIfaceAddr, addr2),
	     favep->faVisitorTimeRemaining));

    mip3dbg(("Adding direct, local route for visitor %s through %s.\n",
	     ntoa(favep->faVisitorHomeAddr, addr1),
	     ntoa(favep->faVisitorIfaceAddr, addr2)));
    if ((val = routeadd(favep->faVisitorHomeAddr,
			favep->faVisitorIfaceAddr, 0)) < 0) {
        mip0dbg(("routeadd failed ... %s\n", err2str(val)));
    }

    mip3dbg(("Enabling decapsulation of inner pkts sent to %s\n",
	     ntoa(favep->faVisitorHomeAddr, addr1)));
    /* TODO: decapadd should take an addr and flags rather than a string */
    if((val = decapadd(favep->faVisitorHomeAgentAddr,
					   favep->faVisitorCOAddr,
					   favep->faVisitorHomeAddr, "dst")) < 0) {
        mip0dbg(("decapadd failed ... %s\n", err2str(val)));
    }
    sendInfoToAllGUIs();

}


/* Process a registration reply received at a foreign agent. */
void FAprocessRegReply(unsigned char *pkt, int pktLen, u32 src,
		       unsigned short srcPort, int ifIdx)
{
    int i;
    int code = 0;
    int extCnt;
    int val;
    int newLen = sizeof(regReply);
    boolean routeExists;
    unsigned char *extp;
    unsigned char extType[MAX_EXPECTED_EXTENSIONS];
    unsigned char *extIdx[MAX_EXPECTED_EXTENSIONS];
    regReply *repPtr = (regReply *) pkt;
    FaVisitorEntry *favePtr;

    mip2dbg(("\n---- %s (%s) ----\n", sprintTime(currentTime, 30),
	     sprintRelativeTime(relativeTime, 30)));
    mip2dbg(("FA got reg reply [MN %s, HA %s, Code %d]\n",
	     ntoa(repPtr->homeAddr, addr1), ntoa(repPtr->haAddr, addr2),
	     repPtr->code));
    mip2dbg(("                 [Lifetime %d sec, ID %0#10x : %0#10x]\n",
             (u32) ntohs(repPtr->regLifetime), ntohl(repPtr->IDHigh),
	     ntohl(repPtr->IDLow)));

    mip3dbg(("FAprocessRegReply called for pkt:\n"));
    if (mipdebug > 2) printBuffer(pkt, pktLen);
    mip3dbg(("\n"));

    favePtr = findPendingFAVE(faVisitorTable, faVisitorTableSize,
			      repPtr->homeAddr, ntohl(repPtr->IDLow));
    if (favePtr == NULL) {
        mip1dbg(("Did not find matching pending request.\n\n"));
	return;
    }

    /* Are extensions ok and in the right order? */
    extp = (pkt + sizeof(regReply));

    extCnt = MAX_EXPECTED_EXTENSIONS;
    if ((mkRegExtList(extp, (pktLen - sizeof(regReply)),
		      &extCnt, extType, extIdx) < 0) ||
	(FaHaOkExtList(extCnt, extType) == FALSE)) {
        code = FA_POORLY_FORMED_REQUEST;
	faPoorlyFormedRequestsCnt++;
        goto reply;
    };

    mip3dbg(("Found %d extensions:", extCnt));
    for (i = 0; i < extCnt; i++) {
        mip3dbg((" <%d, %d, ...>", extType[i], *(extIdx[i]+1)));
    }
    mip3dbg(("\n"));

    /* What about HA-FA authentication? */
    /*
     * If a Home agent Foreign agent authentication extension exists
     * check it. For now, we don't expect one and if we find one
     * we consider it an authentication failure.
     */
    for (i = 0; i < extCnt; i++) {
        if (extType[i] == REG_FH_AUTH_EXT_TYPE) {
	    code = HA_FA_AUTH_FAILURE;
	    faHAAuthFailureCnt++;
	    goto reply;
	}
    }

    /* ... it is a good reply. No need to change the code in pkt. */
    faRegRepliesRecvdCnt++;

    /*
     * TODO: Remove everything after the MN-HA auth ext, add any new ones
     * and relay the request. For now, we simply strip everything beyond
     * the MN-HA auth.
     */

    for (i = 0; i < extCnt; i++) {
        if (extType[i] == REG_MH_AUTH_EXT_TYPE) break;
    }
    newLen = (extIdx[i] - pkt) + 2 + *(extIdx[i]+1);

 reply:
    /* Do we ned to overwrite the code field? */
    if (code)
        repPtr->code = code;

    /*
     * TODO: Handle extensions properly. For now, we include no
     * extensions in denials and therefore don't need to change
     * newLen.
     */

    /*
     * Make sure that, if needed, a route is first created and then
     * deleted to convey rejections to the mobile node.
     */

    routeExists = isAcceptedVisitor(faVisitorTable, faVisitorTableSize,
				    repPtr->homeAddr,
				    favePtr->faVisitorIfaceAddr);

    if (routeExists == FALSE) {
        mip3dbg(("Adding temporary route for visitor %s through %s\n",
		 ntoa(repPtr->homeAddr, addr1),
		 ntoa(favePtr->faVisitorIfaceAddr, addr2)));
        if ((val = routeadd(repPtr->homeAddr,
			    favePtr->faVisitorIfaceAddr, 0)) < 0) {
	    mip0dbg(("routeadd failed ... %s\n", err2str(val)));
        }
    }


    if (sendUDPmessage(maAdvConfigTable[ifIdx].maIfaceUnicastSock, pkt,
		       newLen, favePtr->faVisitorAddr,
		       favePtr->faVisitorPort) == 0) {

        mip2dbg(("\n---- %s (%s) ----\n", sprintTime(currentTime, 30),
		 sprintRelativeTime(relativeTime, 30)));
	mip2dbg(("FA relayed reg reply to %s.%d (Code %d)\n",
		 ntoa(favePtr->faVisitorAddr, addr1),
		 favePtr->faVisitorPort, repPtr->code));
        faRegReqRelayedCnt++;

	mip3dbg(("FA relayed reply packet:\n"));
	if (mipdebug > 2) printBuffer(pkt, newLen);
	mip3dbg(("\n"));

    } else {
        mip1dbg(("sendto failed at FA while relaying reply.\n"));
    }

    if (routeExists == FALSE) {
        mip3dbg(("Deleting temporary route for visitor %s through %s\n",
		 ntoa(repPtr->homeAddr, addr1),
		 ntoa(favePtr->faVisitorIfaceAddr, addr2)));
        if ((val = routedel(repPtr->homeAddr,
			    favePtr->faVisitorIfaceAddr)) < 0) {
	    mip0dbg(("routedel failed ... %s\n", err2str(val)));
        }
    }

    if ((repPtr->code == 0) || (repPtr->code == 1)) {
        if (repPtr->regLifetime == 0) {
	    /* Deregistration ... delete visitor entries for this MN */
	    delFAVE(faVisitorTable, &faVisitorTableSize, repPtr->homeAddr,
		    favePtr->faVisitorCOAddr);
	} else {
	    /*
	     * Delete prior accepted entries for this MN and COA pair
	     * and make the "pending" entry "accepted". The COA is
	     * available in the faVisitorCOAddr field of favePtr.
	     */
	    acceptFAVE(faVisitorTable, &faVisitorTableSize, repPtr->homeAddr,
		       favePtr, ntohs(repPtr->regLifetime));
	}
    } else {
        /*
	 * For denials, simply delete pending entry unless it
	 * corresponds to an HA discovery request; In that case
	 * let the periodic timer delete the request.
	 * TODO: It is better to look at the HA field in the
	 *       request but since we don't have the MN's home
	 *       netmask, we rely on the returned code.
	 */
        if (repPtr->code != HA_UNKNOWN_HOME_AGENT)
	    delFAVEptr(faVisitorTable, &faVisitorTableSize, favePtr);
    }

    printf("\n\n------ FA Registration Request Received & Processed -----\n");
    printf("HA Binding Table Size %d\n", haBindingTableSize);
    printf("HA Mobile Node Table Size %d\n", haMobileNodeTableSize);
    printf("Filter Table Size %d\n", boundMNTableSize);
    printf("FA Visitor Table Size %d\n", faVisitorTableSize);
    printf("FA Unreachable Table Size %d\n", faUnreachableTableSize);
    printf("---------------------------------------------------------\n\n");

    mip2dbg(("\n\n"));
    sendInfoToAllGUIs();
}

main(int argc, char *argv[])
{
    struct timeval tv;
    FILE *fd;

    mip0dbg(("\nMobile IP agent started [%s] ...\n", MIPAGENT_VERSION));
    mip2dbg(("\n---- %s (%s) ----\n", sprintTime(currentTime, 30),
	     sprintRelativeTime(relativeTime, 30)));

    if ((fd = fopen(CONF_FILE_NAME, "r")) == NULL) {
        mip0dbg(("Could not open %s ... exiting.\n", CONF_FILE_NAME));
	exit(-1);
    }

    Initialize(fd);

    while (1) {
        getAndDispatchMessages();

        /* This sleep added for Linux. Couldn't use sleep(1)
	 * as it has an adverse interaction with alarm.
	 * TODO: Linux should define its own fixed sleep.
	 * Sleeping now is good because it relinquishes
	 * control of the CPU instead of staying in a tight loop.
	 */

/* Modified by Uni. of Sheffield */
        tv.tv_sec = 0; tv.tv_usec = (granularity/2);
        select(FD_SETSIZE, NULL, NULL, NULL, &tv);

	if (timerWentOff) {
            totalTime += granularity;
            maSendAdvertisements(maAdvConfigTable,maAdvConfigTableSize, faBusy);
            if ((periodicInterval*1000000) < totalTime) {
               doPeriodicTask(periodicInterval);
               totalTime = 0;
            }
	    timerWentOff = 0;
	}
    }
}


