/*
 * An implementation of Mobile-IP as described in RFC 2002.
 * Author: Vipul Gupta (vgupta@eng.sun.com)
 * agentID.c : Routines to manipulate 64-bits IDs used for replay protection.
 */

#pragma ident	"@(#)agentID.c 1.2 98/08/19"

#include "mip.h"
#include "agent.h"

extern int IDfreshnessSlack;
extern u32 randomLong();
extern u32 CurrentTimeNTPSec();

void HAinitID(u32 *IDHigh, u32 *IDLow, int ReplayStyle) 
{
   if (ReplayStyle == NONCES)  {
        *IDHigh = randomLong();
        *IDLow = 0;
   } else if (ReplayStyle == TIMESTAMPS)  {
        *IDHigh = CurrentTimeNTPSec() - IDfreshnessSlack;
        *IDLow = randomLong();
   } else  {
        *IDHigh = 0;
        *IDLow = 0;
   }
}

boolean isIDgreater(u32 StoredIDHigh, u32 StoredIDLow,
		    u32 IDHigh, u32 IDLow)
{
    if ((IDHigh > StoredIDHigh) || 
	((IDHigh == StoredIDHigh) && (IDLow > StoredIDLow)))
        return(TRUE);
    else
        return(FALSE);
}
 
boolean isIDfresh(u32 IDHigh, u32 IDLow)
{
    long diff;

    diff = (long) (IDHigh - CurrentTimeNTPSec());

    if (diff < 0) 
        diff = (0 - diff);
    return((diff < IDfreshnessSlack) ? TRUE : FALSE);
}

boolean HAisIDok(u32 StoredIDHigh, u32 StoredIDLow,
		 u32 IDHigh, u32 IDLow, int ReplayStyle)
{
   if (ReplayStyle == NONCES)  {
       return((StoredIDHigh == IDHigh) ? TRUE : FALSE);
   } else if (ReplayStyle == TIMESTAMPS)  {
       if (isIDgreater(StoredIDHigh, StoredIDLow, IDHigh, IDLow) &&
	   isIDfresh(IDHigh, IDLow))
      	   return(TRUE);
       else
      	   return(FALSE);
   } else if (ReplayStyle == NONE) {
       return(TRUE);
   } else {
       return(FALSE); 
   }
}

void HAnewID(u32 *newIDHigh, u32 *newIDLow, u32 IDHigh, 
	     u32 IDLow, int ReplayStyle, boolean IDmatched)
{
    if (ReplayStyle == NONCES) {	
	*newIDHigh = randomLong();
    } else if (ReplayStyle == TIMESTAMPS)  {
        *newIDHigh = IDmatched ? IDHigh : CurrentTimeNTPSec();
    } else  {
        *newIDHigh = 0;
    }
    *newIDLow = IDLow;
}

void HAstoreID(u32 *StoredIDHigh, u32 *StoredIDLow, u32 IDHigh, 
	       u32 IDLow, int ReplayStyle, boolean IDmatched)
{
    if ((ReplayStyle == NONCES) || 
        ((ReplayStyle == TIMESTAMPS) && IDmatched)) {
        *StoredIDHigh = IDHigh;
        *StoredIDLow = IDLow;
    }
}

