/*

  Sun Public License Notice

 The contents of this file are subject to the Sun Public License Version
 1.0 (the "License"). You may not use this file except in compliance with
 the License. A copy of the License is available at http://www.sun.com/

 The Original Code is the Sunlabs Mobile IP.
 The Initial Developer of the Original Code is Vipul Gupta.
 Portions created by Vipul Gupta are Copyright (C) 2003.
 All Rights Reserved.

 Contributor(s): Niko A. Fikouras (niko@comnets.uni-bremen.de)
		 Asanga D. Udugama (adu@comnets.uni-bremen.de)

*/

/*
 * An implementation of Mobile-IP as described in RFC 2002.
 * Author: Vipul Gupta (vgupta@eng.sun.com)
 * agentInit.c : Various initialization routines for a mobility agent.
 */

#pragma ident "@(#)agentInit.c 1.9 98/08/19"

#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <errno.h>

/* #include <unistd.h>  Required in LINUX for signal() according to the
 * man pages. But works without it. */

#include "mip.h"
#include "agent.h"

/* Common to all mobility agents */
extern MaAdvConfigEntry maAdvConfigTable[DEFAULT_MAX_INTERFACES];
extern int maAdvConfigTableSize;

/* Foreign Agent specific data structures. */
extern FaVisitorEntry faVisitorTable[DEFAULT_MAX_VISITORS];
extern int faVisitorTableSize;
extern FaUnreachableEntry faUnreachableTable[DEFAULT_MAX_UNREACHABLES];
extern int faUnreachableTableSize;

/* Home Agent specific data structures. */
extern HaMobileNodeEntry haMobileNodeTable[DEFAULT_MAX_MOBILE_NODES];
extern int haMobileNodeTableSize;
extern HaBindingEntry haBindingTable[DEFAULT_MAX_BINDINGS];        
extern int haBindingTableSize;
extern DomainInfo domainInfo;

/* Other external declarations */
extern int  mipdebug;
extern int  IDfreshnessSlack;
extern int  regLifetime;
extern int  advLifetime;
extern int  periodicInterval;
/* modified by Uni. of Sheffield */
extern int  granularity;
extern int  advertiseOnBcast;
extern int  timerWentOff;
/* modified by Uni. of Sheffield */
extern int  totalTime;
extern char msg[], currentTime[], relativeTime[], addr1[], addr2[];
extern int  ICMPsock;
extern int  BcastSock;

/* Modified by Uni. of Sheffield */
struct itimerval tvi;

extern int     hwAddrRead(unsigned char *, char *);
extern char    *ntoa(u32, char *);
extern char    *sprintTime(char *, int);
extern char    *sprintRelativeTime(char *, int);
extern void    HAinitID(u32 *, u32 *, int);
extern int     hexConvert(char *, int, char *);
extern boolean isBlank(char *, int);
extern int     getWord(char *, int, char *);
extern void    printProtectedDomainInfo(DomainInfo);

extern int GUIRecvActive;
extern int GUIRecvPortNum;


/* to store the filters that have been defined for each connected mobile node */
extern BoundMN boundMNTable[DEFAULT_MAX_BINDINGS];			   // one for each MN bound
extern int boundMNTableSize;

/* OS-specific initialization for Mobile IP */
extern void InitNet() ;
extern void OScleanup();

int daemonInit()
{
    pid_t pid;

    if ((pid = fork()) < 0)
        return(-1);
    else if (pid != 0)
        exit(0);         /* parent exits */

    /* child continues */
    setsid();            /* become session leader */
    chdir("/");          /* change working directory */
    umask(0);            /* clear file mode creation mask */
    return(0);
}

int getProtectedDomainInfo(FILE *fd, int *lineNum)
{
    char line[256];
    int state = 0;
    char str1[30], str2[30], word[30];
    int int1, i;
 
    domainInfo.addrIntervalCnt = 0;
    domainInfo.firewallCnt = 0;
    while (fgets(line, 256, fd) != NULL) {
        /* we allow comments starting with # in column 1 */
	(*lineNum)++;
        if ((line[0] == '#') || isBlank(line, strlen(line))) 
	    continue;

        switch (state) {
        case 0 : /* Get number of address intervals */
            if ((sscanf(line, "%d", &int1) != 1) ||
	       (int1 < 1) || (int1 > MAX_ADDR_INTERVALS)) {
	        mip0dbg(("Problem reading # of address intervals on line %d.\n",
	                *lineNum));
	        return(-1);
	    } else {
	        domainInfo.addrIntervalCnt = int1;
                i = 0;
	        state = 1;
            }
	    break;

        case 1 : /* Get the address and netmask */
	    if ((sscanf(line, "%s %s", str1, str2) != 2) || 
		(inet_addr(str1) == -1) ||
		(inet_addr(str2) == -1)) {
		mip0dbg(("Improper arguments on line %d.\n", *lineNum));
		return(-1);
	    } else {
		domainInfo.addr[i] = inet_addr(str1);
	        domainInfo.netmask[i] = inet_addr(str2);
		i++;
		if (i == domainInfo.addrIntervalCnt)
		    state = 2;
	    }
	    break;

	case 2 : /* Get the list of firewalls */
            i = 0;
	    while (((getWord(line, i + 1, word) >= 0) && (i < MAX_FIREWALLS))) {
	        if (inet_addr(word) == -1) {
	            mip0dbg(("Improper firewall address value %s on line %d.\n",
			      word, *lineNum));
		    return(-1);
                } else {
	            domainInfo.fwAddr[i] = inet_addr(word);
		    i++;
		}
	    }
            domainInfo.firewallCnt = i;
            return(0);
	}
    }
    if (state == 1) {
	mip0dbg(("Could not find firewall addresses.\n"));
	return(-1);
    } else 
	return(0);
}


boolean badAdvFlags(char flags)
{
    char implemented;

    implemented = (ADV_IS_HOME_AGENT | ADV_IS_FOREIGN_AGENT);

    /* Only FAs can have the R bit set. */
    if ((flags & ADV_REGISTRATION_REQUIRED) &&
	((flags & ADV_IS_FOREIGN_AGENT) == 0)) {
        return(TRUE);
    }
 
    if (flags & ~implemented)
        return(TRUE);
    else
        return(FALSE);
}

int getAttributeValue(char *str1, int int1, int lineNum)
{
    if (strcasecmp(str1, "debugLevel") == 0) {
        if (int1 < 0 || int1 > 3) {
	    mip0dbg(("Bad debugLevel %d on line %d.\n", int1, lineNum));
	    return(-1);
	} else {
	    mipdebug = int1;
	}
    }

    if (strcasecmp(str1, "IDfreshnessSlack") == 0) {
        if (int1 < 0 || int1 > 3600) {
	    mip0dbg(("Bad IDfreshnessSlack %d on line %d.\n",
		    int1, lineNum));
	    return(-1);
	} else {
	    IDfreshnessSlack = int1;
	}
    }

    if (strcasecmp(str1, "regLifetime") == 0) {
        if (int1 < 0) {
	    mip0dbg(("Bad regLifetime %d on line %d.\n",
		    int1, lineNum));
	    return(-1);
	} else {
	    regLifetime = int1;
	}
    }

    if (strcasecmp(str1, "advLifetime") == 0) {
        if (int1 < 0 || int1 > 600) {
	    mip0dbg(("Bad advLifetime %d on line %d.\n",
		    int1, lineNum));
	    return(-1);
	} else {
	    advLifetime = int1;
	}
    }

    if (strcasecmp(str1, "periodicInterval") == 0) {
        if ((int1 < 0) || (int1 > advLifetime/3)) {
	    mip0dbg(("Bad periodicInterval %d on line %d.\n",
		    int1, lineNum));
	    return(-1);
	} else {
	    periodicInterval = int1;
	}
    }

/* Modified by Uni. of Sheffield to enable agents to operate */
/* in granularities of less than a second.                   */

    if (strcasecmp(str1, "granularity") == 0) {
        if (int1 < 10000) {
	    mip0dbg(("Bad granularity %d on line %d. Granularity must be at least 10ms.\n",
		    int1, lineNum));
	    return(-1);
	} else {
            tvi.it_interval.tv_sec = 0;
            tvi.it_interval.tv_usec = 0;

            tvi.it_value.tv_sec = 0;
            tvi.it_value.tv_usec = int1;
	}
    }

    if (strcasecmp(str1, "advertiseOnBcast") == 0) {
        if (int1 < 0 || (int1 > 1)) {
	    mip0dbg(("Bad advertiseOnBcast flag %d on line %d.\n",
		    int1, lineNum));
	    return(-1);
	} else {
	    advertiseOnBcast = int1;
	}
    }

    if (strcasecmp(str1, "GUIRecvActive") == 0) {
       if ((int1 != 0) && (int1 != 1)) {
          mip0dbg(("Bad GUIRecvActive %d on line %d.\n", int1, lineNum));
          return(-1);
       }
       else {
            GUIRecvActive = int1;
       }
       return(0);
    }

    if (strcasecmp(str1, "GUIRecvPortNum") == 0) {
       if (int1 <= 2048) {
          mip0dbg(("Bad GUIRecvPortNum %d on line %d.\n", int1, lineNum));
          return(-1);
       }
       else {
            GUIRecvPortNum = int1;
       }
       return(0);
    }

    return(0);
}



/* Read configuration information for the mobility agent from file */
int ReadConfigInfo(FILE *fd,
		   MaAdvConfigEntry maceTable[], int *maceTableSize,
		   HaMobileNodeEntry hmneTable[], int *hmneTableSize)
{
    int  i, j;
    int  state = 0;
    int  lineNum = 0;
    char line[256];
    char str1[20], str2[20], str3[20], str4[20], str5[20];
    int  int1, int2;
    long SPI;
    char keyStr[2*MAX_KEY_LEN + 2];
    MaAdvConfigEntry  macEntry;
    HaMobileNodeEntry hmnEntry;

    while ((fgets(line, 256, fd) != NULL) && (state != 6) && (state != 7)) {
        /* we allow comments starting with # in column 1 */
        lineNum++;
        if ((line[0] == '#') || isBlank(line, strlen(line))) continue;

        switch (state) {
	case 0:  /* expect to see a version string */
	    if ((sscanf(line, "%s %d", str1, &int1) != 2) ||
		(strcasecmp(str1, "version") != 0) ||
		(int1 != 1)) {
	        mip0dbg(("Bad version in configuration file on line %d.\n",
			 lineNum));
	        state = 7;
	    } else {
	        state = 1;
	    }

	    break;

	case 1:  /* expect (optional) attribute-value pairs */
	    if (sscanf(line, "%s %d", str1, &int1) == 2) {
	        if (getAttributeValue(str1, int1, lineNum) < 0) 
		    state = 7;
		break;
	    }
		
	    /* Note that if we are unable to read a string-integer
	     * pair, we try to read the next part of the configuration
	     * file by falling through to case 2.
	     */

	case 2:  /* expect to see number of MIP interfaces */
	    if ((sscanf(line, "%d", &int1) != 1) ||
		(int1 < 1) || (int1 > DEFAULT_MAX_INTERFACES)) {
	        mip0dbg(("Problem reading # of MIP interfaces on line %d.\n",
			 lineNum));
	        state = 7;
	    } else {
	        *maceTableSize = int1;
	        state = 3;
		i = 0;
	    }
	    
	    break;

	case 3:  /* expect to see one line for each MIP interface */
	    if ((sscanf(line, "%s %s %s %s %x %d",
			str1, str2, str3, str4, &int1, &int2) != 6) ||
		(strlen(str1) >= MAX_IFNAME_LEN - 1) ||
		(inet_addr(str2) == -1) || (inet_addr(str3) == -1) ||
		(hwAddrRead((unsigned char *)str5, str4) == -1) ||
		badAdvFlags((char) int1)) {
	        mip0dbg(("Problem reading interface info on line %d.\n",
			 lineNum));
	        state = 7; 
	    } else {
	        strncpy(macEntry.maIfaceName, str1, MAX_IFNAME_LEN);
	        macEntry.maIfaceAddr = inet_addr(str2);
		macEntry.maIfaceNetmask = inet_addr(str3);
		memcpy(macEntry.maIfaceHWaddr, str5, MAX_HWADDR_LEN);
		macEntry.maAdvMaxRegLifetime = regLifetime;
		/* 
		 * TODO: Under Solaris, pkts for LINK_BCAST_ADDR seem to
		 * be sent on all of a host's interfaces. Don't know if this
		 * can be controlled using some options similar to 
		 * IP_MULTICAST_IF.
		 */
		macEntry.maAdvAddr = (advertiseOnBcast == 1) ? 
		  inet_addr(LINK_BCAST_ADDR) : inet_addr(LINK_MCAST_ADDR);
		macEntry.maAdvMaxInterval = DEFAULT_MAX_INTERVAL;
		macEntry.maAdvMinInterval = DEFAULT_MIN_INTERVAL;
		macEntry.maAdvMaxAdvLifetime = advLifetime;
		macEntry.maAdvSeqNum = 0;
		macEntry.maAdvServiceFlags = (char) int1;
		macEntry.maAdvPrefixLenInclusion = (boolean) int2;
		bcopy(&macEntry, &(maceTable[i]), sizeof(macEntry));
		/* printMaAdvConfigEntry(&(maceTable[i])); */
		i++;
		if (i == *maceTableSize) state = 4;
	    }
		
	    break;

	case 4:  /* expect to see number of supported mobile nodes */
	    if ((sscanf(line, "%d", &int1) != 1) ||
		(int1 < 0) || (int1 > DEFAULT_MAX_MOBILE_NODES)) {
	        mip0dbg(("Problem reading # of mobile nodes on line %d.\n",
			 lineNum));
	        state = 7;
	    } else {
	        *hmneTableSize = int1;
		i = 0;
		if (i == *hmneTableSize) 
		    state = 6;
		else
		    state = 5;
	    }

	    break;


	case 5:  /* expect one line for each mobile node  */
	    if ((sscanf(line, "%s %s %ld %d %d %s",
			str1, str2, &SPI, &int1, &int2, keyStr) != 6) ||
		(inet_addr(str1) == -1) || (inet_addr(str2) == -1) ||
		(int1 < 0) || (int1 > 2) || (int2 > MAX_KEY_LEN)) {
	        mip0dbg(("Problem reading mobile node info on line %d.\n",
			 lineNum));
	        state = 7;
	    } else {
	        hmnEntry.haMnAddr = inet_addr(str1);
		hmnEntry.haBindingIfaceAddr = inet_addr(str2);
		hmnEntry.haMnBindingCnt = 0;
		hmnEntry.haMnMSA.mipSecAddr = hmnEntry.haMnAddr;
		hmnEntry.haMnMSA.mipSecSPI = SPI;
		hmnEntry.haMnMSA.mipSecReplayMethod = int1;
		HAinitID(&hmnEntry.haMnRegIDHigh, &hmnEntry.haMnRegIDLow, 
			 hmnEntry.haMnMSA.mipSecReplayMethod);
		hmnEntry.haMnMSA.mipSecAlgorithmType = MD5;
		hmnEntry.haMnMSA.mipSecAlgorithmMode = PREFIXSUFFIX;
		hmnEntry.haMnMSA.mipSecKeyLen = int2;
		if (hexConvert((char *)hmnEntry.haMnMSA.mipSecKey,
			      int2, keyStr) < 0) {
		    mip0dbg(("Problem reading key on line %d.\n",
			     lineNum));
		    state = 7;
		}
    
		bcopy(&hmnEntry, &hmneTable[i], sizeof(hmnEntry));
		/* printHaMobileNodeEntry(&hmnEntry); */
		i++;
		if (i == *hmneTableSize) {
		    if (getProtectedDomainInfo(fd, &lineNum) < 0) {
		        state = 7;
		    } else {
			state = 6;
		    }
		}
	    }
		
	    break;
	}
    }

    if (state == 6) {
        mip0dbg(("Configuration file read successfully.\n"));
	return(0);
    } else if (state == 7) {
        mip0dbg(("Configuration file had problems.\n"));
	return(-1);
    } else {
        mip0dbg(("Configuration file seems incomplete.\n"));
	return(-1);
    }
}

void InitHaBindingTable(HaBindingEntry habTable[], int *habTableSize)
{
    int i;

    for (i = 0; i < DEFAULT_MAX_BINDINGS; i++)
        memset(&habTable[i], 0, sizeof(HaBindingEntry));
    *habTableSize = 0;
}

void InitFaVisitorTable(FaVisitorEntry faveTable[], int *faveTableSize)
{
    int i;

    for (i = 0; i < DEFAULT_MAX_VISITORS; i++)
        memset(&faveTable[i], 0, sizeof(FaVisitorEntry));
    *faveTableSize = 0;
}

void InitFaUnreachableTable(FaUnreachableEntry faueTable[], int *faueTableSize)
{
    int i;

    for (i = 0; i < DEFAULT_MAX_UNREACHABLES; i++)
        memset(&faueTable[i], 0, sizeof(FaUnreachableEntry));
    *faueTableSize = 0;
}

/* initialize the table that holds all the filter data related to bindings*/
void InitFilterTable(BoundMN bmn[], int *fltCnt) {
	int i, j, k;

	for(i = 0; i < DEFAULT_MAX_BINDINGS; i++) {
		bmn[i].mnAddr = 0;
		bmn[i].mnFilterCount = 0;
		bmn[i].filtersChanged = FALSE;

		for(j = 0; j < MAX_FILTERS; j++) {
			bmn[i].mnFilters[j].filterModuleCount = 0;
			bmn[i].mnFilters[j].seq = (-1);
			bmn[i].mnFilters[j].markValue = (-1);

			for(k = 0; k < MAX_FILTER_MODULES_PER_FILTER; k++) {
				bmn[i].mnFilters[j].controlModules[k].boundCoA = 0;
				bmn[i].mnFilters[j].controlModules[k].weight = (-1);
			}
		}
	}
	*fltCnt = 0;
}

#ifndef IPPROTO_ICMP
#define IPPROTO_ICMP  1
#endif

/* Create all the sockets used by a mobility agent */
int InitSockets(int *icmpSockId, int *bcastSockId,
		MaAdvConfigEntry maceTable[], int maceTableSize)
{
    int i;
    int sid;
    int enable = 1;
    struct sockaddr_in sa;
    struct ip_mreq imr;

    mip2dbg(("Creating ICMP socket.\n"));
    if ((*icmpSockId = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP)) < 0) {
        perror("0:socket()"); 
        mip0dbg(("Error: Could not create ICMP socket in InitSockets().\n"));
	return (-1);
    }

    /* Just in case we advertise on 255.255.255.255, enable bcast */
    enable = 1;
    if (setsockopt(*icmpSockId, SOL_SOCKET, SO_BROADCAST,
		   (char *)&enable, sizeof(int)) < 0) {
        perror("SO_BROADCAST on ICMPsock failed");
	return(-1);
    }

    /* 
     * We also bind a UDP socket to 255.255.255.255, port 434
     * for dynamic HA discovery. We do so because some
     * routers (incorrectly) alter the destination from subnet bcast
     * addr to 255.255.255.255.
     */
    if ((*bcastSockId = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
        perror("1:socket()");
	mip0dbg(("Error: Could not create bcast UDP socket.\n"));
	return (-1);
	}

    enable = 1;
    if (setsockopt(*bcastSockId, SOL_SOCKET, SO_REUSEADDR,
		   (char *)&enable, sizeof(int)) < 0) {
        perror("1:setsockopt1()");
    }

    sa.sin_family = AF_INET;
    sa.sin_port = htons(MIP_PORT);
    sa.sin_addr.s_addr = inet_addr(LINK_BCAST_ADDR);

    mip2dbg(("Binding UDP socket to %s port %d.\n",
	     ntoa(sa.sin_addr.s_addr, addr1), ntohs(sa.sin_port)));
    if (bind(*bcastSockId, (struct sockaddr *) &sa, sizeof(sa)) < 0) {
        perror("1:bind()");
	mip0dbg(("Error: Could not bind bcast socket.\n"));
	return(-1);
    }

    /* 
     * TODO: We also need a socket to monitor 224.0.0.11 on port 434.
     */
    for (i = 0; i < maceTableSize; i++) {

	/* 
	 * Join multicast groups so we can receive ICMP messages
	 * on this interface sent to 224.0.0.1 and 224.0.0.2.
	 */

	imr.imr_multiaddr.s_addr = inet_addr("224.0.0.1");
	imr.imr_interface.s_addr = maceTable[i].maIfaceAddr;
	if (setsockopt(*icmpSockId, IPPROTO_IP, IP_ADD_MEMBERSHIP,
		       (char *) &imr, sizeof(struct ip_mreq)) < 0) {
	    perror("Could not join multicast group 224.0.0.1\n");
	} else {
	    mip2dbg(("Joined 224.0.0.1 on interface %s.\n",
		     ntoa(maceTable[i].maIfaceAddr, addr1)));
	}

	imr.imr_multiaddr.s_addr = inet_addr("224.0.0.2");
	imr.imr_interface.s_addr = maceTable[i].maIfaceAddr;
	if (setsockopt(*icmpSockId, IPPROTO_IP, IP_ADD_MEMBERSHIP, 
		       (char *) &imr, sizeof(struct ip_mreq)) < 0) {
	    perror("Could not join multicast group 224.0.0.2\n");
	} else {
	    mip2dbg(("Joined 224.0.0.2 on interface %s.\n",
		     ntoa(maceTable[i].maIfaceAddr, addr1)));
	}

        /* Create and bind the socket to monitor the unicast interface addr */

        if ((sid = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
	    perror("2:socket()");
	    mip0dbg(("Error: Could not create unicast UDP socket.\n"));
	    return (-1);
	}

	if (setsockopt(sid, SOL_SOCKET, SO_REUSEADDR, 
		       (char *)&enable, sizeof(int)) < 0) {
	    perror("2:setsockopt1()");
	}

	sa.sin_family = AF_INET;
	sa.sin_port = htons(MIP_PORT);
	sa.sin_addr.s_addr = maceTable[i].maIfaceAddr;

	mip2dbg(("Binding UDP socket to %s port %d.\n",
		 ntoa(sa.sin_addr.s_addr, addr1), ntohs(sa.sin_port)));
	if (bind(sid, (struct sockaddr *) &sa, sizeof(sa)) < 0) {
	    perror("2:bind()");
	    mip0dbg(("Error: Could not bind unicast socket.\n"));
	    return(-1);
	}

	maceTable[i].maIfaceUnicastSock = sid;
	
        /* Create and bind the socket to monitor the bcast interface addr */
        if ((sid = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
	    perror("3:socket()");
	    mip0dbg(("Error: Could not create broadcast UDP socket.\n"));
	    return (-1);
	}	

	if (setsockopt(sid, SOL_SOCKET, SO_REUSEADDR, 
		       (char *)&enable, sizeof(int)) < 0) {
	    perror("3:setsockopt()");
	}

	sa.sin_family = AF_INET;
	sa.sin_port = htons(MIP_PORT);
	sa.sin_addr.s_addr = (maceTable[i].maIfaceAddr &
			      maceTable[i].maIfaceNetmask) |
	                      ~maceTable[i].maIfaceNetmask;
	mip2dbg(("Binding UDP socket to %s port %d.\n",
		 ntoa(sa.sin_addr.s_addr, addr1), ntohs(sa.sin_port)));
	if (bind(sid, (struct sockaddr *) &sa, sizeof(sa)) < 0) {
	    perror("3:bind");
	    mip0dbg(("Error: Could not bind broadcast socket.\n"));
	    return(-1); 
	}
	maceTable[i].maIfaceBcastSock = sid;
    }

    return 0;
}

/* Schedule doPeriodicTask once every periodicInterval seconds. */
/* ARGSUSED */
void DoAlarm(int junk) 
{
    timerWentOff = 1;
    if (signal(SIGALRM, DoAlarm) == SIG_ERR) {
        perror("0:signal()");
        mip0dbg(("Error: Failed to set alarm in DoAlarm().\n"));
    }
    /*alarm(periodicInterval);*/
/* Modified by Uni. of Sheffield to enable agents to operate */
/* in granularities of less than a second.                   */
     setitimer(ITIMER_REAL,&tvi,0);     
}

void Finalize(int junk)
{
    int i;

    mip2dbg(("\n---- %s (%s) ----\n", sprintTime(currentTime, 30),
	     sprintRelativeTime(relativeTime, 30)));
    mip2dbg(("<<< Cleaning up mobility bindings >>>\n"));
    for (i = 0; i < DEFAULT_MAX_BINDINGS; i++) {
        if (haBindingTable[i].haBindingTimeRemaining) {
	    delHABEidx(haBindingTable, &haBindingTableSize, i);
	}
    }

    mip2dbg(("<<< Cleaning up accepted visitor entries >>>\n"));
    for (i = 0; i < DEFAULT_MAX_VISITORS; i++) {
        if ((faVisitorTable[i].faVisitorRegIsAccepted == TRUE) &&
	    (faVisitorTable[i].faVisitorTimeRemaining)) {
	    delFAVEptr(faVisitorTable, &faVisitorTableSize,
		       &(faVisitorTable[i]));
	}
    }

    OScleanup();
    termInfoModule();

    mip2dbg(("Terminated by signal (SIGTERM or SIGINT)\n"));
    exit(-1);
}

/* Set up handlers for various signals */
void SetSigHandlers()
{
    if (signal(SIGINT, Finalize) == SIG_ERR) {
        perror("1:signal()");
	mip0dbg(("Error: failed to set handler for SIGINT.\n"));
    }

    if (signal(SIGTERM, Finalize) == SIG_ERR) {
        perror("2:signal()");
	mip0dbg(("Error: failed to set handler for SIGTERM.\n"));
    }
}

void showConfigInfo()
{
    int i;

    mip3dbg(("DebugLevel                   : %d\n", mipdebug));
    mip3dbg(("PeriodicInterval             : %d\n", periodicInterval));
    mip3dbg(("Granularity                  : %d\n", granularity));
    mip3dbg(("AdvLifetime                  : %d\n", advLifetime));
    mip3dbg(("RegLifetime                  : %d\n", regLifetime));
    mip3dbg(("IDfreshnessSlack             : %d\n", IDfreshnessSlack));
    mip3dbg(("\n"));
    printMaAdvConfigTable(maAdvConfigTable, maAdvConfigTableSize);
    mip3dbg(("\n"));
    printHaMobileNodeTable(haMobileNodeTable, haMobileNodeTableSize);
    mip3dbg(("\n"));
    printProtectedDomainInfo(domainInfo);
    mip3dbg(("\n"));
}

void Initialize(FILE *fd)
{
    /* Initialize itself as a daemon */
    /*
     * daemonInit();
     */

    /* Initialize random number generator */
    randomInit();

    /* Initialize maAdvConfigTable and haMobileNodeTable from file */
    if (ReadConfigInfo(fd, maAdvConfigTable, &maAdvConfigTableSize,
		       haMobileNodeTable, &haMobileNodeTableSize) < 0) {
        mip0dbg(("Error: Start up configuration unsuccessful.\n"));
        exit(-1);
    }

    showConfigInfo();

    /* Initialize other tables as empty */
    InitHaBindingTable(haBindingTable, &haBindingTableSize);
    InitFaVisitorTable(faVisitorTable, &faVisitorTableSize);
    InitFaUnreachableTable(faUnreachableTable, &faUnreachableTableSize);
    InitFilterTable(boundMNTable, &boundMNTableSize);

    /* OS-specific initialization of tunneling module etc */
    InitNet();

    /* OS-neutral, socket initialization */
    if (InitSockets(&ICMPsock, &BcastSock, maAdvConfigTable,
		    maAdvConfigTableSize) < 0)  {
        mip0dbg(("InitSockets failed.\n"));
        exit(-1);
    }

    initInfoModule();
    
    /* Set up periodic house keeping and other chores */
    DoAlarm(0);
    SetSigHandlers();
}
