/*

  Sun Public License Notice

 The contents of this file are subject to the Sun Public License Version
 1.0 (the "License"). You may not use this file except in compliance with
 the License. A copy of the License is available at http://www.sun.com/

 The Initial Developer of the Original Code is Asanga D. Udugama
 Portions created by Asanga D. Udugama are Copyright (C) 2003.
 All Rights Reserved.

 Contributor(s): none

*/

/*
 * An implementation of Filers for Mobile-IP as described in
 * draft-nomad-mobileip-filters-05.txt.
 * Author: Asanga D. Udugama (adu@comnets.uni-bremen.de)
 * filters.c : Filter management routines used in Sunlabs Mobile IP.
 */

#pragma ident	"@(#)filters.c 1.2 2002/10/23"

#include <stdio.h>
#include "mip.h"
#include "agent.h"

#define MAX_MARKS_SUPPORTED	256

extern int  mipdebug;

extern void cvtCharArrayToShortInt(short int *, unsigned char *);
extern void cvtCharArrayTou32(u32 *, unsigned char *);
extern int deleteAllFilters(u32);
extern int applyFilter(u32, u32, char *);
extern int applyDefaultFilter(u32, u32, int);
extern int flushRouteCache();

int extractFilters(int extCnt, unsigned char extType[], unsigned char *extIdx[],
					BoundMN bmn[], int *bts, u32 mnAddr, u32 coAddr);
void dispFilters(BoundMN bmn[], u32 mnAddr);
int haveFiltersChanged(BoundMN bmn[], u32 mnAddr);
void FAapplyFilters(BoundMN bmn[], u32 mnAddr);
int HAapplyFilters(BoundMN bmn[], HaBindingEntry habe[], u32 mnAddr);
u32 getDefaultFilterCoA(BoundMN bmn[], HaBindingEntry habe[], u32 mnAddr);
int arrangeFiltersSequentially(BoundMN [], u32);
int sortFilters(BoundMN [], u32);
int removeSelectedFilters(BoundMN bmn[], int *bmnCnt, u32 mnAddr, u32 coAddr);
int addFilterTableEntry(BoundMN bmn[], u32 mnAddr);
int getMarkValue(u32 mnAddr, int fltSeq);
void freeMarkValue(u32 mnAddr, int fltSeq);
void setFilterOp(int opCode, u32 mnAddr, u32 coAddr, int seq, int markValue, int pos) ;

// variables used in holding mnaddr & coaddr for
// marks used in iptables
BindingMark bmark[MAX_MARKS_SUPPORTED];
int bmarkInit = TRUE;

/* to store the operations for filters that were changed */
MNFltOpEntry mnFltOpTable[DEFAULT_MAX_BINDINGS];
int fltOpInit = TRUE;

/* The function extracts the filters from the registration request received
 * and does the operation required on the filter table
 *
 */
int extractFilters(int extCnt, unsigned char extType[], unsigned char *extIdx[],
			BoundMN bmn[], int *bts, u32 mnAddr, u32 coAddr) {

	int i, j, k, usable, free, existing, fltCnt;
	char *fltExt[MAX_FILTERS];

	Filter cfltr;

	controlExt *controlExtPtr;
	delExt *delExtPtr;
	bafExt *bafExtPtr;
	protoExt *protoExtPtr;
	srcAddrExt *srcAddrExtPtr;
	srcNetExt *srcNetExtPtr;
	srcPortExt *srcPortExtPtr;
	srcPortRangeExt *srcPortRangeExtPtr;
	freeFormExt *freeFormExtPtr;

	dstAddrExt *dstAddrExtPtr;
	dstNetExt *dstNetExtPtr;
	dstPortExt *dstPortExtPtr;
	dstPortRangeExt *dstPortRangeExtPtr;

	BAFExtData  *bafedp;
	ProtoExtData *protoedp;
	SrcAddrExtData *srcaddredp;
	SrcNetExtData *srcnetedp;
	SrcPortExtData *srcportedp;
	SrcPortRangeExtData *srcportrangeedp;
	FreeFormExtData *freeformedp;

	DstAddrExtData *dstaddredp;
	DstNetExtData *dstnetedp;
	DstPortExtData *dstportedp;
	DstPortRangeExtData *dstportrangeedp;

	int len, action, found;
	char str1[30];
	unsigned int int1;
	int l, m, n;

	// find the filter table index to use for the
	// specified MN Addr
	usable = (-1); free = (-1); existing = (-1);
	for(i = 0; i < DEFAULT_MAX_BINDINGS; i++) {
		if(bmn[i].mnAddr == 0 && free == (-1))
			free = i;
		else if(bmn[i].mnAddr == mnAddr) {
			existing = i;
			break;
		}
	}
	if(free == (-1) && existing == (-1))
		return FALSE;
	if(existing != (-1))
		usable = existing;
	else
		usable = free;


	// extract and place the Filter Ext in a separate
	// table

	j = (-1);
	for(i = 0; i < extCnt; i++) {
		if(extType[i] >= FLT_DEL_EXT && extType[i] <=  FLT_DSTPORTRNG_EXT) {
			j++;
			fltExt[j] = extIdx[i];
		}
	}
	fltCnt = ++j;

 	// no filters with the packet, so return (graceful rtn value)
	if(fltCnt <= 0)
		return TRUE;

	// if totally new MN, then increment boundMN table size and set
	// other variables
	if(existing == (-1)) {
		(*bts)++;
		bmn[usable].mnAddr = mnAddr;
		bmn[usable].mnFilterCount = 0;
	}

	// set filters changed status to activate the
	// filter application method
	bmn[usable].filtersChanged = TRUE;

	i = 0;
	while(i < fltCnt) {

		// initialize the filter variable
		cfltr.filterModuleCount = 0;

		cfltr.seq = (-1);
		cfltr.markValue = (-1);

		for(j = 0; j < MAX_FILTER_MODULES_PER_FILTER; j++) {
			cfltr.filterModules[j].filterType = 0;
			cfltr.filterModules[j].filterPtr = (char *) NULL;

			cfltr.controlModules[j].boundCoA = 0;
			cfltr.controlModules[j].weight = (-1);
		}

		j = 0;

		printf("FILTER COUNT IS %d \n", fltCnt);
		// get the individual Filter modules related to one filter
		for(; (i < fltCnt) && !(*fltExt[i] == FLT_CTL_EXT || *fltExt[i] == FLT_DEL_EXT); i++) {
			cfltr.filterModules[j].filterType = *fltExt[i];

			// extract the filter
			switch(*fltExt[i]) {
				case FLT_BAF_EXT:
					bafExtPtr = (bafExt *) fltExt[i];
					if((cfltr.filterModules[j].filterPtr =
								(char *) malloc(sizeof(BAFExtData))) == NULL)
						return FALSE;
					bafedp = (BAFExtData *) cfltr.filterModules[j].filterPtr;
					bafedp->DSCP = (int) (bafExtPtr->DSCP >> 2);
					break;

				case FLT_PROTO_EXT:
					protoExtPtr = (protoExt *) fltExt[i];
					if((cfltr.filterModules[j].filterPtr =
								(char *) malloc(sizeof(ProtoExtData))) == NULL)
						return FALSE;
					protoedp = (ProtoExtData *) cfltr.filterModules[j].filterPtr;
					protoedp->protoNum = protoExtPtr->protoNum;
					break;

				case FLT_SRCADR_EXT:
					srcAddrExtPtr = (srcAddrExt *) fltExt[i];
					if((cfltr.filterModules[j].filterPtr =
								(char *) malloc(sizeof(SrcAddrExtData))) == NULL)
						return FALSE;
					srcaddredp = (SrcAddrExtData *) cfltr.filterModules[j].filterPtr;
					cvtCharArrayTou32(&(srcaddredp->srcAddr), srcAddrExtPtr->srcAddr);
					break;

				case FLT_SRCNET_EXT:
					srcNetExtPtr = (srcNetExt *) fltExt[i];
					if((cfltr.filterModules[j].filterPtr =
								(char *) malloc(sizeof(SrcNetExtData))) == NULL)
						return FALSE;
					srcnetedp = (SrcNetExtData *) cfltr.filterModules[j].filterPtr;
					cvtCharArrayTou32(&(srcnetedp->srcAddr), srcNetExtPtr->srcNet.addr);
					cvtCharArrayTou32(&(srcnetedp->srcNetMask), srcNetExtPtr->srcNet.netMask);
					break;

				case FLT_SRCPORT_EXT:
					srcPortExtPtr = (srcPortExt *) fltExt[i];
					if((cfltr.filterModules[j].filterPtr =
								(char *) malloc(sizeof(SrcPortExtData))) == NULL)
						return FALSE;
					srcportedp = (SrcPortExtData *) cfltr.filterModules[j].filterPtr;
					cvtCharArrayToShortInt(&(srcportedp->portNum), srcPortExtPtr->portNum);
					break;

				case FLT_SRCPORTRNG_EXT:
					srcPortRangeExtPtr = (srcPortRangeExt *) fltExt[i];
					if((cfltr.filterModules[j].filterPtr =
								(char *) malloc(sizeof(SrcPortRangeExtData))) == NULL)
						return FALSE;
					srcportrangeedp = (SrcPortRangeExtData *) cfltr.filterModules[j].filterPtr;
					cvtCharArrayToShortInt(&(srcportrangeedp->minPortNum),
									srcPortRangeExtPtr->portRange.minPortNum);
					cvtCharArrayToShortInt(&(srcportrangeedp->maxPortNum),
									srcPortRangeExtPtr->portRange.maxPortNum);
					break;

				case FLT_FRFORM_EXT:
					freeFormExtPtr = (freeFormExt *) fltExt[i];
					if((cfltr.filterModules[j].filterPtr = (char *) malloc(sizeof(FreeFormExtData))) == NULL)
						return FALSE;
					freeformedp = (FreeFormExtData *) cfltr.filterModules[j].filterPtr;
					cvtCharArrayToShortInt(&(freeformedp->offset), freeFormExtPtr->offset);
					len = ((freeFormExtPtr->length - 2) / 2);
					*(freeformedp->value) == (char) NULL;
					*(freeformedp->mask) == (char) NULL;

					for(k = 0; k < len; k++) {
						int1 = (int) freeFormExtPtr->valueNmask[k];
						sprintf(str1, "%02x", int1);
						strcat(freeformedp->value, str1);
						int1 = (int) freeFormExtPtr->valueNmask[(k + len)];
						sprintf(str1, "%02x", int1);
						strcat(freeformedp->mask, str1);
					}
					break;

				// new filters

				case FLT_DSTADR_EXT:
					dstAddrExtPtr = (dstAddrExt *) fltExt[i];
					if((cfltr.filterModules[j].filterPtr =
								(char *) malloc(sizeof(DstAddrExtData))) == NULL)
						return FALSE;
					dstaddredp = (DstAddrExtData *) cfltr.filterModules[j].filterPtr;
					cvtCharArrayTou32(&(dstaddredp->dstAddr), dstAddrExtPtr->dstAddr);
					break;

				case FLT_DSTNET_EXT:
					dstNetExtPtr = (dstNetExt *) fltExt[i];
					if((cfltr.filterModules[j].filterPtr =
								(char *) malloc(sizeof(DstNetExtData))) == NULL)
						return FALSE;
					dstnetedp = (DstNetExtData *) cfltr.filterModules[j].filterPtr;
					cvtCharArrayTou32(&(dstnetedp->dstAddr), dstNetExtPtr->dstNet.addr);
					cvtCharArrayTou32(&(dstnetedp->dstNetMask), dstNetExtPtr->dstNet.netMask);
					break;

				case FLT_DSTPORT_EXT:
					dstPortExtPtr = (dstPortExt *) fltExt[i];
					if((cfltr.filterModules[j].filterPtr =
								(char *) malloc(sizeof(DstPortExtData))) == NULL)
						return FALSE;
					dstportedp = (DstPortExtData *) cfltr.filterModules[j].filterPtr;
					cvtCharArrayToShortInt(&(dstportedp->portNum), dstPortExtPtr->portNum);
					break;

				case FLT_DSTPORTRNG_EXT:
					dstPortRangeExtPtr = (dstPortRangeExt *) fltExt[i];
					if((cfltr.filterModules[j].filterPtr =
								(char *) malloc(sizeof(DstPortRangeExtData))) == NULL)
						return FALSE;
					dstportrangeedp = (DstPortRangeExtData *) cfltr.filterModules[j].filterPtr;
					cvtCharArrayToShortInt(&(dstportrangeedp->minPortNum),
									dstPortRangeExtPtr->portRange.minPortNum);
					cvtCharArrayToShortInt(&(dstportrangeedp->maxPortNum),
									dstPortRangeExtPtr->portRange.maxPortNum);
					break;


			}
			j++;
		}

		// the next extension definitlely should be a Control Ext or Delete Ext
		if(i >= fltCnt || !((*fltExt[i] == FLT_CTL_EXT) || (*fltExt[i] == FLT_DEL_EXT))) {
			printf("PROBLEM at filter ext code");
			return FALSE;
		}

		// get delete ext and set filter op
		if(*fltExt[i] == FLT_DEL_EXT) {
			delExtPtr = (delExt *) fltExt[i];
			for(l = 0; l < delExtPtr->length; l++) {
				for(m = 0; m < MAX_FILTERS; m++) {
					if(bmn[usable].mnFilters[m].seq == delExtPtr->idx[l]) {
						setFilterOp(FLTOP_DELETE, mnAddr, coAddr,
								bmn[usable].mnFilters[m].seq,
								bmn[usable].mnFilters[m].markValue, (-1));

						// initialize entry
						bmn[usable].mnFilters[m].filterModuleCount = 0;
						bmn[usable].mnFilters[m].seq = (-1);
						bmn[usable].mnFilters[m].markValue = (-1);

						for(n = 0; n < MAX_FILTER_MODULES_PER_FILTER; n++) {
							bmn[usable].mnFilters[m].controlModules[n].boundCoA = 0;
							bmn[usable].mnFilters[m].controlModules[n].weight = (-1);
						}

						bmn[usable].mnFilterCount--;
						
						break;
					}
				}
			}
			goto maintain;
		}

		// get the Control Ext related to the above filter modules
		controlExtPtr = (controlExt *) fltExt[i];

		cfltr.filterModuleCount = j;

		cfltr.controlModules[0].boundCoA = coAddr;
		cfltr.controlModules[0].weight = controlExtPtr->flags; // flags indicate weight
		//cfltr.target = (controlExtPtr->flags & 0x0f); // flags = 4 bits ACT, 4 bits TG
		//action = ((controlExtPtr->flags >> 4) & 0x0f);

		cfltr.seq = controlExtPtr->idx;
		cfltr.markValue = getMarkValue(mnAddr, cfltr.seq); // get packet marking value


		// filter built, now add to the, filter table & op table

		if(cfltr.filterModuleCount <= 0) { // no filter modules in the reg. request, only control ext

			// means addition of a new weight distribution
			// to an existing filter
			for(j = 0; j < MAX_FILTERS; j++) {
				if(bmn[usable].mnFilters[j].seq != cfltr.seq)
					continue;

				free = (-1);
				for(k = 0; k < MAX_FILTER_MODULES_PER_FILTER; k++) {
					if(bmn[usable].mnFilters[j].controlModules[k].boundCoA == coAddr)
						break;
					if(free == (-1) && bmn[usable].mnFilters[j].controlModules[k].weight == (-1))
						free = k; // finds a free slot, for later use
				}

				if(k >= MAX_FILTER_MODULES_PER_FILTER) { // means CoA was not present
					if(free != (-1)) {
						bmn[usable].mnFilters[j].controlModules[free].weight =
									cfltr.controlModules[0].weight;
						bmn[usable].mnFilters[j].controlModules[free].boundCoA =
									cfltr.controlModules[0].boundCoA;
						setFilterOp(FLTOP_SHARE, mnAddr, coAddr, cfltr.seq, cfltr.markValue, (-1));
					} // else, something wrong

				} else { // means CoA present, therefore overwrite
					bmn[usable].mnFilters[j].controlModules[k].weight =
							cfltr.controlModules[k].weight;
					setFilterOp(FLTOP_SHARE, mnAddr, coAddr, cfltr.seq, cfltr.markValue, (-1));
				}
			}
		} else { // filter modules present in the reg. request

			// means new filter to add or replace an existing
			found = FALSE;
			for(j = 0; j < MAX_FILTERS; j++) {
				if(bmn[usable].mnFilters[j].seq == cfltr.seq) {
					found = TRUE;
					break;
				}
			}

			if(!found) { // filter is not present, so append as new filter
				bmn[usable].mnFilterCount++;
				memcpy(&(bmn[usable].mnFilters[MAX_FILTERS - 1]), &cfltr, sizeof(Filter));
				setFilterOp(FLTOP_ADD, mnAddr, coAddr, cfltr.seq, cfltr.markValue, (-1));

			} else { // filter present
				// then, first delete existing
				setFilterOp(FLTOP_DELETE, mnAddr, bmn[usable].mnFilters[j].controlModules[0].boundCoA,
						cfltr.seq, bmn[usable].mnFilters[j].markValue, (-1));

				// and then add the filter
				memcpy(&(bmn[usable].mnFilters[j]), &cfltr, sizeof(Filter));
				setFilterOp(FLTOP_ADD, mnAddr, coAddr, cfltr.seq, cfltr.markValue, (-1));
			}
		}

		maintain:

		// rearrange the filters
		arrangeFiltersSequentially(bmn, mnAddr);

		// sort the filters
		sortFilters(bmn, mnAddr);


		i++;
	}

	return TRUE;
}


/* This function displays the filters associated with the
 * given MN
 */
void dispFilters(BoundMN bmn[], u32 mnAddr) {
	int i, j, elem;
	char addr1[25], addr2[25], fltType[256];
	BAFExtData  *bafedp;
	ProtoExtData *protoedp;
	SrcAddrExtData *srcaddredp;
	SrcNetExtData *srcnetedp;
	SrcPortExtData *srcportedp;
	SrcPortRangeExtData *srcportrangeedp;
	FreeFormExtData *freeformedp;

	DstAddrExtData *dstaddredp;
	DstNetExtData *dstnetedp;
	DstPortExtData *dstportedp;
	DstPortRangeExtData *dstportrangeedp;

	// find the filters associated with the MN
	elem = (-1);
	for(i = 0; i < DEFAULT_MAX_BINDINGS; i++) {
		if(bmn[i].mnAddr == mnAddr) {
			elem = i;
			break;
		}
	}
	if(elem == (-1))
		return;

	// display header info
	mip3dbg(("\n\n----------Filter Data------------ \n"));
	mip3dbg(("MN Address               : %s\n", ntoa(mnAddr, addr1)));
	mip3dbg(("Number of Filters        : %d\n", bmn[elem].mnFilterCount));

	// display data of each filter
	for(i = 0; i < bmn[elem].mnFilterCount; i++) {
		mip3dbg(("  Filter Module Count    : %d\n", bmn[elem].mnFilters[i].filterModuleCount));
		//mip3dbg(("  Bound CoA              : %s\n", ntoa(bmn[elem].mnFilters[i].boundCoA, addr1)));
		//mip3dbg(("  Target                 : %d\n", bmn[elem].mnFilters[i].target));
		mip3dbg(("  Seq Number             : %d\n\n", bmn[elem].mnFilters[i].seq));

		for(j = 0; j < bmn[elem].mnFilters[i].filterModuleCount; j++) {
			switch(bmn[elem].mnFilters[i].filterModules[j].filterType) {
				case FLT_BAF_EXT:
					strcpy(fltType, "Behavior Aggregate Filter Extension");
					mip3dbg(("    Filter Type          : %s\n", fltType));

					bafedp = (BAFExtData *) bmn[elem].mnFilters[i].filterModules[j].filterPtr;

					mip3dbg(("    DSCP Value           : %d\n\n", bafedp->DSCP));

					break;

				case FLT_PROTO_EXT:
					strcpy(fltType, "Protocol Extension");
					mip3dbg(("    Filter Type          : %s\n", fltType));

					protoedp = (ProtoExtData *) bmn[elem].mnFilters[i].filterModules[j].filterPtr;
					mip3dbg(("    Protocol Number      : %d\n\n", protoedp->protoNum));

					break;

				case FLT_SRCADR_EXT:
					strcpy(fltType, "Source Address Extension");
					mip3dbg(("    Filter Type          : %s\n", fltType));

					srcaddredp = (SrcAddrExtData *) bmn[elem].mnFilters[i].filterModules[j].filterPtr;
					mip3dbg(("    Source Address       : %s\n\n", ntoa(srcaddredp->srcAddr, addr1)));

					break;

				case FLT_SRCNET_EXT:
					strcpy(fltType, "Source Network Extension");
					mip3dbg(("    Filter Type          : %s\n", fltType));

					srcnetedp = (SrcNetExtData *) bmn[elem].mnFilters[i].filterModules[j].filterPtr;
					mip3dbg(("    Source Address & Mask: %s/%s\n\n",
										ntoa(srcnetedp->srcAddr, addr1),
										ntoa(srcnetedp->srcNetMask, addr2)));

					break;

				case FLT_SRCPORT_EXT:
					strcpy(fltType, "Source Port Extension");
					mip3dbg(("    Filter Type          : %s\n", fltType));

					srcportedp = (SrcPortExtData *) bmn[elem].mnFilters[i].filterModules[j].filterPtr;
					mip3dbg(("    Port Number          : %d\n\n", srcportedp->portNum));

					break;

				case FLT_SRCPORTRNG_EXT:
					strcpy(fltType, "Source Port Range Extension");
					mip3dbg(("    Filter Type          : %s\n", fltType));

					srcportrangeedp = (SrcPortRangeExtData *) bmn[elem].mnFilters[i].filterModules[j].filterPtr;
					mip3dbg(("    Port Number Range    : %d-%d\n\n",
										srcportrangeedp->minPortNum,
										srcportrangeedp->maxPortNum));

					break;

				case FLT_FRFORM_EXT:
					strcpy(fltType, "Free-Form Extension");
					mip3dbg(("     Filter Type            : %s\n", fltType));
					freeformedp = (FreeFormExtData *) bmn[elem].mnFilters[i].filterModules[j].filterPtr;
					mip3dbg(("     Offset, Value & Mask   : %d %s %s\n\n",
										freeformedp->offset,
										freeformedp->value,
										freeformedp->mask));

					break;

				// new filters

				case FLT_DSTADR_EXT:
					strcpy(fltType, "Destination Address Extension");
					mip3dbg(("    Filter Type          : %s\n", fltType));

					dstaddredp = (DstAddrExtData *) bmn[elem].mnFilters[i].filterModules[j].filterPtr;
					mip3dbg(("    Destination Address  : %s\n\n", ntoa(dstaddredp->dstAddr, addr1)));

					break;

				case FLT_DSTNET_EXT:
					strcpy(fltType, "Destination Network Extension");
					mip3dbg(("    Filter Type          : %s\n", fltType));

					dstnetedp = (DstNetExtData *) bmn[elem].mnFilters[i].filterModules[j].filterPtr;
					mip3dbg(("    Destin Address & Mask: %s/%s\n\n",
										ntoa(dstnetedp->dstAddr, addr1),
										ntoa(dstnetedp->dstNetMask, addr2)));

					break;

				case FLT_DSTPORT_EXT:
					strcpy(fltType, "Destination Port Extension");
					mip3dbg(("    Filter Type          : %s\n", fltType));

					dstportedp = (DstPortExtData *) bmn[elem].mnFilters[i].filterModules[j].filterPtr;
					mip3dbg(("    Port Number          : %d\n\n", dstportedp->portNum));

					break;

				case FLT_DSTPORTRNG_EXT:
					strcpy(fltType, "Destination Port Range Extension");
					mip3dbg(("    Filter Type          : %s\n", fltType));

					dstportrangeedp = (DstPortRangeExtData *) bmn[elem].mnFilters[i].filterModules[j].filterPtr;
					mip3dbg(("    Port Number Range    : %d-%d\n\n",
										dstportrangeedp->minPortNum,
										dstportrangeedp->maxPortNum));

					break;

				default:
					strcpy(fltType, "Cannot determine filter type");
					mip3dbg(("     Filter Type            : %s\n", fltType));

					break;
			}

		}
		mip3dbg(("\n"));

		// display distribution weights
		for(j = 0; j < MAX_FILTER_MODULES_PER_FILTER; j++) {
			if(bmn[elem].mnFilters[i].controlModules[j].weight == (-1))
				break;
			mip3dbg(("     Bound CoA & Weight   : %s %d \n",
						ntoa(bmn[elem].mnFilters[i].controlModules[j].boundCoA, addr1),
						bmn[elem].mnFilters[i].controlModules[j].weight));
		}
	}
}

/* This function checks to see if the filters of the specified
 * MN has changed
 */
int haveFiltersChanged(BoundMN bmn[], u32 mnAddr) {
	int i;

	for(i = 0; i < DEFAULT_MAX_BINDINGS; i++) {
		if((bmn[i].mnAddr == mnAddr) && bmn[i].filtersChanged)
			return TRUE;
	}
	return FALSE;
}

/* This function adds a blank entry to the filter table,
 * for the given MN. Has to be used when adding a new binding
 */
int addFilterTableEntry(BoundMN bmn[], u32 mnAddr) {
	int i, elem, free;

	elem = (-1);
	free = (-1);
	for(i = 0; i < DEFAULT_MAX_BINDINGS; i++) {
		if(bmn[i].mnAddr == mnAddr) {
			elem = i;
			break;
		}
		if(bmn[i].mnAddr == 0 && free == (-1))
			free = i;
	}

	if(elem == (-1) && free == (-1))
		return FALSE;

	if(elem == (-1)) {
 		bmn[free].filtersChanged = TRUE;
		bmn[free].mnAddr = mnAddr;
		bmn[free].mnFilterCount = 0;
	} else
 		bmn[elem].filtersChanged = TRUE;

	return TRUE;
}


/* This function applies the filters as an FA for the
 * the specified MN (before applying all existing filters
 * are removed
 */
void FAapplyFilters(BoundMN bmn[], u32 mnAddr) {
	int i;

	for(i = 0; i < DEFAULT_MAX_BINDINGS; i++) {
		if((bmn[i].mnAddr == mnAddr) && bmn[i].filtersChanged) {
			bmn[i].filtersChanged = FALSE;
			break;
		}
	}
}

/* This function applies the filters as an HA for the
 * the specified MN (before applying all existing filters
 * are removed). If it cannot apply any or all filters due
 * to any reason, it returns FALSE
 */
u32 prevDefaultFtlCoA = 0;

int HAapplyFilters(BoundMN bmn[], HaBindingEntry habe[], u32 mnAddr) {
	int i, j, elem, opelem, k, pos, target;
	char str[128], *sptr, addr1[32], addr2[32], str2[128];
	BAFExtData  *bafedp;
	ProtoExtData *protoedp;
	SrcAddrExtData *srcaddredp;
	SrcNetExtData *srcnetedp;
	SrcPortExtData *srcportedp;
	SrcPortRangeExtData *srcportrangeedp;
	FreeFormExtData *freeformedp;

	DstAddrExtData *dstaddredp;
	DstNetExtData *dstnetedp;
	DstPortExtData *dstportedp;
	DstPortRangeExtData *dstportrangeedp;

	u32 defaultFtlCoA;

	// get the default CoA
	defaultFtlCoA = getDefaultFilterCoA(bmn, habe, mnAddr);

	// if default filter coa is 0, then mn at home
	if(defaultFtlCoA != 0 && defaultFtlCoA != prevDefaultFtlCoA) {
		setFilterOp(FLTOP_ADD, mnAddr, defaultFtlCoA, DEFAULT_FILTER_SEQ,
						  getMarkValue(mnAddr, DEFAULT_FILTER_SEQ), (-1));
	}

	// find the element number of the filter set for the
	// given mnAddr
	for(i = 0; i < DEFAULT_MAX_BINDINGS; i++) {
		if(bmn[i].mnAddr == mnAddr) {
			elem = i;
			break;
		}
	}
	if(i >= DEFAULT_MAX_BINDINGS)
		return FALSE;

	// find the element number of the filter operations
	// set for the given mn Addr
	for(i = 0; i < DEFAULT_MAX_BINDINGS; i++) {
		if(mnFltOpTable[i].mnaddr == mnAddr) {
			opelem = i;
			break;
		}
	}
	if(i >= DEFAULT_MAX_BINDINGS)
		return FALSE;

	// remove the delete instruction for the default filter if
	// the CoA of defaultFlt not changed or if default Flt CoA
	// is 0 (meaning at home or program terminated)
	if((defaultFtlCoA != 0 && defaultFtlCoA != prevDefaultFtlCoA)
		 || defaultFtlCoA == 0) {
		 //leave the filter op table as it is
	} else {
		pos = MAX_FILTERS - 1;
		mnFltOpTable[opelem].fltOpTable[pos].opCode = (-1);
		mnFltOpTable[opelem].fltOpTable[pos].boundCoA = 0;
		mnFltOpTable[opelem].fltOpTable[pos].seq = (-1);
		mnFltOpTable[opelem].fltOpTable[pos].markValue = (-1);
	}
	prevDefaultFtlCoA = defaultFtlCoA;


	// create filter rules, one at a time, goes in the reverse order
	// as rule setting assigns filter priorities backwards
	for(i = (MAX_FILTERS - 1); i >= 0; i--) {

		if(mnFltOpTable[opelem].fltOpTable[i].opCode == (-1))
			continue;

		switch(mnFltOpTable[opelem].fltOpTable[i].opCode) {
			case FLTOP_DELETE:
				sprintf(str, " %d %d ", FLTOP_DELETE,
							mnFltOpTable[opelem].fltOpTable[i].markValue);
				//applyFilter(mnAddr, mnFltOpTable[opelem].fltOpTable[i].boundCoA, str);
				applyFilter(mnAddr, mnFltOpTable[opelem].fltOpTable[i].seq, str);

				break;

			case FLTOP_SHARE:
			case FLTOP_ADD:
				// do the same for add or share

				// this is default filter ; therefore special
				if(mnFltOpTable[opelem].fltOpTable[i].seq == DEFAULT_FILTER_SEQ) {
					sprintf(str, " %d %d %d %d %d %s %d", FLTOP_ADD,
								(-1),
								TG_ACCEPT,
								mnFltOpTable[opelem].fltOpTable[i].markValue,
								1,
								ntoa(mnFltOpTable[opelem].fltOpTable[i].boundCoA, addr1),
								100);
					applyFilter(mnAddr, mnFltOpTable[opelem].fltOpTable[i].seq, str);
					//applyFilter(mnAddr, mnFltOpTable[opelem].fltOpTable[i].boundCoA, str);
					break;
				}

				// find the filter details
				for(j = 0; j < MAX_FILTERS; j++) {
					if(bmn[elem].mnFilters[j].seq
						== mnFltOpTable[opelem].fltOpTable[i].seq)
						break;
				}
				if(j >= MAX_FILTERS)
					return FALSE;

				if(bmn[elem].mnFilters[j].filterModuleCount <= 0)
					continue;

				// build a string containing info of a filter, for example, following
				// filter data is placed in a string as,
				//
				// operation=add, modules=2, target=accept, mark=5, coa-count=2,
				// coa&weights=134.102.158.89 5 134.102.158.73 10
				// filter-modules=src-adr 134.102.158.1 src-port 12000
				//
				// str=2 2 2 5 2 134.102.158.89 5 134.102.158.73 10 69 134.102.158.1 71 12000

				*str2 = '\0';
				for(k = 0; k < MAX_FILTER_MODULES_PER_FILTER; k++) {
					if(bmn[elem].mnFilters[j].controlModules[k].weight == (-1))
						break;
					sprintf(str2 + strlen(str2), " %s %d ",
							ntoa(bmn[elem].mnFilters[j].controlModules[k].boundCoA, addr1),
							bmn[elem].mnFilters[j].controlModules[k].weight);

				}

				// decide what to do with flow, accept or drop
				if(bmn[elem].mnFilters[j].controlModules[0].weight == 0)
					target = TG_DROP;
				else
					target = TG_ACCEPT;

				sprintf(str, " %d %d %d %d %d ", mnFltOpTable[opelem].fltOpTable[i].opCode,
							bmn[elem].mnFilters[j].filterModuleCount,
							target,
							bmn[elem].mnFilters[j].markValue,
							k);
				strcat(str, str2);

				for(k = 0; k < bmn[elem].mnFilters[j].filterModuleCount; k++) {
					sptr = str + strlen(str);
					switch(bmn[elem].mnFilters[j].filterModules[k].filterType) {
						case FLT_BAF_EXT:
							bafedp = (BAFExtData *)
								bmn[elem].mnFilters[j].filterModules[k].filterPtr;
							sprintf(sptr, " %d %d ", FLT_BAF_EXT, bafedp->DSCP);
							break;
						case FLT_PROTO_EXT:
							protoedp = (ProtoExtData *)
								bmn[elem].mnFilters[j].filterModules[k].filterPtr;
							sprintf(sptr, " %d %d ", FLT_PROTO_EXT, protoedp->protoNum);
							break;
						case FLT_SRCADR_EXT:
							srcaddredp = (SrcAddrExtData *)
								bmn[elem].mnFilters[j].filterModules[k].filterPtr;
							sprintf(sptr, " %d %s ", FLT_SRCADR_EXT, ntoa(srcaddredp->srcAddr, addr1));
							break;
						case FLT_SRCNET_EXT:
							srcnetedp = (SrcNetExtData *)
								bmn[elem].mnFilters[j].filterModules[k].filterPtr;
							sprintf(sptr, " %d %s %s ", FLT_SRCNET_EXT,
										ntoa(srcnetedp->srcAddr, addr1),
										ntoa(srcnetedp->srcNetMask, addr2));
							break;
						case FLT_SRCPORT_EXT:
							srcportedp = (SrcPortExtData *)
								bmn[elem].mnFilters[j].filterModules[k].filterPtr;
							sprintf(sptr, " %d %d ", FLT_SRCPORT_EXT, srcportedp->portNum);
							break;
						case FLT_SRCPORTRNG_EXT:
							srcportrangeedp = (SrcPortRangeExtData *)
								bmn[elem].mnFilters[j].filterModules[k].filterPtr;
							sprintf(sptr, " %d %d %d ", FLT_SRCPORTRNG_EXT,
										srcportrangeedp->minPortNum,
										srcportrangeedp->maxPortNum);
							break;
						case FLT_FRFORM_EXT:
							printf("Free Form not implemented, yet \n");
							// not implemented
							break;

						// new filters
						case FLT_DSTPORT_EXT:
							dstportedp = (DstPortExtData *)
								bmn[elem].mnFilters[j].filterModules[k].filterPtr;
							sprintf(sptr, " %d %d ", FLT_DSTPORT_EXT, dstportedp->portNum);
							break;
						case FLT_DSTPORTRNG_EXT:
							dstportrangeedp = (DstPortRangeExtData *)
								bmn[elem].mnFilters[j].filterModules[k].filterPtr;
							sprintf(sptr, " %d %d %d ", FLT_DSTPORTRNG_EXT,
										dstportrangeedp->minPortNum,
										dstportrangeedp->maxPortNum);
							break;

						default:
							printf("Non-implemented filter in table \n");
							break;
					}
				}
				printf("FILTER STR = ");
				printf(str);
				printf("\n");

				// call filter application function with the mnAddr & seq (filter index)
				applyFilter(mnAddr, mnFltOpTable[opelem].fltOpTable[i].seq, str);

				break;
			default:
				printf("Non-implemented operation \n");
				break;
		}
		mnFltOpTable[opelem].fltOpTable[i].opCode = (-1);
		mnFltOpTable[opelem].fltOpTable[i].boundCoA = 0;
		mnFltOpTable[opelem].fltOpTable[i].seq = (-1);
		mnFltOpTable[opelem].fltOpTable[i].markValue = (-1);
	}

	// flush routing cache to effect the new routing entries
	flushRouteCache();

	// update the states once filters have been applied
	bmn[elem].filtersChanged = FALSE;

	// place a delete in the operations table, so that the deletion
	// of default filter is done automatically ; place it at end
	setFilterOp(FLTOP_DELETE, mnAddr, defaultFtlCoA, DEFAULT_FILTER_SEQ,
						  getMarkValue(mnAddr, DEFAULT_FILTER_SEQ), (MAX_FILTERS - 1));

	return TRUE;
}

/*
 * find the default filter to route any trafic that does not
 * conform to the existing filter rules, default filter is the
 * binding that has no filters (or the first binding)
 */
u32 getDefaultFilterCoA(BoundMN bmn[], HaBindingEntry habe[], u32 mnAddr) {
	int fltTableElem, i, j, found;
	u32 firstCoA;

	// find the element in the filter table
	fltTableElem = (-1);
	for(i = 0;i < DEFAULT_MAX_BINDINGS; i++) {
		if(bmn[i].mnAddr == mnAddr) {
			fltTableElem = i;
			break;
		}
	}

	// check with the binding table to find out what CoA
	// to use as the default filter
	firstCoA = 0;
	for(i = 0; i < DEFAULT_MAX_BINDINGS; i++) {
		// some other MN entry or blank entry
		if((habe[i].haBindingMN != mnAddr) || (habe[i].haBindingTimeRemaining == 0))
			continue;

		// if no filters are available, first binding's CoA is
		// taken as the default filter
		if(fltTableElem == (-1)	|| bmn[fltTableElem].mnFilterCount <= 0) {
			return habe[i].haBindingCOA;
		}

		//found = FALSE;
		//for(j = 0; j < MAX_FILTERS; j++) {
		//	if(bmn[fltTableElem].mnFilters[j].boundCoA == habe[i].haBindingCOA) {
		//		found = TRUE;
		//		break;
		//	}
		//}
		//if(!found)
		//	return habe[i].haBindingCOA;

		// if all the bindings have filters the first binding's
		// CoA is taken as default filter
		if(firstCoA == 0)
			firstCoA = habe[i].haBindingCOA;
	}

	return firstCoA;
}

int removeSelectedFilters(BoundMN bmn[], int *bmnCnt, u32 mnAddr, u32 coAddr) {
	int elem, i, j, found;

	// find the filterset related to mnAddr
	elem = (-1);
	for(i = 0; i < DEFAULT_MAX_BINDINGS; i++) {
		if(bmn[i].mnAddr == mnAddr) {
			elem = i;
			break;
		}
	}
	if(elem == (-1))
		return FALSE;

	if(bmn[elem].mnFilterCount <= 0)
		return FALSE;

	// find and delete the entries matching the CoA
	// in the table
	for(i = 0; i < MAX_FILTERS; i++) {

		// no filter, go to next
		if(bmn[elem].mnFilters[i].seq == (-1))
			continue;

		// find if CoA present
		found = FALSE;
		for(j = 0; j < MAX_FILTER_MODULES_PER_FILTER; j++) {
			if(bmn[elem].mnFilters[i].controlModules[j].weight == (-1))
				break;

			if(coAddr == bmn[elem].mnFilters[i].controlModules[j].boundCoA) {
				found = TRUE;
				break;
			}
		}
		// if CoA not present, go to next
		if(!found)
			continue;

		// if CoA found, delete filter
		setFilterOp(FLTOP_DELETE, mnAddr, coAddr,
						bmn[elem].mnFilters[i].seq,
						bmn[elem].mnFilters[i].markValue, (-1));


		// release the mark associated with a binding
		freeMarkValue(mnAddr, bmn[elem].mnFilters[i].seq);

		// once filter is deleted, init filter slot
		bmn[elem].mnFilters[i].filterModuleCount = 0;
		for(j = 0; j < MAX_FILTER_MODULES_PER_FILTER; j++) {
			if(bmn[elem].mnFilters[i].controlModules[j].weight == (-1))
				break;

			bmn[elem].mnFilters[i].controlModules[j].boundCoA = 0;
			bmn[elem].mnFilters[i].controlModules[j].weight = (-1);
		}
		bmn[elem].mnFilters[i].seq = (-1);
		bmn[elem].mnFilters[i].markValue = (-1);
		bmn[elem].mnFilterCount--;
	}


	// if no filters for the mn, remove entry
	//if(bmn[elem].mnFilterCount <= 0) {
	//	bmn[elem].mnAddr = 0;
	//	bmn[elem].mnFilterCount = 0;
	//	bmn[elem].filtersChanged = FALSE;
	//	(*bmnCnt)--;
	//}

	// rearrange the filters
	arrangeFiltersSequentially(bmn, mnAddr);

	// sort the filters
	sortFilters(bmn, mnAddr);

	return TRUE;
}

/* Function to sort the filter table according to the sequence
 * number of each filter
 */
int sortFilters(BoundMN bmn[], u32 mnAddr) {
	int elem, i, j;
	Filter tempFlt;

	// find the filterset related to mnAddr
	elem = (-1);
	for(i = 0; i < DEFAULT_MAX_BINDINGS; i++) {
		if(bmn[i].mnAddr == mnAddr) {
			elem = i;
			break;
		}
	}
	if(elem == (-1))
		return TRUE;

	if(bmn[elem].mnFilterCount <= 0)
		return TRUE;

	// perform a bubble sort
	for(i = 0; i < (MAX_FILTERS - 1); i++) {
		if(bmn[elem].mnFilters[i].seq == (-1))
			continue;
		for(j = (i + 1); j < MAX_FILTERS; j++) {
			if(bmn[elem].mnFilters[j].seq == (-1))
				continue;
			if(bmn[elem].mnFilters[i].seq <= bmn[elem].mnFilters[j].seq)
				continue;
			memcpy(&tempFlt, &(bmn[elem].mnFilters[j]), sizeof(Filter));
			memcpy(&(bmn[elem].mnFilters[j]),&(bmn[elem].mnFilters[i]),
									sizeof(Filter));
			memcpy(&(bmn[elem].mnFilters[i]), &tempFlt, sizeof(Filter));
		}
	}
}

/* Function to rearrange the filters removing the gaps
 * in between them.
 */
int arrangeFiltersSequentially(BoundMN bmn[], u32 mnAddr) {
	int elem, i, j, k;

	// find the filterset related to mnAddr
	elem = (-1);
	for(i = 0; i < DEFAULT_MAX_BINDINGS; i++) {
		if(bmn[i].mnAddr == mnAddr) {
			elem = i;
			break;
		}
	}
	if(elem == (-1))
		return TRUE;

	if(bmn[elem].mnFilterCount <= 0)
		return TRUE;

	// look at the elements and shift each one down, if a slot is available
	for(i = 0; i < (MAX_FILTERS - 1); i++) {
		if(bmn[elem].mnFilters[i].seq != (-1))
			continue;

		for(j = (i + 1); j < MAX_FILTERS; j++) {
			if(bmn[elem].mnFilters[j].seq != (-1))
				break;
		}
		if(j < MAX_FILTERS) {
			memcpy(&(bmn[elem].mnFilters[i]),&(bmn[elem].mnFilters[j]),
									sizeof(Filter));
			bmn[elem].mnFilters[j].filterModuleCount = 0;

			for(k = 0; k < MAX_FILTER_MODULES_PER_FILTER; k++) {
				if(bmn[elem].mnFilters[j].controlModules[k].weight == (-1))
					break;
				bmn[elem].mnFilters[j].controlModules[k].boundCoA = 0;
				bmn[elem].mnFilters[j].controlModules[k].weight = (-1);
			}
			bmn[elem].mnFilters[j].seq = (-1);
			bmn[elem].mnFilters[j].markValue = (-1);
		}
	}
}

/* Function to find a mark number to use for marking
 * packets using iptables. A mark is assigned to a
 * binding to send any packet conforming to any
 * filter that is associated with a binding
 */
int getMarkValue(u32 mnAddr, int fltSeq) {
	int i, free;

	if(bmarkInit) {
		bmarkInit = FALSE;
		for(i = 0; i < MAX_MARKS_SUPPORTED; i++) {
			bmark[i].mnaddr = 0;
			bmark[i].fltSeq = (-1);
		}
	}

	free = (-1);
	for(i = 0; i < MAX_MARKS_SUPPORTED; i++) {
		if(bmark[i].mnaddr == mnAddr && bmark[i].fltSeq == fltSeq)
			return (i + 1);
		if(free == (-1) && bmark[i].mnaddr == 0 && bmark[i].fltSeq == (-1))
			free = i;
	}

	if(free != (-1)) {
		bmark[free].mnaddr = mnAddr;
		bmark[free].fltSeq = fltSeq;
		return (free + 1);
	}

	return (-1);
}

/* Free the mark value assigned to a binding
 */
void freeMarkValue(u32 mnAddr, int fltSeq) {
	int i;

	for(i = 0; i < MAX_MARKS_SUPPORTED; i++) {
		if(bmark[i].mnaddr == mnAddr && bmark[i].fltSeq == fltSeq) {
			bmark[i].mnaddr = 0;
			bmark[i].fltSeq = (-1);
			return;
		}
	}
	return;
}


/* Set the filter operations table with the filter operation
 */
void setFilterOp(int opCode, u32 mnAddr, u32 coAddr, int seq, int markValue, int pos) {
	int elem, free, i, j;

	if(fltOpInit) {
		fltOpInit = FALSE;
		for(i = 0; i < DEFAULT_MAX_BINDINGS; i++) {
			mnFltOpTable[i].mnaddr = 0;
			for(j = 0; j < MAX_FILTERS; j++) {
				mnFltOpTable[i].fltOpTable[j].opCode = (-1);
				mnFltOpTable[i].fltOpTable[j].boundCoA = 0;
				mnFltOpTable[i].fltOpTable[j].seq = (-1);
				mnFltOpTable[i].fltOpTable[j].markValue = (-1);
			}
		}
	}


	free = elem = (-1);
	for(i = 0; i < DEFAULT_MAX_BINDINGS; i++) {
		if(mnFltOpTable[i].mnaddr == mnAddr) {
			elem = i;
			break;
		}
		if(free == (-1) && mnFltOpTable[i].mnaddr == 0)
			free = i;
	}
	if(elem == (-1) && free == (-1))
		return;

	if(elem == (-1))
		elem = free;

	mnFltOpTable[elem].mnaddr = mnAddr;

	// when inserting, if position not given use next available else insert
	// at the specified position
	if(pos < 0) {

		for(i = (MAX_FILTERS - 2); i >= 0; i--) {
		//for(i = 0; i < MAX_FILTERS; i++) {
			if(mnFltOpTable[elem].fltOpTable[i].opCode != (-1))
				continue;

			mnFltOpTable[elem].fltOpTable[i].opCode = opCode;
			mnFltOpTable[elem].fltOpTable[i].boundCoA = coAddr;
			mnFltOpTable[elem].fltOpTable[i].seq = seq;
			mnFltOpTable[elem].fltOpTable[i].markValue = markValue;
			break;
		}
	} else {
		mnFltOpTable[elem].fltOpTable[pos].opCode = opCode;
		mnFltOpTable[elem].fltOpTable[pos].boundCoA = coAddr;
		mnFltOpTable[elem].fltOpTable[pos].seq = seq;
		mnFltOpTable[elem].fltOpTable[pos].markValue = markValue;
	}

}
