/*

  Sun Public License Notice

 The contents of this file are subject to the Sun Public License Version
 1.0 (the "License"). You may not use this file except in compliance with
 the License. A copy of the License is available at http://www.sun.com/

 The Initial Developer of the Original Code is Asanga D. Udugama
 Portions created by Asanga D. Udugama are Copyright (C) 2003.
 All Rights Reserved.

 Contributor(s): none

*/

/*
 * An implementation of Filers for Mobile-IP as described in
 * draft-nomad-mobileip-filters-05.txt.
 * Author: Asanga Udugama (adu@comnets.uni-bremen.de)
 * info.c : Information passing routines used in Mobile IP.
 */

#pragma ident   "@(#)info.c 1.1 2003/10/27"

#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pthread.h>

#include "mip.h"
#include "agent.h"


#define MAX_GUI_CONNECTIONS	20
#define DEFAULT_GUI_PORT	26000

// all the information variables that will
// be transmitted
extern int mipdebug;
extern int IDfreshnessSlack;
extern int regLifetime;
extern int advLifetime;
extern int periodicInterval;
extern int granularity;
extern int advertiseOnBcast;

extern FaVisitorEntry faVisitorTable[DEFAULT_MAX_VISITORS];
extern int faVisitorTableSize;

extern HaBindingEntry haBindingTable[DEFAULT_MAX_BINDINGS];
extern int haBindingTableSize;

extern BoundMN boundMNTable[DEFAULT_MAX_BINDINGS];
extern int boundMNTableSize;

extern HaMobileNodeEntry haMobileNodeTable[DEFAULT_MAX_MOBILE_NODES];
extern int haMobileNodeTableSize;


// holds info related to each connected GUI
typedef struct connectedgui {
	u32 ipAddr;
	int port;
	long dataSent;
} ConnectedGUI;
ConnectedGUI connGUITable[MAX_GUI_CONNECTIONS];
int connGUICount = 0;

int GUIRecvActive = 0;
// socket variables
int GUIRecvPortNum = DEFAULT_GUI_PORT;
struct sockaddr_in GUIRecvIPAddr;
int GUIRecvSocket;
int GUISendSocket;

// thread variables
static pthread_t monitorGUIs;
static pthread_mutex_t connGUITableUpdateLock;
static pthread_attr_t monitorGUIsAttr;
int monitorGUIsDummy = 1;

// function prototypes
int initInfoModule();
void *monitorGUIConnections(void *ignoreThis);
int sendInfoToGUI(ConnectedGUI *cgui);
int sendInfoToAllGUIs();
char *buildSystemInfo(char *buffer);
char *buildFAVisitorInfo(char *buffer);
char *buildHABindingInfo(char *buffer);
char *buildBoundMNInfo(char *buffer);
char *buildHAMobileNodeInfo(char *buffer);
int termInfoModule();

/**
* Creates the sockets and threads to send information to
* GUIs
*/
int initInfoModule() {
	int i;

	// if info module not activated
	if(!GUIRecvActive)
		return;

	for(i = 0; i < MAX_GUI_CONNECTIONS; i++) {
		connGUITable[i].ipAddr = 0;
		connGUITable[i].port = 0;
		connGUITable[i].dataSent = 0;
	}

	GUIRecvSocket = socket(PF_INET, SOCK_DGRAM, 0);
	bzero(&GUIRecvIPAddr, sizeof(GUIRecvIPAddr));
	GUIRecvIPAddr.sin_family = AF_INET;
	GUIRecvIPAddr.sin_port = htons(GUIRecvPortNum);
	GUIRecvIPAddr.sin_addr.s_addr = INADDR_ANY;
	if ( bind(GUIRecvSocket, (struct sockaddr*)&GUIRecvIPAddr, sizeof(GUIRecvIPAddr)) != 0 ) {
        	mip0dbg(("Error: Could not create GUI listen socket in initInfoModule().\n"));
		return (-1);
	}

	GUISendSocket = socket(PF_INET, SOCK_DGRAM, 0);

	pthread_mutex_init(&connGUITableUpdateLock, NULL);
	pthread_attr_init(&monitorGUIsAttr);
	pthread_attr_setdetachstate(&monitorGUIsAttr, PTHREAD_CREATE_DETACHED);
	if ((pthread_create(&monitorGUIs, &monitorGUIsAttr,
			monitorGUIConnections, (void *) &monitorGUIsDummy) != 0)) {
		mip1dbg(("Error while creating monitorGUIs thread.\n"));
		return (-1);
	}

	return(0);
}

/**
* Thread function that waits for connections from GUIs
*/
void *monitorGUIConnections(void *ignoreThis) {
	int i = (-1), len;
	char buffer[64], chIPAddr[32], chPort[16], dummy[24];
	struct sockaddr_in addr;

	while(1) {
		len = recv(GUIRecvSocket, buffer, sizeof(buffer), 0);

		// if not GET, forget it
		if(strncmp(buffer, "GET", 3) != 0)
			continue;

		// if tooooooo many connections, forget it too
		if((i + 1) >= MAX_GUI_CONNECTIONS)
			continue;

		buffer[len] = '\0';
		if((sscanf(buffer, "%s %s %s", dummy, chIPAddr, chPort)) != 3)
			continue;

		// lock your self in a room, and do as you please
		pthread_mutex_lock(&connGUITableUpdateLock);

		i++;
		connGUICount++;

		connGUITable[i].ipAddr = inet_addr(chIPAddr);
		connGUITable[i].port = atoi(chPort);

		sendInfoToGUI(&connGUITable[i]);

		pthread_mutex_unlock(&connGUITableUpdateLock);
	}
}

/**
* Function to send info to a specific GUI
*/
int sendInfoToGUI(ConnectedGUI *cgui) {
	char buffer[2048];
	struct sockaddr_in addr;

	// if info module not activated
	if(!GUIRecvActive)
		return;

	if(cgui == NULL || cgui->ipAddr == 0)
		return (-1);

	bzero(&addr, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(cgui->port);
	addr.sin_addr.s_addr = cgui->ipAddr;

	// build system info & send
	buildSystemInfo(buffer);
	sendto(GUISendSocket, buffer, strlen(buffer), 0, (struct sockaddr *) &addr, sizeof(addr));

	// build FA visitor info
	buildFAVisitorInfo(buffer);
	sendto(GUISendSocket, buffer, strlen(buffer), 0, (struct sockaddr *) &addr, sizeof(addr));

	// build HA binding info
	buildHABindingInfo(buffer);
	sendto(GUISendSocket, buffer, strlen(buffer), 0, (struct sockaddr *) &addr, sizeof(addr));

	// build Bound MN info
	buildBoundMNInfo(buffer);
	sendto(GUISendSocket, buffer, strlen(buffer), 0, (struct sockaddr *) &addr, sizeof(addr));

	// build FA visitor info
	buildHAMobileNodeInfo(buffer);
	sendto(GUISendSocket, buffer, strlen(buffer), 0, (struct sockaddr *) &addr, sizeof(addr));

	return 0;
}

/**
* Function to send all the information to all connected
* GUIs
*/
int sendInfoToAllGUIs() {
	int i;

	// if info module not activated
	if(!GUIRecvActive)
		return;

	// lock & send info
	pthread_mutex_lock(&connGUITableUpdateLock);

	for(i = 0; i < connGUICount; i++) {
		if(connGUITable[i].ipAddr == 0 || connGUITable[i].port == 0)
			continue;

		sendInfoToGUI(&connGUITable[i]);
	}

	pthread_mutex_unlock(&connGUITableUpdateLock);
}

/**
* This function fills the system information in to the sent
* buffer
*/
char *buildSystemInfo(char *buffer) {

	strcpy(buffer, "Sysinfo@");
	sprintf(buffer + strlen(buffer), "mipdebug %d;", mipdebug);
	sprintf(buffer + strlen(buffer), "IDfreshnessSlack %d;", IDfreshnessSlack);
	sprintf(buffer + strlen(buffer), "regLifetime %d;", regLifetime);
	sprintf(buffer + strlen(buffer), "advLifetime %d;", advLifetime);
	sprintf(buffer + strlen(buffer), "periodicInterval %d;", periodicInterval);
	sprintf(buffer + strlen(buffer), "granularity %d;", granularity);
	sprintf(buffer + strlen(buffer), "advertiseOnBcast %d:", advertiseOnBcast);

	return buffer;
}


/**
* Function to fill the FA vistor information to the
* sent buffer
*/
char *buildFAVisitorInfo(char *buffer) {
	int i;
	char chAddr1[24];

	strcpy(buffer, "FaVisitorEntry@");
	for(i = 0; i < faVisitorTableSize; i++) {
		sprintf(buffer + strlen(buffer), "%s;", ntoa(faVisitorTable[i].faVisitorAddr, chAddr1));
		sprintf(buffer + strlen(buffer), "%s;", ntoa(faVisitorTable[i].faVisitorIfaceAddr, chAddr1));
		sprintf(buffer + strlen(buffer), "%d;", faVisitorTable[i].faVisitorRegIsAccepted);
		sprintf(buffer + strlen(buffer), "%d;", faVisitorTable[i].faVisitorRegFlags);
		sprintf(buffer + strlen(buffer), "%d;", faVisitorTable[i].faVisitorPort);
		sprintf(buffer + strlen(buffer), "%s;", ntoa(faVisitorTable[i].faVisitorHomeAddr, chAddr1));
		sprintf(buffer + strlen(buffer), "%s;", ntoa(faVisitorTable[i].faVisitorHomeAgentAddr, chAddr1));
		sprintf(buffer + strlen(buffer), "%s;", ntoa(faVisitorTable[i].faVisitorCOAddr, chAddr1));
		sprintf(buffer + strlen(buffer), "%d;", faVisitorTable[i].faVisitorTimeGranted);
		sprintf(buffer + strlen(buffer), "%d;", faVisitorTable[i].faVisitorTimeRemaining);
		sprintf(buffer + strlen(buffer), "%d;", faVisitorTable[i].faVisitorRegIDHigh);
		sprintf(buffer + strlen(buffer), "%d:", faVisitorTable[i].faVisitorRegIDLow);
	}
	if(faVisitorTableSize <= 0)
		strcat(buffer, "Null");

	return buffer;
}

/**
* Function to fill the HA binding information to the
* sent buffer
*/
char *buildHABindingInfo(char *buffer) {
	int i;
	char chAddr1[24];

	strcpy(buffer, "HaBindingEntry@");
	for(i = 0; i < haBindingTableSize; i++) {
		sprintf(buffer + strlen(buffer), "%s;", ntoa(haBindingTable[i].haBindingMN, chAddr1));
		sprintf(buffer + strlen(buffer), "%s;", ntoa(haBindingTable[i].haBindingCOA, chAddr1));
		sprintf(buffer + strlen(buffer), "%s;", ntoa(haBindingTable[i].haBindingSrcAddr, chAddr1));
		sprintf(buffer + strlen(buffer), "%d;", haBindingTable[i].haBindingTimeGranted);
		sprintf(buffer + strlen(buffer), "%d;", haBindingTable[i].haBindingTimeRemaining);
		sprintf(buffer + strlen(buffer), "%d;", haBindingTable[i].haBindingSrcPort);
		sprintf(buffer + strlen(buffer), "%d:", haBindingTable[i].haBindingRegFlags);
	}
	if(haBindingTableSize <= 0)
		strcat(buffer, "Null");

	return buffer;
}

/**
* Function to fill the filter information to the
* sent buffer
*/
char *buildBoundMNInfo(char *buffer) {
	char tbuffer[256], chAddr1[24];
	int i, found, j, k, pick;
	BAFExtData  *bafedp;
	ProtoExtData *protoedp;
	SrcAddrExtData *srcaddredp;
	SrcNetExtData *srcnetedp;
	SrcPortExtData *srcportedp;
	SrcPortRangeExtData *srcportrangeedp;
	DstPortExtData *dstportedp;
	DstPortRangeExtData *dstportrangeedp;
	FreeFormExtData *freeformedp;
	int boundCnt;

	strcpy(buffer, "BoundMN@");
	for(i = 0; i < DEFAULT_MAX_BINDINGS; i++) {
		if(boundMNTable[i].mnAddr == 0)
			continue;

		sprintf(buffer + strlen(buffer), "%s;", ntoa(boundMNTable[i].mnAddr, chAddr1));
		sprintf(buffer + strlen(buffer), "%d;", boundMNTable[i].mnFilterCount);

		for(j = 0; j < MAX_FILTERS; j++) {
			if(boundMNTable[i].mnFilters[j].seq == (-1))
				continue;

			// get filter module info
			sprintf(buffer + strlen(buffer), "%d;", boundMNTable[i].mnFilters[j].filterModuleCount);
			for(k = 0; k < boundMNTable[i].mnFilters[j].filterModuleCount; k++) {

				sprintf(buffer + strlen(buffer), "%d;",
					boundMNTable[i].mnFilters[j].filterModules[k].filterType);
				switch(boundMNTable[i].mnFilters[j].filterModules[k].filterType) {
					case FLT_BAF_EXT:
						bafedp = (BAFExtData *) boundMNTable[i].mnFilters[j].filterModules[k].filterPtr;
						sprintf(tbuffer, "%d", bafedp->DSCP);
						break;
					case FLT_PROTO_EXT:
						protoedp = (ProtoExtData *) boundMNTable[i].mnFilters[j].filterModules[k].filterPtr;
						sprintf(tbuffer, "%d", protoedp->protoNum);
						break;
					case FLT_SRCADR_EXT:
						srcaddredp = (SrcAddrExtData *) boundMNTable[i].mnFilters[j].filterModules[k].filterPtr;
						sprintf(tbuffer, "%s", ntoa(srcaddredp->srcAddr, chAddr1));
						break;
					case FLT_SRCNET_EXT:
						srcnetedp = (SrcNetExtData *) boundMNTable[i].mnFilters[j].filterModules[k].filterPtr;
						sprintf(tbuffer, "%s/%d", ntoa(srcnetedp->srcAddr, chAddr1),
								numBitsInMask(srcnetedp->srcNetMask));
						break;
					case FLT_SRCPORT_EXT:
						srcportedp = (SrcPortExtData *) boundMNTable[i].mnFilters[j].filterModules[k].filterPtr;
						sprintf(tbuffer, "%d", srcportedp->portNum);
						break;
					case FLT_SRCPORTRNG_EXT:
						srcportrangeedp = (SrcPortRangeExtData *) boundMNTable[i].mnFilters[j].filterModules[k].filterPtr;
						sprintf(tbuffer, "%d-%d", srcportrangeedp->minPortNum,
									srcportrangeedp->maxPortNum);
						break;
					case FLT_DSTPORT_EXT:
						dstportedp = (DstPortExtData *) boundMNTable[i].mnFilters[j].filterModules[k].filterPtr;
						sprintf(tbuffer, "%d", dstportedp->portNum);
						break;
					case FLT_DSTPORTRNG_EXT:
						dstportrangeedp = (DstPortRangeExtData *) boundMNTable[i].mnFilters[j].filterModules[k].filterPtr;
						sprintf(tbuffer, "%d-%d", dstportrangeedp->minPortNum,
									dstportrangeedp->maxPortNum);
						break;
					case FLT_FRFORM_EXT:
					default:
						sprintf(tbuffer, "not-supported");
					break;
				}
				sprintf(buffer + strlen(buffer), "%s;", tbuffer);
			}


			// get bound ifc info
			boundCnt = 0;
			for(k = 0; k < MAX_FILTER_MODULES_PER_FILTER; k++) {
				if(boundMNTable[i].mnFilters[j].controlModules[k].weight == (-1))
					continue;
				boundCnt++;
			}
			sprintf(buffer + strlen(buffer), "%d;", boundCnt);
			for(k = 0; k < MAX_FILTER_MODULES_PER_FILTER; k++) {
				if(boundMNTable[i].mnFilters[j].controlModules[k].weight == (-1))
					continue;

				sprintf(buffer + strlen(buffer), "%s;",
					ntoa(boundMNTable[i].mnFilters[j].controlModules[k].boundCoA, chAddr1));
				sprintf(buffer + strlen(buffer), "%d;",
					boundMNTable[i].mnFilters[j].controlModules[k].weight);
			}
			sprintf(buffer + strlen(buffer), "%d;", boundMNTable[i].mnFilters[j].seq);
		}
	}
	if(boundMNTableSize <= 0)
		strcat(buffer, "Null");

	return buffer;
}


/**
* Function to fill the Mobile Node info to the
* sent buffer
*/
char *buildHAMobileNodeInfo(char *buffer) {
	int i;
	char chAddr1[24];

	strcpy(buffer, "HaMobileNodeEntry@");

	for(i = 0; i < haMobileNodeTableSize; i++) {
		sprintf(buffer + strlen(buffer), "%s;",ntoa(haMobileNodeTable[i].haMnAddr, chAddr1));
		sprintf(buffer + strlen(buffer), "%s;",ntoa(haMobileNodeTable[i].haBindingIfaceAddr, chAddr1));
		sprintf(buffer + strlen(buffer), "%d;",haMobileNodeTable[i].haMnRegIDHigh);
		sprintf(buffer + strlen(buffer), "%d:",haMobileNodeTable[i].haMnRegIDLow);
	}
	if(haMobileNodeTableSize <= 0)
		strcat(buffer, "Null");

	return buffer;
}


/**
* Function to terminate info module
*/
int termInfoModule() {

	// if info module not activated
	if(!GUIRecvActive)
		return;

	// terminate thread
	pthread_kill(monitorGUIs, 0);

	// close sockets
	close(GUIRecvSocket);
	close(GUISendSocket);
}
