/*

  Sun Public License Notice

 The contents of this file are subject to the Sun Public License Version
 1.0 (the "License"). You may not use this file except in compliance with
 the License. A copy of the License is available at http://www.sun.com/

 The Original Code is the Sunlabs Mobile IP.
 The Initial Developer of the Original Code is Vipul Gupta.
 Portions created by Vipul Gupta are Copyright (C) 2003.
 All Rights Reserved.

 Contributor(s): Niko A. Fikouras (niko@comnets.uni-bremen.de)
		 Asanga D. Udugama (adu@comnets.uni-bremen.de)

*/

/*
 * An implementation of Mobile-IP as described in RFC 2002.
 * Author: Vipul Gupta (vgupta@eng.sun.com)
 * mip.h : Header file common to Mobile IP nodes and agents.
 */

#ifndef _MIP_H
#define _MIP_H

#pragma ident	"@(#)mip.h 1.2 98/11/30"

#ifdef __cplusplus
extern "C" {
#endif

/* This file contains definitions for structures used in all
 * Mobile IP-aware entities.
 */

#define MIPAGENT_VERSION "Sun Labs Beta 1.2 (Nov 1998)"

#ifdef MIP_DEBUG
#define mip0dbg(a)	printf a
#define	mip1dbg(a)	if (mipdebug) printf a
#define	mip2dbg(a)	if (mipdebug>1) printf a
#define	mip3dbg(a)	if (mipdebug>2) printf a
#else
#define	mip0dbg(a)	printf a
#define	mip1dbg(a)	/* */
#define	mip2dbg(a)	/* */
#define	mip3dbg(a)	/* */
#endif

#define MAX_KEY_LEN   32

#include <sys/time.h>
typedef unsigned int u32;
typedef char boolean;
typedef struct timeval TimeStamp;

typedef struct msae {
    u32           mipSecAddr;
    u32           mipSecSPI;
    int           mipSecAlgorithmType;
    int           mipSecAlgorithmMode;
    int           mipSecKeyLen;
    unsigned char mipSecKey[MAX_KEY_LEN];
    int           mipSecReplayMethod;
} MipSecAssocEntry;


typedef struct msve {
    u32           mipSecViolatorAddr;
    u32           mipSecViolationCounter;
    int           mipSecRecentViolationSPI;
    TimeStamp     mipSecRecentViolationTime;
    int           mipSecRecentViolationIDLow;
    int           mipSecRecentViolationIDHigh;
    int           mipSecRecentViolationReason;
} MipSecViolationEntry;

/* Flags used in mobility agent advertisements */
#define ADV_REGISTRATION_REQUIRED  0x80
#define ADV_IS_BUSY                0x40
#define ADV_IS_HOME_AGENT          0x20
#define ADV_IS_FOREIGN_AGENT       0x10
#define ADV_MIN_ENCAP              0x08
#define ADV_GRE_ENCAP              0x04
#define ADV_VJ_COMPRESSION         0x02

/* Possible addresses for agent advertisements */
#define LINK_MCAST_ADDR            "224.0.0.1"
#define LINK_BCAST_ADDR            "255.255.255.255"

/* Boolean values */
#define TRUE                       1
#define FALSE                      0

/* Flags used in registrations */
#define REG_SIMULTANEOUS_BINDINGS  0x80
#define REG_FWD_BROADCASTS         0x40
#define REG_DECAPSULATION_BY_MN    0x20
#define REG_MIN_ENCAP              0x10
#define REG_GRE_ENCAP              0x08
#define REG_VJ_COMPRESSION         0x04

/* Rejection codes from Foreign Agent */
#define FA_REASON_UNSPECIFIED      64
#define FA_ADM_PROHIBITED          65
#define FA_INSUFFICIENT_RESOURCES  66
#define FA_MN_AUTH_FAILURE         67 
#define	FA_HA_AUTH_FAILURE         68 
#define FA_REG_LIFETIME_TOO_LONG   69
#define	FA_POORLY_FORMED_REQUEST   70
#define	FA_POORLY_FORMED_REPLY     71
#define FA_ENCAP_UNAVAILABLE       72
#define FA_VJ_UNAVAILABLE          73
#define FA_HA_UNREACHABLE          88

/* Rejection codes from Home Agent */
#define HA_REASON_UNSPECIFIED      128
#define HA_ADM_PROHIBITED          129
#define HA_INSUFFICIENT_RESOURCES  130
#define HA_MN_AUTH_FAILURE         131
#define HA_FA_AUTH_FAILURE         132
#define HA_ID_MISMATCH             133
#define HA_POORLY_FORMED_REQUEST   134
#define HA_TOO_MANY_SIMULTANEOUS   135
#define HA_UNKNOWN_HOME_AGENT      136


#define NONE                        0
/* Authentication algorithm types */
#define MD5                         1

/* Authentication algorithm modes */
#define PREFIXSUFFIX                1

/* Replay method style */
#define TIMESTAMPS                  1
#define NONCES                      2

/* Encapsulation style */
#define IPIP                        1
#define GRE                         2
#define MINIMAL                     3

/* ICMP messages (we define them here for portability) */
#define ICMP_DEST_UNREACHABLE       3
#define ICMP_ROUTER_ADVERTISEMENT   9
#define ICMP_ROUTER_SOLICITATION   10

typedef struct icmphdr {
    unsigned char  type;
    unsigned char  code;
    unsigned short checksum;
    union {
        struct {
            unsigned char u_adv_num_addr;
	    unsigned char u_adv_addr_entry_size;
	    unsigned short u_adv_lifetime;
	} u_adv;
        unsigned long u_unused;
    } icmphdr_u;
} icmph;

#define icmpAdvNumAddr         icmphdr_u.u_adv.u_adv_num_addr
#define icmpAdvAddrEntrySize   icmphdr_u.u_adv.u_adv_addr_entry_size
#define icmpAdvLifetime        icmphdr_u.u_adv.u_adv_lifetime

/* Mobile IP Agent Advertisement Extension */

#define ADV_EXT_TYPE                16
#define ADV_PREFIX_EXT_TYPE         19
#define ADV_PADDING_EXT_TYPE         0

typedef struct aaext {
    unsigned char  type;
    unsigned char  length;
    unsigned short seqNum;
    unsigned short regLifetime;
    unsigned char  advFlags;
    unsigned char  reserved;
} advExt;

/* Mobile IP Registration Request and Reply */

#define REG_REQUEST_TYPE             1
#define REG_REPLY_TYPE               3
#define REG_MH_AUTH_EXT_TYPE        32
#define REG_MF_AUTH_EXT_TYPE        33
#define REG_FH_AUTH_EXT_TYPE        34

typedef struct rreq {
    unsigned char  type;   
    unsigned char  regFlags;
    unsigned short regLifetime;
    u32            homeAddr;
    u32            haAddr;
    u32            COAddr;
    u32            IDHigh;
    u32            IDLow;
} regRequest;

typedef struct rrep {
    unsigned char  type;   
    unsigned char  code;
    unsigned short regLifetime;
    u32            homeAddr;
    u32            haAddr;
    u32            IDHigh;
    u32            IDLow;
} regReply;

typedef struct authext {
    unsigned char  type;
    unsigned char  length;
    unsigned short SPIhi;
    unsigned short SPIlo;
} authExt;

/* Filter data definition for Mobile IP */

/* Mobile IP filter definitions */
#define FLT_DEL_EXT		65
#define FLT_CTL_EXT		66
#define FLT_BAF_EXT		67
#define FLT_PROTO_EXT		68
#define FLT_SRCADR_EXT		69
#define FLT_SRCNET_EXT		70
#define FLT_SRCPORT_EXT		71
#define FLT_SRCPORTRNG_EXT	72
#define FLT_FRFORM_EXT		73
#define FLT_DSTADR_EXT		74
#define FLT_DSTNET_EXT		75
#define FLT_DSTPORT_EXT		76
#define FLT_DSTPORTRNG_EXT	77

/* Possible Target Actions on Filters */
#define TG_DROP			0        //  Drop incoming packets without notification
//#define TG_REJECT		1        //  Reject incoming packets with notification
#define TG_ACCEPT      		2        //  Accept incoming packets
//#define TG_ACCEPT_AVOID_RO	3        //  Accept incoming packets but avoid route optimization
//#define TG_MASQ	      		4        //  Masquerade


#define MAX_FLT_CRITERIA_IN_ENTRY	5
#define MAX_FREEFORM_VAL_SIZE		20
#define MAX_FREEFORM_MASK_SIZE		20

#ifndef MAX_FILTERS
#define MAX_FILTERS			10
#endif

//#define ACT_INSERT_AT_BEGINING	0	// Insert at the beginning of existing
                 			//   filters for the registered address
//#define ACT_APPEND_AT_END	1	// append at the end of existing filters
					//   for the registered address
//#define ACT_DELETE		2	// Delete from an exiting filter
//#define ACT_REPLACE      	3	// Replace entry in an existing filter
//#define ACT_FLUSH      		4	// Flush all filter entries for address


typedef struct controlext {
	unsigned char type;
	unsigned char length;
	unsigned char flags;
	unsigned char idx;
} controlExt;

typedef struct delext {
	unsigned char type;
	unsigned char length;
	unsigned char idx[MAX_FILTERS];
} delExt;

typedef struct bafext {
	unsigned char type;
	unsigned char length;
	unsigned char DSCP;
} bafExt;

typedef struct protoext {
	unsigned char type;
	unsigned char length;
	unsigned char protoNum;
} protoExt;

typedef struct srcaddrext {
	unsigned char type;
	unsigned char length;
	unsigned char srcAddr[4];
} srcAddrExt;

struct netEntry {
	unsigned char addr[4];
	unsigned char netMask[4];
};

typedef struct srcnetext {
	unsigned char type;
	unsigned char length;
	struct netEntry srcNet;
} srcNetExt;

typedef struct srcportext {
	unsigned char type;
	unsigned char length;
	unsigned char portNum[2];
} srcPortExt;

struct portRangeEntry {
	unsigned char minPortNum[2];
	unsigned char maxPortNum[2];
};

typedef struct srcportrangeext {
	unsigned char type;
	unsigned char length;
	struct portRangeEntry portRange;
} srcPortRangeExt;

typedef struct freeformext {
	unsigned char type;
	unsigned char length;
	unsigned char offset[2];
	unsigned char valueNmask[(MAX_FREEFORM_VAL_SIZE/2) + (MAX_FREEFORM_MASK_SIZE/2)];
} freeFormExt;

// new filters

typedef struct dstaddrext {
	unsigned char type;
	unsigned char length;
	unsigned char dstAddr[4];
} dstAddrExt;

typedef struct dstnetext {
	unsigned char type;
	unsigned char length;
	struct netEntry dstNet;
} dstNetExt;

typedef struct dstportext {
	unsigned char type;
	unsigned char length;
	unsigned char portNum[2];
} dstPortExt;

typedef struct dstportrangeext {
	unsigned char type;
	unsigned char length;
	struct portRangeEntry portRange;
} dstPortRangeExt;




#ifdef __cplusplus
}
#endif

#endif /* _MIP_H */
