/*
 * Copyright (c) 1996 by Sun Microsystems, Inc.
 */
#ifndef _VTUNL_H
#define _VTUNL_H

#pragma ident	"@(#)vtunl.h 1.1 97/04/09 SMI"

#ifdef __cplusplus
extern "C" {
#endif

/* 
 * This file contains all of the "public" definitions for structures
 * and constants associated with the vtunl.
 */


/* Vtunl-specific ioctls are first. */

#define VTUNL_IOCTL                 0x1018

/*
 * Ioctls that set/get configuration info from 
 * individual vtunl instances. These ioctls MUST be
 * issued on the stream containing the vtunl instance
 * of interest.
 */

#define VTUNL_SETCONF_TRANSPORT     (VTUNL_IOCTL + 1)
#define VTUNL_SETCONF_DLSUSER       (VTUNL_IOCTL + 2)
#define VTUNL_GETCONF     (VTUNL_IOCTL + 3)

/* 
 * Ioctls for encapsulation services. These and others that 
 * follow may be issued on any vtunl instance.
 */
#define VTUNL_ENCAP_ADD   (VTUNL_IOCTL + 4)
#define VTUNL_ENCAP_REM   (VTUNL_IOCTL + 5)
#define VTUNL_ENCAP_OK    (VTUNL_IOCTL + 6)
#define VTUNL_ENCAP_SIZE  (VTUNL_IOCTL + 7)
#define VTUNL_ENCAP_GET   (VTUNL_IOCTL + 8)

/* Ioctls for decapsulation services */
#define VTUNL_DECAP_ADD   (VTUNL_IOCTL + 9)
#define VTUNL_DECAP_REM   (VTUNL_IOCTL + 10)
#define VTUNL_DECAP_OK    (VTUNL_IOCTL + 11)
#define VTUNL_DECAP_SIZE  (VTUNL_IOCTL + 12)
#define VTUNL_DECAP_GET   (VTUNL_IOCTL + 13)

/* Ioctls to set/get debug level */
#define VTUNL_DEBUG_SET   (VTUNL_IOCTL + 14)
#define VTUNL_DEBUG_GET   (VTUNL_IOCTL + 15)

/* Ioctls to set/get upper bound on PMTU for a tunnel exit. We 
 * won't need these once the IP module correctly informs UPPER 
 * LAYER PROTOCOLS of the need to fragment large packets.
 */
#define VTUNL_MAXPMTU_SET (VTUNL_IOCTL + 16)
#define VTUNL_MAXPMTU_GET (VTUNL_IOCTL + 17)

/* This should eventually go in <netinet/in.h> */
#define IPPROTO_MINENCAP    5

/*
 * Flag bits passed with registration requests. Only one of 
 * MINIMALENCAP or GREENCAP may be set at any time. If neither is
 * set, IPIPENCAP is assumed. For now, only IPIPENCAP and 
 * SIMULTANEOUS bindings are supported. The ordering here 
 * corresponds to the sequence in which these flags appear in 
 * Mobile-IP registrations: SBDMGV.
 */

#define SIMULTANEOUS        0x0080
#define FWDBROADCASTS       0x0040
#define USINGDYNAMIC        0x0020
#define MINIMALENCAP        0x0010
#define GREENCAP            0x0008
#define VJCOMPRESSION       0x0004

 
/*
 * Vtunl-specific ioctls use special structures to exchange 
 * information between a user process and a vtunl module. Each 
 * ioctl class needs its own structure so we end up using a union.
 *
 * The VTUNL_{GET|SET}_DEBUG ioctls use an integer field rather 
 * than a vtunl_req structure.
 *
 * All address fields are assumed to be in network byter order.
 * For VTUNL_DECAP_ADD, the least significant bit-pair in 
 * vr_decap_flags field is set to one of DECAP_IF_DST, DECAP_IF_SRC
 * or DECAP_IF_EITHER.
 */

#define DECAP_IF_MASK       0x0003

#define DECAP_IF_DST        0x0000
#define DECAP_IF_SRC        0x0001
#define DECAP_IF_EITHER     0x0002

struct vtunl_req {
        union {
  	        /* 
		 * The fields in encap_req are used by 
		 * VTUNL_{EN|DE}CAP_{ADD|REM|OK}.
		 */
 	        struct encap_reqs {
		        u32        target_addr;  /* address encapsulated */
		        u32        t_exit_addr;  /* tunnel exit */
		        u16        encap_flags;  /* control encapsulation */
		} encap_req;

		/* These fields are used by VTUNL_SETCONF ioctls. */

	        struct sconf_reqs {
		        u32        address;     /* tunnel entry/exit 
						 * or DLPI interface addr */
		        u32        addr_mask;   /* netmask for DLPI device */
		        u_char     ppa;         /* ppa for DLPI device */
                        u_char     ip_protocol; /* protocol subscribed to */
		        u16        mtu_bound;   /* upper bound on PMTU for the
						 * exit specified in address */
		} sconf_req;

	} vtunl_requ;

#define vr_target_addr     vtunl_requ.encap_req.target_addr
#define vr_t_exit_addr     vtunl_requ.encap_req.t_exit_addr
#define vr_encap_flags     vtunl_requ.encap_req.encap_flags
#define vr_decap_flags     vtunl_requ.encap_req.encap_flags
#define vr_address         vtunl_requ.sconf_req.address
#define vr_addr_mask       vtunl_requ.sconf_req.addr_mask
#define vr_ppa             vtunl_requ.sconf_req.ppa
#define vr_ip_protocol     vtunl_requ.sconf_req.ip_protocol
#define vr_mtu_bound       vtunl_requ.sconf_req.mtu_bound
};

/* Valid values for the type field in a vtunl_conf structure. */
#define VTUNL_INST_HEAD       0
#define VTUNL_INST_TRANSPORT  1
#define VTUNL_INST_DLSUSER    2
#define VTUNL_INST_UNKNOWN    3

/* Valid values for the state field in a vtunl_conf structure. */
#define VTUNL_STATE_INIT      0
#define VTUNL_STATE_ATTACHED  1
#define VTUNL_STATE_BOUND     2
#define VTUNL_STATE_MULTI     3
#define VTUNL_STATE_BAD       4

/* 
 * We use this as the target address when specifying a default
 * tunnel with VTUNL_ENCAP_ADD or VTUNL_ENCAP_REM. An entry
 * created with this target address is used to encapsulate packets
 * for all those destinations that do not have their own entries
 * in the encap_table.
 */
#define DEFAULT_ADDR		((u32) htonl(0x7fffffff)) /* 127.255.255.255 */
#define VTUNL_DUMMY_ADDR	(u_long)0xfffffffe      /* 255.255.255.254 */

/* This is used by the VTUNL_GETCONF ioctl. */
#define MAXMNAMESZ  11
struct vtunl_conf {
        int    type;        /* vtunl instance type */
        u_char ppa;         /* ppa of DLS provider */
        u_char ip_protocol; /* ip protocol subscribed to */
        u32    address;     /* IPv4 address of DLS provider or tunnel src */
        u32    addr_mask;   /* netmask assigned to DLS provider */
        ulong  last_req;    /* last request sent to ip or DLS provider */
        ulong  last_resp;   /* last response from ip or DLS provider */
        ulong  state;       /* as determined by msg exchange with ip */
                            /* or the DLS provider below vtunl instance */
        char   modname[MAXMNAMESZ+1];
                            /* either "ip" or name of DLS provider,eg "le" */
};

#ifdef __cplusplus
}
#endif

#endif /* _VTUNL_H */
