/*
 * Copyright (c) 1996 by Sun Microsystems, Inc.
 */
#ifndef _VTUNL_IOCTL_H
#define _VTUNL_IOCTL_H

#pragma ident	"@(#)vtunl_impl.h 1.1 97/04/09 SMI"

#ifdef __cplusplus
extern "C" {
#endif

#include "vtunl.h"

/* 
 * Contains the definitions of all structures used to store the vtunl's
 * state and procedures that manipulate it. Definitions that should be
 * user visible are in vtunl.h.
 */

#ifdef VTUNL_DEBUG
#define vtunl0dbg(a)	printf a
#define	vtunl1dbg(a)	if (vtunldebug) printf a
#define	vtunl2dbg(a)	if (vtunldebug>1) printf a
#define	vtunl3dbg(a)	if (vtunldebug>2) printf a
#else
#define	vtunl0dbg(a)	/* */
#define	vtunl1dbg(a)	/* */
#define	vtunl2dbg(a)	/* */
#define	vtunl3dbg(a)	/* */
#endif

#define IPIP_DEF_TTL        32      /* default TTL for an IPIP header  */
#define VTUNL_TIMER_PERIOD  (drv_usectohz(60000000)) 
                                    /* do house keeping tasks every minute.*/
#define VTUNL_MAXDLSDU      1500    /* Maximum size for Data Link Service 
				     * Data Unit, MTU advertised to IP */

/*
 * This structure stores all instance-specific information for
 * a vtunl module.
 */
struct vtunl_inst {
        struct vtunl_inst  *prev;  /* previous member in list */
        struct vtunl_inst  *next;  /* next member in list */
        queue_t            *q;     /* read q for this vtunl instance */
        struct vtunl_conf  vcs;    /* config. info. for this instance */
};

/* 
 * There are two kinds of targets, those that get encapsulation service
 * and those that get decapsulation service. Additionally, there are
 * tunnel exit points. All three structures are kept in hash tables
 * hashed on IPv4 addresses. For efficiency, we have a fourth hash
 * table of hosts that have requested forwarding of non-unicast packets.
 */

#define MAXEXITCOUNT  8

struct bcast_target {
        struct bcast_target  *bcast_target_next;  /* next target, same hash */
        u32                  bcast_target_addr;   /* its address */
        krwlock_t            decap_target_rwlock; /* lock */
        /* pointer to its encap_target entry */
        struct encap_target  *bcast_target_entry; 
};

struct decap_target {
        struct decap_target  *decap_target_next;  /* next target, same hash */
        u32                  decap_target_addr;   /* its address */
        krwlock_t            decap_target_rwlock; /* lock */

        /* 
         * The least significant bit pair in flags indicates what the
         * decap_target is allowed to be in the decapsulated
         * datagram. Possible values are: DECAP_IF_DST, DECAP_IF_SRC,
         * DECAP_IF_EITHER.
         */
        u16                  decap_target_flags;               
};

struct encap_target {
        struct encap_target  *encap_target_next;  /* next target, same hash */
        u32                  encap_target_addr;   /* its address */
        krwlock_t            encap_target_rwlock; /* lock */
        short int            encap_target_exit_count; /* num of exits in use */
        u16                  encap_target_encapflags[MAXEXITCOUNT];
        /* flags SBDMGVxx  are kept in lower byte */
        /* u16                 encap_target_lifetime[MAXEXITCOUNT]; */
        /* if we support registration timers within kernel --
         * not used for multicasting. May not be a good idea. */
        struct t_exit       *encap_target_exit[MAXEXITCOUNT];
};

struct t_exit {
        struct t_exit   *t_exit_next;     /* next tunl hashed to same value */
        u32             t_exit_addr;      /* its address */
        krwlock_t       t_exit_rwlock;    /* lock */
        int             t_exit_ref_count; /* num of encap targets using it */
        u16             t_exit_pmtu;      /* tunnel's Path MTU */
        u16             t_exit_maxpmtu;   /* current bound on path MTU */
        u16             t_exit_state;   
 
  /* 
   * Of the 16 bits in this field, currently only four are used. The
   * most significant bit is set when an ICMP host unreachable ought
   * to be sent for targets using this exit. The three least significant
   * bits are used as a 3-bit counter. The counter is used to discover
   * PMTU increases, i.e. the PMTU is increased whenever the counter
   * expires. The counter is set to 001 after an increase and to 111
   * after a PMTU decrease. As a result, PMTU increases are attempted
   * sooner immediately following a successful increase than after a
   * PMTU decrease.
   */
};

struct decap_fill {
        u32        addr;
        u16        flags;
};

struct exit_fill {
        u32        addr;
        u16        cur_pmtu;
        u16        max_pmtu;
        u16        flags;
};


/*
 * To avoid having three different kinds of hash tables, we define
 * the following structure and cast the other three to this for all
 * hash table manipulation. The top three fields in all 
 * these strcutures are of the same type.
 */

struct hash_entry {
        struct hash_entry   *hash_entry_next;
        u32                 hash_entry_addr;
        krwlock_t           hash_entry_lock;
};
  
#define HASH_TBL_SIZE   256

struct hash_table {
        int                 size;
        struct hash_entry   *hash_entries[HASH_TBL_SIZE];
};

#define HASH_BIT_COUNT      8

#ifdef _BIG_ENDIAN
#define HASHIT(l1)                                          \
		((u32)((l1 >> HASH_BIT_COUNT) ^ (l1))	    \
		& ~(~0 << HASH_BIT_COUNT))
#else
#define	HASHIT(l1)					    \
		((u32)((l1 << HASH_BIT_COUNT) ^ (l1)) >>    \
		(u32)(32 - HASH_BIT_COUNT))
#endif 

#ifdef __cplusplus
}
#endif
  
#endif /* _VTUNL_IOCTL_H */



