/*
 * Copyright (c) 1996 by Sun Microsystems, Inc.
 */

#pragma ident	"@(#)vtunlddi.c 1.2 97/08/29 SMI"

/*  
 *  Author: Vipul Gupta (vgupta@eng.sun.com)
 *
 *  Based on the atun code in the IPv6 distribution. - V.G.
 *
 *  Compile as:
 *      cc -g -D_KERNEL -I. -c vtunl.c vtunlddi.c
 *      ld -r -o vtunl vtunl.o vtunlddi.o
 */

#include <sys/types.h>
#include <sys/conf.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/stream.h>
#include <sys/errno.h>
#include <sys/modctl.h>
#include <sys/kmem.h>
#include <sys/stat.h>
#include <sys/cmn_err.h>
#include <inet/common.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>

#include "vtunl_impl.h"

extern queue_t *timer_q;
extern queue_t *vtunl_transport_q;
extern int vtunldebug;
extern int vtunldecapall;
extern u16 vtunl_default_pmtu;
extern int total_exit_count;
extern struct vtunl_inst *head_vip;
extern struct hash_table bcast_table, decap_table, encap_table, texit_table;
extern void hash_tbl_init(struct hash_table *);
extern void hash_tbl_clean(struct hash_table *, int);

extern	int	nulldev();
extern	int	nodev();
extern  int     vtunldebug;
extern	struct mod_ops mod_driverops;
extern	struct streamtab vtunlinfo;

static int vtunl_driver_info(dev_info_t *, ddi_info_cmd_t, void *, void **);
static int vtunl_driver_attach(dev_info_t *,  ddi_attach_cmd_t);
static int vtunl_driver_identify(dev_info_t *);
static int vtunl_driver_detach(dev_info_t *,  ddi_detach_cmd_t);

/* inter-module dependencies */
/* did not find another occurrence in either vtunl.c or this file - VG */
char _depends_on[] = "drv/ip";

#define VTUNL_DEV 1

#ifdef VTUNL_DEV
/*
 * Device Access functions. Contains entry points for the character 
 * and block ops for the device driver. Unsupported entry points
 * are specified as "nodev" (9F) and others that do nothing but simply
 * return success are marked "nulldev"(9F). nodev always returns
 * ENXIO and nulldev always returns 0. nulldev can be called from any 
 * context. nodev only from user context.
 *
 * cb_flag:        D_NEW indicates new style driver, D_MP indicates 
 *                 safe for multithreading, D_MTPERMOD indicates
 *                 module/driver has an inner perimeter that encloses all
 *                 the module's/driver's queues (mt-streams(9F)).
 *
 * cb_str          is set to NULL for non-streams drivers, else it
 *                 points to a streamtab structure.
 *
 * others:         Fields (*cb_read)() (cb_ops(9S)) and (*cb_write)() 
 *                 are not mentioned here and therefore get initialized
 *                 to null values. 
 *
 */
static struct cb_ops vtunl_driver_ops = {
	nulldev,		/* cb_open */
	nulldev,		/* cb_close */
	nodev,			/* cb_strategy */
	nodev,			/* cb_print */
	nodev,			/* cb_dump */
	nodev,			/* cb_read */
	nodev,			/* cb_write */
	nodev,			/* cb_ioctl */
	nodev,			/* cb_devmap */
	nodev,			/* cb_mmap */
	nodev,			/* cb_segmap */
	nochpoll,		/* cb_chpoll */
	ddi_prop_op,		/* cb_prop_op */
	&vtunlinfo,		/* cb_stream */
	(int)(D_NEW|D_MP|D_MTPERMOD),		/* cb_flag */
	
        CB_REV                  /* cb_rev */
};

/*
 * NOTE:  If multithreading support in vtunl is increased beyond
 *        the D_MTPERMOD level, we should change the allocation of
 *        buffers used to convert u32 values to internet address
 *        strings.
 */

/*
 * Device Configuration. Refer to dev_ops (9S). Contains common fields
 * and pointers to the bus_ops and/or cb_ops.
 *
 *   devo_rev            Driver  build  version.  Set   this   to
 *                       DEVO_REV.
 *
 *   devo_refcnt         Driver reference count. Set this to 0.
 *
 *   devo_getinfo        Get  device  driver   information   (see
 *                       getinfo(9E)).
 *
 *   devo_identify       Determine if a driver is associated with
 *                       a device (see identify(9E)).
 *
 *   devo_probe          Probe device (see probe(9E)).
 *
 *   devo_attach         Attach   driver   to    dev_info    (see
 *                       attach(9E)).
 *
 *   devo_detach         Detach/prepare  driver  to  unload  (see
 *                       detach(9E)). A nodev here prevents the
 *                       driver from being unloaded.
 *
 *   devo_reset          Reset device.  (Not  supported  in  this
 *                       release.)  Set this to nodev.
 *
 *   devo_cb_ops         Pointer to cb_ops(9S) structure for leaf
 *                       drivers.
 *
 *   devo_bus_ops        Pointer to bus operations structure  for
 *                       nexus drivers.  Set this to NULL if this
 *                       is for a leaf driver.
 */
static struct dev_ops vtunl_mops = {
	DEVO_REV,		/* devo_rev */
	0,			/* devo_refcnt */
	vtunl_driver_info,	/* devo_getinfo */
	vtunl_driver_identify,	/* devo_identify */
	nulldev,		/* devo_probe */
	vtunl_driver_attach,	/* devo_attach */
	vtunl_driver_detach,	/* devo_detach */
	nodev,			/* devo_reset */
	&vtunl_driver_ops,	/* devo_cb_ops */
	(struct bus_ops *)NULL	/* devo_bus_ops */
};

/*
 * Linkage structure for loadable drivers (man -s 9S modldrv).
 * Desrcibes a module in more detail (Writing Device Drivers 
 * Chap. titles Autoconfiguration). 
 */
static struct modldrv modldrv = {
	&mod_driverops,	 /* *drv_modops contains &mod_driverops
                            for a device driver. Type of module */
	"IPv4 tunneling Streams driver", /* This is displayed
						 by modinfo (1M) */
	&vtunl_mops,		/* Points to the dev_ops structure.*/
};
#endif


extern struct mod_ops mod_strmodops;

/* See man -s 9S fmodsw. Contains info for STREAMS modules. */
static struct fmodsw vtunl_fmodsw = {
	"vtunl",                      /* char f_name[FMNAMESZ + 1]; */
	&vtunlinfo,                   /* struct streamtab *f_str */
	(D_NEW | D_MP | D_MTPERMOD)  /* int f_flag */
	};

/* Linkage structure for loadable STREAMS modules, see modlstrmod(9S).*/
static struct modlstrmod modlstrmod = {
        &mod_strmodops,              /* strmod_modops must always 
                                      * point to mod_strmodops
                                      * this identifies the module as
                                      * a loadable STREAMS module */
        "IPv4 tunneling Streams module",     /* char *strmod_linkinfo */
        &vtunl_fmodsw                 /* struct fmodsw *strmod_fmodsw */
	};

/*
 * See man -s 9S modlinkage. Provided by the module writer to the
 * routines which install, remove and retrieve information
 * from a module, e.g. _init (9E), _fini (9E) and _info (9E).
 *
 *   int ml_rev       Revision number of the loadable modules system
 *                    must be MOD_REV_1.
 *
 *   void *ml_linkage[4]  Null terminated array of pointers to linkage
 *                        structures. For driver modules ther is only 
 *                        one.
 *
 */
static struct modlinkage modlinkage = {
        MODREV_1,
        (void *) &modlstrmod,
#ifdef VTUNL_DEV
        (void *) &modldrv,
#endif
        NULL
	};


/*
 * This is where all one-time resource allocations should occur. 
 * Per-instance allocations should be left to the attach routine.
 * Called before any other routine in a loadable module. Any setup
 * performed before mod_install (9F) should be undone if mod_install
 * fails. Structures referred to by modlinkage should not be altered
 * after the call to mod_install().
 */
_init()
{
	head_vip = (struct vtunl_inst *) kmem_zalloc(sizeof(struct vtunl_inst),
						     KM_NOSLEEP);
	if (head_vip == NULL) return(-1);

	head_vip->prev = head_vip;
	head_vip->next = head_vip;
	head_vip->q = NULL;
	head_vip->vcs.type = VTUNL_INST_HEAD;
	head_vip->vcs.state = VTUNL_STATE_INIT;

        timer_q = NULL;
	vtunl_transport_q = NULL;
	vtunldebug = 0;
	vtunldecapall = 0;
	vtunl_default_pmtu = VTUNL_MAXDLSDU;
	total_exit_count = 0;

        hash_tbl_init(&bcast_table);
        hash_tbl_init(&decap_table);
        hash_tbl_init(&encap_table);
        hash_tbl_init(&texit_table);

	return (mod_install(&modlinkage));
}

/*
 * Prepares a loadable module for unloading. Undo work performed
 * in _init (e.g. allocation of global mutexes) here after mod_remove
 * (9F) returns success.
 */
int
_fini()
{
   int val;
   
        if ((val = mod_remove(&modlinkage)) == 0) {
           hash_tbl_clean(&bcast_table, sizeof(struct bcast_target));
           hash_tbl_clean(&decap_table, sizeof(struct decap_target));
           hash_tbl_clean(&encap_table, sizeof(struct encap_target));
           hash_tbl_clean(&texit_table, sizeof(struct t_exit));
	   kmem_free(head_vip, sizeof(struct vtunl_inst));
        }
	return(val);
}

/* Returns information about a loadable module. */
_info(struct modinfo *modinfop)
{
	return (mod_info(&modlinkage, modinfop));
}


/* We use this to save the dev_info node. Typically, the attach
 * function saves the dev_info_t pointer it is passed in a field 
 * of a per-instance state structure.
 */
static	dev_info_t	* vtunl_driver_dev_info;

/* The attach routine allocates per-instance soft state, registers
 * interrupts, maps device registers, initializes per-instance
 * mutexes and conditional variables, also creates minor device
 * node for the instance.
 *
 * It is important to return DDI_FAILURE if (cmd != DDI_ATTACH).
 *
 * A successful return from probe is required before attach is
 * called.
 */

/* ARGSUSED */
static int
vtunl_driver_attach(dev_info_t *devi, ddi_attach_cmd_t cmd)
{
        if (cmd != DDI_ATTACH) return (DDI_FAILURE);
	vtunl_driver_dev_info = devi;
	if (ddi_create_minor_node(devi, "vtunl", S_IFCHR,
	    0, NULL, CLONE_DEV) == DDI_FAILURE) {
		ddi_remove_minor_node(devi, NULL);
		return (-1);
	}
        vtunl3dbg(("vtunl_driver_attach successful\n"));
	return (DDI_SUCCESS);
}

/* Free resources allocated by the attach routine.
 *
 * It is important to return DDI_FAILURE if (cmd != DDI_DETACH). 
 * Make sure no callback functions are pending before returning 
 * DDI_SUCCESS. Cancel callbacks if needed.
 */

/* ARGSUSED */
static int
vtunl_driver_detach(dev_info_t *devi, ddi_detach_cmd_t cmd)
{
        if (cmd != DDI_ATTACH) return (DDI_FAILURE);
        vtunl3dbg(("vtunl_driver_detach successful\n"));
	return (DDI_SUCCESS);
}


/*
 * When infocmd is DDI_INFO_DEVT2DEVINFO, return the _dev_info_t
 * pointer in result. When it is DDI_INFO_DEVT2INSTANCE, return
 * the instance number associated with the dev_t argument. 
 * 
 * Drivers that do not export a cb_ops structure are not required to
 * provide a getinfo entry point.
 *
 */

/* ARGSUSED */
static int
vtunl_driver_info(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg,
	void **result)
{
	register int error;

	switch (infocmd) {
	case DDI_INFO_DEVT2DEVINFO:
		if (vtunl_driver_dev_info == NULL) {
			error = DDI_FAILURE;
		} else {
			*result = (void *) vtunl_driver_dev_info;
			error = DDI_SUCCESS;
		}
		break;
	case DDI_INFO_DEVT2INSTANCE:
		*result = (void *)0;
		error = DDI_SUCCESS;
		break;
	default:
		error = DDI_FAILURE;
	}
        vtunl3dbg(("vtunl_driver_info returning %s\n",
                    ((error == DDI_SUCCESS) ? "success" : "failure")));
	return (error);
}

/*
 * This is only for backward compatibility and according to
 * man -s 9E identify, should not be implemented. 
 */
static int
vtunl_driver_identify(dev_info_t *devi)
{
        vtunl3dbg(("vtunl_driver_identify called\n"));
	if (strcmp((char *)ddi_get_name(devi), "vtunl") == 0)
		return (DDI_IDENTIFIED);
	return (DDI_NOT_IDENTIFIED);
}


