/*

  Sun Public License Notice

 The contents of this file are subject to the Sun Public License Version
 1.0 (the "License"). You may not use this file except in compliance with
 the License. A copy of the License is available at http://www.sun.com/

 The Initial Developer of the Original Code is Asanga D. Udugama
 Portions created by Asanga D. Udugama are Copyright (C) 2003.
 All Rights Reserved.

 Contributor(s): Niko A. Fikouras (niko@comnets.uni-bremen.de)

*/

/*
 * An implementation of Filers for Mobile-IP as described in
 * draft-nomad-mobileip-filters-05.txt.
 * Author: Asanga D. Udugama (adu@comnets.uni-bremen.de)
 * filters.c : Filter management routines used in Sunlabs Mobile IP.
 */

#pragma ident   "@(#)filters.c 1.1 2002/10/20"

#include <stdio.h>
#include "mip.h"
#include "mobile.h"

int appendFilterExt(unsigned char buffer[], int buflen, Filter fs[], int fsCnt,
						FilterOp fo[], int *foCnt, char *ifcname);
int setFilterForHomeBinding(Filter fs[], int fsCnt, FilterOp fo[], int *foCnt, char *ifcname);
int setFilterForNewBinding(Filter fs[], int fsCnt, FilterOp fo[], int *foCnt, char *ifcname, int action, RegReqEntry rrt[]);
int applyReverseFilters(Filter fs[], InterfaceState ist[], char *ifaceName);
int isFilterAppliable(int elem, Filter fs[], InterfaceState ist[], char *ifaceName);
u32 getValidRouterAddr(char *ifaceName, InterfaceState ist[]);
void deleteAllFilters(Filter fs[]);
void showFilterInfo();

extern void cvtShortIntToCharArray(unsigned char *ca, short int si);
extern void cvtu32ToCharArray(char *ca, u32 li);
extern int applyFilter(char *, char *, u32);
extern int flushRouteCache();

/*
 * Append filter extensions to a registration reply using the operations
 * table.  Only filters relating to the given interface are appended.
 * Returns total number of bytes in the extension.
 */
int appendFilterExt(unsigned char buffer[], int buflen, Filter fs[],
		  int fsCnt, FilterOp fo[], int *foCnt, char *ifcname) {
	int bufpos = buflen, i, j, k, elem, seq;
	unsigned char flgs;
	char str1[30];
	unsigned int int1;
	controlExt *controlExtPtr;
	delExt *delExtPtr;
	bafExt *bafExtPtr;
	protoExt *protoExtPtr;
	srcAddrExt *srcAddrExtPtr;
	srcNetExt *srcNetExtPtr;
	srcPortExt *srcPortExtPtr;
	srcPortRangeExt *srcPortRangeExtPtr;

	dstAddrExt *dstAddrExtPtr;
	dstNetExt *dstNetExtPtr;
	dstPortExt *dstPortExtPtr;
	dstPortRangeExt *dstPortRangeExtPtr;

	freeFormExt *freeFormExtPtr;
	BAFExtData  *bafedp;
	ProtoExtData *protoedp;
	SrcAddrExtData *srcaddredp;
	SrcNetExtData *srcnetedp;
	SrcPortExtData *srcportedp;
	SrcPortRangeExtData *srcportrangeedp;

	DstAddrExtData *dstaddredp;
	DstNetExtData *dstnetedp;
	DstPortExtData *dstportedp;
	DstPortRangeExtData *dstportrangeedp;

	FreeFormExtData *freeformedp;



	// filters have not changed
	if(*foCnt <= 0)
		return 0;

	// loop looks at the operations table and appends
	// filter and control extension packets to registration
	// request
	for(i = 0; i < MAX_FILTERS; i++) {

			// find the element number of the filter table
			// that corresponds to the index (fo[i].idx)
			// in operations table; if idx number is incorrect
			// return 0, and do not append the packets
		if(fo[i].idx == (-1)) break;
		elem = -1;
		for(j = 0; j < MAX_FILTERS; j++) {
			if(fo[i].idx == fs[j].idx) {
				elem = j;
				break;
			}
		}

		if (fo[i].action != ACT_DELETE) {
			if(elem < 0)
				return 0;

			seq = -1;
			for(j = 0; j < MAX_FILTERS; j++) {
				if(strcmp(ifcname, fs[elem].controlModules[j].boundIfaceName) == 0) {
					seq = j;
					break;
				}
			}

			if(seq < 0)
				return 0;
		}

			// append filter extension packet setting the operation
			// as specified in the
		switch(fo[i].action) {
			case ACT_REPLACE:
				showFilterInfo();
				if (fs[elem].markValue == -1) {
					fs[elem].markValue = 1;
					for(j = 0; j < fs[elem].filterModuleCount; j++) {
						switch(fs[elem].filterModules[j].filterType) {
							case FLT_BAF_EXT:
								bafExtPtr = (bafExt *) (buffer + bufpos);
								bafExtPtr->type = FLT_BAF_EXT;
								bafExtPtr->length = 1;
								bafedp = (BAFExtData *) fs[elem].filterModules[j].filterPtr;
								bafExtPtr->DSCP = (unsigned char) (bafedp->DSCP << 2);
								bufpos = bufpos + (2 + 1);
								break;

							case FLT_PROTO_EXT:
								protoExtPtr = (protoExt *) (buffer + bufpos);
								protoExtPtr->type = FLT_PROTO_EXT;
								protoExtPtr->length = 1;
								protoedp = (ProtoExtData *) fs[elem].filterModules[j].filterPtr;
								protoExtPtr->protoNum = (unsigned char) protoedp->protoNum;
								bufpos = bufpos + (2 + 1);
								break;

							case FLT_SRCADR_EXT:
								srcAddrExtPtr = (srcAddrExt *) (buffer + bufpos);
								srcAddrExtPtr->type = FLT_SRCADR_EXT;
								srcAddrExtPtr->length = 4;
								srcaddredp = (SrcAddrExtData *) fs[elem].filterModules[j].filterPtr;
								cvtu32ToCharArray(srcAddrExtPtr->srcAddr, srcaddredp->srcAddr);
								bufpos = bufpos + (2 + 4);
								break;

							case FLT_SRCNET_EXT:
								srcNetExtPtr = (srcNetExt *) (buffer + bufpos);
								srcNetExtPtr->type = FLT_SRCNET_EXT;
								srcNetExtPtr->length = 4 * 2;
								srcnetedp = (SrcNetExtData *)
										fs[elem].filterModules[j].filterPtr;

								cvtu32ToCharArray(srcNetExtPtr->srcNet.addr,
										srcnetedp->srcAddr);
								cvtu32ToCharArray(srcNetExtPtr->srcNet.netMask,
										srcnetedp->srcNetMask);
								bufpos = bufpos + (2 + (4 * 2));
								break;

							case FLT_SRCPORT_EXT:
								srcPortExtPtr = (srcPortExt *) (buffer + bufpos);
								srcPortExtPtr->type = FLT_SRCPORT_EXT;
								srcPortExtPtr->length = 2;
								srcportedp = (SrcPortExtData *)
										fs[elem].filterModules[j].filterPtr;

								cvtShortIntToCharArray(srcPortExtPtr->portNum,
											srcportedp->portNum);
								bufpos = bufpos + (2 + 2);
								break;

							case FLT_SRCPORTRNG_EXT:
								srcPortRangeExtPtr = (srcPortRangeExt *) (buffer + bufpos);
								srcPortRangeExtPtr->type = FLT_SRCPORTRNG_EXT;
								srcPortRangeExtPtr->length = 2 * 2;
								srcportrangeedp = (SrcPortRangeExtData *)
										fs[elem].filterModules[j].filterPtr;

								cvtShortIntToCharArray(srcPortRangeExtPtr->portRange.minPortNum,
											srcportrangeedp->minPortNum);
								cvtShortIntToCharArray(srcPortRangeExtPtr->portRange.maxPortNum,
											srcportrangeedp->maxPortNum);
								bufpos = bufpos + (2 + (2 * 2));
								break;

							case FLT_FRFORM_EXT:
								freeFormExtPtr = (freeFormExt *) (buffer + bufpos);
								freeFormExtPtr->type = FLT_FRFORM_EXT;
								freeformedp = (FreeFormExtData *)
										fs[elem].filterModules[j].filterPtr;
								freeFormExtPtr->length = 2 + ((freeformedp->length / 2) * 2);
								cvtShortIntToCharArray(freeFormExtPtr->offset, freeformedp->offset);

								for(k = 0; k < (freeformedp->length / 2); k++) {
									strcpy(str1, " ");
									strncat(str1,&(freeformedp->value[k * 2]), 2);
									int1 = 0;
									sscanf(str1, "%02x", &int1);
									freeFormExtPtr->valueNmask[k] = (char) int1;
								}

								for(k = 0; k < (freeformedp->length / 2); k++) {
									strcpy(str1, " ");
									strncat(str1,&(freeformedp->mask[k * 2]), 2);
									int1 = 0;
									sscanf(str1, "%02x", &int1);
									freeFormExtPtr->valueNmask[(freeformedp->length / 2) + k]
												= (char) int1;
								}
								bufpos = bufpos + (2 + 2 + freeformedp->length);
								break;


							case FLT_DSTPORT_EXT:
								dstPortExtPtr = (dstPortExt *) (buffer + bufpos);
								dstPortExtPtr->type = FLT_DSTPORT_EXT;
								dstPortExtPtr->length = 2;
								dstportedp = (DstPortExtData *)
										fs[elem].filterModules[j].filterPtr;

								cvtShortIntToCharArray(dstPortExtPtr->portNum,
											dstportedp->portNum);
								bufpos = bufpos + (2 + 2);
								break;

							case FLT_DSTPORTRNG_EXT:
								dstPortRangeExtPtr = (dstPortRangeExt *) (buffer + bufpos);
								dstPortRangeExtPtr->type = FLT_DSTPORTRNG_EXT;
								dstPortRangeExtPtr->length = 2 * 2;
								dstportrangeedp = (DstPortRangeExtData *)
										fs[elem].filterModules[j].filterPtr;

								cvtShortIntToCharArray(dstPortRangeExtPtr->portRange.minPortNum,
											dstportrangeedp->minPortNum);
								cvtShortIntToCharArray(dstPortRangeExtPtr->portRange.maxPortNum,
											dstportrangeedp->maxPortNum);
								bufpos = bufpos + (2 + (2 * 2));
								break;

							default:
								break;
						}
					}
				}
				// append control extension packet filters
				controlExtPtr = (controlExt *) (buffer + bufpos);
				controlExtPtr->type = FLT_CTL_EXT;
				controlExtPtr->length = 2;
				controlExtPtr->weight = fs[elem].controlModules[seq].weight;
				controlExtPtr->idx = (unsigned char) fo[i].idx;
				bufpos = bufpos + sizeof(controlExt);
				break;

			case ACT_DELETE:
				// append control extension packet filters
				delExtPtr = (delExt *) (buffer + bufpos);
				delExtPtr->type = FLT_DEL_EXT;
				delExtPtr->length = 1;
				delExtPtr->idx[0] = (unsigned char) fo[i].idx;
				bufpos = bufpos + sizeof(delExt) - 9;
				break;

			case ACT_FLUSH:
				// flush all filters
				controlExtPtr = (controlExt *) (buffer + bufpos);
				controlExtPtr->type = FLT_DEL_EXT;
				controlExtPtr->length = 1;
				controlExtPtr->idx = 0;
				bufpos = bufpos + sizeof(controlExt);
				break;
			default:
				break;
		}

		// once filter is attached, the corresponding
		// entry in the operations table is removed
		// and the count is reduced
		(*foCnt)--;

		fo[i].action = (-1);
		fo[i].idx = (-1);
	}

	return (bufpos - buflen);
}

/*
 * Sets the operations table to resend the filters when the MN
 * goes to Colocated CoA mode or FA CoA Mode
 */
int setFilterForNewBinding(Filter fs[], int fsCnt, FilterOp fo[], int *foCnt, char *ifcname, int action,
		RegReqEntry rrt[]) {

	int i, j, k, l, m;

	for(i = 0; i < MAX_FILTERS; i++) {
		// if the interface is the same; i.e. the interface for
		// which the hand-off is occuring
		for(j = 0; j < MAX_FILTER_MODULES_PER_FILTER; j++) {
			if (fs[i].controlModules[j].weight == -1) continue;
			printf("checking for idx: %d\n", i);
			printf("iface: %s candidate iface %s\n", ifcname, fs[i].controlModules[j].boundIfaceName);
			if(strcmp(ifcname, fs[i].controlModules[j].boundIfaceName) == 0) {

				// make sure that the operations table has
				// sufficient free slots to add the delete
				// and append filter instructions
				if(((*foCnt) + 2) >= MAX_FILTERS)
					return (-1);

				// find a blank slot in the operations
				// table to insert the delete
				for(k = 0; k < MAX_FILTERS; k++) {
					if(fo[k].action < 0)
						break;
				}
				if(k >= MAX_FILTERS)
					return (-1);

				fo[k].action = action;
				fo[k].idx = fs[i].idx;
				(*foCnt)++;
				if (action == ACT_REPLACE) {
					printf("\nadd idx: %d for iface %s\n", i, ifcname);
				}
				if (action == ACT_DELETE) {
					fs[i].markValue = (-1);
					printf("\ndel idx: %d for iface %s\n", i, ifcname);
					for(l = 0; l < MAX_FILTER_MODULES_PER_FILTER; l++) {
						if (fs[i].controlModules[l].weight == -1) continue;
						if(strcmp(ifcname, fs[i].controlModules[l].boundIfaceName) 
							!= 0) {
							for (m = 0; m < DEFAULT_MAX_REG_REQ; m++) {
								if((strcmp(ifcname, rrt[m].mnRegIfaceName) 
								!= 0) && (rrt[m].mnRegStatus == REG_ACCEPTED_RENEWABLE)) {
									setFilterForNewBinding(fs, fsCnt, fo, foCnt, rrt[m].mnRegIfaceName, ACT_REPLACE, rrt);
									rrt[m].mnRegNextTime = 0;
									break;
								}
							}
						}
					}
				}
			}
		}
	}
	if (action == ACT_DELETE) {
		for (i = 0; i < DEFAULT_MAX_REG_REQ; i++) {
			if((strcmp(ifcname, rrt[i].mnRegIfaceName) != 0) && 
				(rrt[i].mnRegStatus == REG_ACCEPTED_RENEWABLE)) {
					rrt[i].mnRegNextTime = 0;
					break;

			}
		}
		showFilterInfo();
	}
	return 0;
}


/*
 * Sets the operations table to flush all the filters when the MN
 * goes to Home state
 */
int setFilterForHomeBinding(Filter fs[], int fsCnt, FilterOp fo[], int *foCnt, char *ifcname) {
	int i;

	for(i = 0; i < MAX_FILTERS; i++) {
		fo[i].action = (-1);
		fo[i].idx = -1;

	}

	*foCnt = 1;
	fo[0].action = ACT_FLUSH;
	fo[0].idx = (-1);

	return 0;
}

/* This function applies the the reverse filters on the MN.  Before
 * applying, it delets any existing filters
 */
int applyReverseFilters(Filter fs[], InterfaceState ist[], char *ifaceName) {
	int i, j;
	char str[128], *sptr, addr1[16], addr2[16];
	BAFExtData  *bafedp;
	ProtoExtData *protoedp;
	SrcAddrExtData *srcaddredp;
	SrcNetExtData *srcnetedp;
	SrcPortExtData *srcportedp;
	SrcPortRangeExtData *srcportrangeedp;
	FreeFormExtData *freeformedp;

	DstAddrExtData *dstaddredp;
	DstNetExtData *dstnetedp;
	DstPortExtData *dstportedp;
	DstPortRangeExtData *dstportrangeedp;

	u32 defaultFtlCoA;

	// remove all the currently applied reverse filters for the given interface
	for(i = 0; i < MAX_FILTERS; i++) {
		for(j = 0; j < MAX_FILTERS; j++) {
			if(strcmp(fs[i].controlModules[j].boundIfaceName, ifaceName) == 0) {
				deleteFilters((i + 1), ifaceName);
				break;
			}
		}
	}

	// create reverse filter rules, one at a time, goes in the reverse order
	// as rule setting assigns filter priorities backwards
	for(i = (MAX_FILTERS - 1); i >= 0; i--) {

		// is the filter appliable ?
		if(isFilterAppliable(i, fs, ist, ifaceName)) {

			// if so, build a string containing each filter module,
			// an example format is,
			// "  2  4  70 134.102.158.1 134.102.158.0 "
			sprintf(str, " %d %d %d ", fs[i].idx, fs[i].filterModuleCount, (i + 1));

			for(j = 0; j < fs[i].filterModuleCount; j++) {
				sptr = str + strlen(str);
				switch(fs[i].filterModules[j].filterType) {
					case FLT_BAF_EXT:
						bafedp = (BAFExtData *)
								fs[i].filterModules[j].filterPtr;
						sprintf(sptr, " %d %d ", FLT_BAF_EXT, bafedp->DSCP);
						break;
					case FLT_PROTO_EXT:
						protoedp = (ProtoExtData *)
								fs[i].filterModules[j].filterPtr;
						sprintf(sptr, " %d %d ", FLT_PROTO_EXT, protoedp->protoNum);
						break;
					case FLT_SRCADR_EXT:
						srcaddredp = (SrcAddrExtData *)
								fs[i].filterModules[j].filterPtr;
						sprintf(sptr, " %d %s ", FLT_SRCADR_EXT, ntoa(srcaddredp->srcAddr, addr1));
						break;
					case FLT_SRCNET_EXT:
						srcnetedp = (SrcNetExtData *)
								fs[i].filterModules[j].filterPtr;
						sprintf(sptr, " %d %s %s ", FLT_SRCNET_EXT,
										ntoa(srcnetedp->srcAddr, addr1),
										ntoa(srcnetedp->srcNetMask, addr2));
						break;
					case FLT_SRCPORT_EXT:
						srcportedp = (SrcPortExtData *)
								fs[i].filterModules[j].filterPtr;
						sprintf(sptr, " %d %d ", FLT_SRCPORT_EXT, srcportedp->portNum);
						break;
					case FLT_SRCPORTRNG_EXT:
						srcportrangeedp = (SrcPortRangeExtData *)
								fs[i].filterModules[j].filterPtr;
						sprintf(sptr, " %d %d %d ", FLT_SRCPORTRNG_EXT,
										srcportrangeedp->minPortNum,
										srcportrangeedp->maxPortNum);
						break;
					case FLT_FRFORM_EXT:
						printf("Non-implemented filter in table \n");
						// not implemented
						break;

					// new filters

					case FLT_DSTPORT_EXT:
						dstportedp = (DstPortExtData *)
								fs[i].filterModules[j].filterPtr;
						sprintf(sptr, " %d %d ", FLT_DSTPORT_EXT, dstportedp->portNum);
						break;
					case FLT_DSTPORTRNG_EXT:
						dstportrangeedp = (DstPortRangeExtData *)
								fs[i].filterModules[j].filterPtr;
						sprintf(sptr, " %d %d %d ", FLT_DSTPORTRNG_EXT,
										dstportrangeedp->minPortNum,
										dstportrangeedp->maxPortNum);
						break;

				}
			}

			// apply the filter as given in string
			applyFilter(ifaceName, str, getValidRouterAddr(ifaceName, ist));
		}
	}

	// flush routing cache to effect the new routing entries
	flushRouteCache();

	return TRUE;
}

/* This function checks whether a filter is appliable or not. A
 * filter is appliable if, the filter belongs to an interface
 * which is up and connected as foreign or colocated
 */
int isFilterAppliable(int elem, Filter fs[], InterfaceState ist[], char *ifaceName) {
	int i, j;

	// if no filter on this slot, return FALSE
	if (fs[elem].idx == -1 || strcmp(fs[elem].controlModules[0].boundIfaceName, ifaceName) != 0)
		return FALSE;

	// the filter termed as the default filter is
	// not set
	if(fs[elem].idx == DEFAULT_FILTER_SEQ)
		return FALSE;

	for(i = 0; i < DEFAULT_MAX_INTERFACES; i++) {
		if(*(ist[i].ifaceName) == (char) NULL)
			continue;

		for(j = 0; j < DEFAULT_MAX_INTERFACES; j++) {
			if(strcmp(fs[elem].controlModules[j].boundIfaceName, ist[i].ifaceName) != 0)
				continue;

			// if the interface is in foreign or colocated
			// this filter is applicable
			if((ist[i].currentState.state == FOREIGN_STATE)
				|| (ist[i].currentState.state == COLOCATED_STATE))
				return TRUE;
		}
	}

	return FALSE;
}

u32 getValidRouterAddr(char *ifaceName, InterfaceState ist[]) {
	int i;

	for(i = 0; i < DEFAULT_MAX_INTERFACES; i++) {
		if(*(ist[i].ifaceName) == (char) NULL)
			continue;

		if(strcmp(ifaceName, ist[i].ifaceName) != 0)
			continue;

		return (ist[i].currentState.routerAddr);

	}

	return (inet_addr("0.0.0.0"));
}

/**
 * Delete all filters applied using applyFilters function
*/
void deleteAllFilters(Filter fs[]) {
	int i;

	for(i = 0; i < MAX_FILTERS; i++) {
		if(*(fs[i].controlModules[0].boundIfaceName) != '\0')
			deleteFilters((i + 1), fs[i].controlModules[0].boundIfaceName);
	}

}
