/*

  Sun Public License Notice

 The contents of this file are subject to the Sun Public License Version
 1.0 (the "License"). You may not use this file except in compliance with
 the License. A copy of the License is available at http://www.sun.com/

 The Initial Developer of the Original Code is Asanga D. Udugama
 Portions created by Asanga D. Udugama are Copyright (C) 2003.
 All Rights Reserved.

 Contributor(s): none

*/

/*
 * An implementation of Filers for Mobile-IP as described in
 * draft-nomad-mobileip-filters-05.txt.
 * Author: Asanga Udugama (adu@comnets.uni-bremen.de)
 * info.c : Information passing routines used in Mobile IP.
 */

#pragma ident   "@(#)info.c 1.1 2002/10/20"

#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pthread.h>

#include "mip.h"
#include "mobile.h"


#define MAX_GUI_CONNECTIONS	20
#define DEFAULT_GUI_PORT	26000

// all the information variables that will
// be transmitted
extern int mipdebug;
extern int periodicInterval;
extern int agentSolicitThreshold;
extern int agentExpireThreshold;
extern int ifacePollThreshold;
extern int regLifetime;
extern int regRoundTrip;
extern int regFlags;
extern int haCooloffPeriod;
extern int faCooloffPeriod;
extern int IDfreshnessSlack;
extern int solicitOnBcast;
extern int retransmissionPolicy[];
extern float renewalPolicy[];
extern char ignoreIfaceList[][MAX_IFNAME_LEN];
extern int ignoreIfaceCnt;
extern int FilterStatus;
extern int movementDetection;
extern int HApriority;
extern int EagerSoliciting;
extern int LazySoliciting;
extern u32 mnHomeAddr;
extern u32 mnHomeNetmask;
extern u32 mnHomeBcastAddr;
extern u32 mnSolicitAddr;
extern InterfaceState ISTable[DEFAULT_MAX_INTERFACES];
extern int ISTableSize;
extern IfaceEntry ifaceTable[];
extern int ifaceTableSize;
extern HomeAgentEntry haTable[];
extern int haTableSize;
extern u32 homeRouterTable[];
extern int homeRouterTableSize;
extern LocalMaEntry localMaTable[];
extern int localMaTableSize;
extern LocalMaEntry *knownLocalMaTable[];
extern RegReqEntry regReqTable[];
extern int regReqTableSize;
extern u32 mnRegRequestsAcceptedCnt;
extern int  ICMPsolicitSock;
extern int  UDPregSock;
extern DomainInfo domainInfo;
extern Filter mnFilters[MAX_FILTERS];
extern int mnFilterCount;
extern FilterOp mnFilterOpTable[MAX_FILTERS];
extern int mnFilterOpCount;

// holds info related to each connected GUI
typedef struct connectedgui {
	u32 ipAddr;
	int port;
	long dataSent;
} ConnectedGUI;
ConnectedGUI connGUITable[MAX_GUI_CONNECTIONS];
int connGUICount = 0;

int GUIRecvActive = 0;
// socket variables
int GUIRecvPortNum = DEFAULT_GUI_PORT;
struct sockaddr_in GUIRecvIPAddr;
int GUIRecvSocket;
int GUISendSocket;

// thread variables
static pthread_t monitorGUIs;
static pthread_mutex_t connGUITableUpdateLock;
static pthread_attr_t monitorGUIsAttr;
int monitorGUIsDummy = 1;

// function prototypes
int initInfoModule();
void *monitorGUIConnections(void *ignoreThis);
int sendInfoToGUI(ConnectedGUI *cgui);
int sendInfoToAllGUIs();
char *buildSystemInfo(char *buffer);
char *buildIfcStateInfo(char *buffer);
char *buildFilterInfo(char *buffer);
int termInfoModule();

/**
* Creates the sockets and threads to send information to
* GUIs
*/
int initInfoModule() {
	int i;

	// if info module not activated
	if(!GUIRecvActive)
		return;

	for(i = 0; i < MAX_GUI_CONNECTIONS; i++) {
		connGUITable[i].ipAddr = 0;
		connGUITable[i].port = 0;
		connGUITable[i].dataSent = 0;
	}

	GUIRecvSocket = socket(PF_INET, SOCK_DGRAM, 0);
	bzero(&GUIRecvIPAddr, sizeof(GUIRecvIPAddr));
	GUIRecvIPAddr.sin_family = AF_INET;
	GUIRecvIPAddr.sin_port = htons(GUIRecvPortNum);
	GUIRecvIPAddr.sin_addr.s_addr = INADDR_ANY;
	if ( bind(GUIRecvSocket, (struct sockaddr*)&GUIRecvIPAddr, sizeof(GUIRecvIPAddr)) != 0 ) {
        	mip0dbg(("Error: Could not create GUI listen socket in initInfoModule().\n"));
		return (-1);
	}

	GUISendSocket = socket(PF_INET, SOCK_DGRAM, 0);

	pthread_mutex_init(&connGUITableUpdateLock, NULL);
	pthread_attr_init(&monitorGUIsAttr);
	pthread_attr_setdetachstate(&monitorGUIsAttr, PTHREAD_CREATE_DETACHED);
	if ((pthread_create(&monitorGUIs, &monitorGUIsAttr,
			monitorGUIConnections, (void *) &monitorGUIsDummy) != 0)) {
		mip1dbg(("Error while creating monitorGUIs thread.\n"));
		return (-1);
	}

	return(0);
}

/**
* Thread function that waits for connections from GUIs
*/
void *monitorGUIConnections(void *ignoreThis) {
	int i = (-1), len;
	char buffer[64], chIPAddr[32], chPort[16], dummy[24];
	struct sockaddr_in addr;

	while(1) {
		len = recv(GUIRecvSocket, buffer, sizeof(buffer), 0);

		// if not GET, forget it
		if(strncmp(buffer, "GET", 3) != 0)
			continue;

		// if tooooooo many connections, forget it too
		if((i + 1) >= MAX_GUI_CONNECTIONS)
			continue;

		buffer[len] = '\0';
		if((sscanf(buffer, "%s %s %s", dummy, chIPAddr, chPort)) != 3)
			continue;

		// lock your self in a room, and do as you please
		pthread_mutex_lock(&connGUITableUpdateLock);

		i++;
		connGUICount++;

		connGUITable[i].ipAddr = inet_addr(chIPAddr);
		connGUITable[i].port = atoi(chPort);

		sendInfoToGUI(&connGUITable[i]);

		pthread_mutex_unlock(&connGUITableUpdateLock);
	}
}

/**
* Function to send info to a specific GUI
*/
int sendInfoToGUI(ConnectedGUI *cgui) {
	char buffer[2048];
	struct sockaddr_in addr;

	// if info module not activated
	if(!GUIRecvActive)
		return;

	if(cgui == NULL || cgui->ipAddr == 0)
		return (-1);

	bzero(&addr, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(cgui->port);
	addr.sin_addr.s_addr = cgui->ipAddr;

	// build system info & send
	buildSystemInfo(buffer);
	sendto(GUISendSocket, buffer, strlen(buffer), 0, (struct sockaddr *) &addr, sizeof(addr));

	// build interface state info
	buildIfcStateInfo(buffer);
	sendto(GUISendSocket, buffer, strlen(buffer), 0, (struct sockaddr *) &addr, sizeof(addr));

	// build filter info
	buildFilterInfo(buffer);
	sendto(GUISendSocket, buffer, strlen(buffer), 0, (struct sockaddr *) &addr, sizeof(addr));

	return 0;
}

/**
* Function to send all the information to all connected
* GUIs
*/
int sendInfoToAllGUIs() {
	int i;

	// if info module not activated
	if(!GUIRecvActive)
		return;

	// lock & send info
	pthread_mutex_lock(&connGUITableUpdateLock);

	for(i = 0; i < connGUICount; i++) {
		if(connGUITable[i].ipAddr == 0 || connGUITable[i].port == 0)
			continue;

		sendInfoToGUI(&connGUITable[i]);
	}

	pthread_mutex_unlock(&connGUITableUpdateLock);
}

/**
* This function fills the system information in to the sent
* buffer
*/
char *buildSystemInfo(char *buffer) {
	char tbuffer[256], chAddr1[24], chAddr2[24], chAddr3[24], chAddr4[24];
	int i;

	sprintf(buffer, "Sysinfo@periodicInterval %d;", periodicInterval);
	sprintf(buffer + strlen(buffer), "agentSolicitThreshold %d;", agentSolicitThreshold);
	sprintf(buffer + strlen(buffer), "agentExpireThreshold %d;", agentExpireThreshold);
	sprintf(buffer + strlen(buffer), "ifacePollThreshold %d;", ifacePollThreshold);
	sprintf(buffer + strlen(buffer), "regLifetime %d;", regLifetime);
	sprintf(buffer + strlen(buffer), "regRoundTrip %d;", regRoundTrip);
	sprintf(buffer + strlen(buffer), "regFlags 0x%02x;", (unsigned char) (regFlags & 0xff));
	sprintf(buffer + strlen(buffer), "haCooloffPeriod %d;", haCooloffPeriod);
	sprintf(buffer + strlen(buffer), "faCooloffPeriod %d;", faCooloffPeriod);
	sprintf(buffer + strlen(buffer), "IDfreshnessSlack %d;", IDfreshnessSlack);


	if(ignoreIfaceCnt <= 0) {
		strcpy(tbuffer, " null");
	} else {
		*tbuffer = '\0';
		for(i = 0; i < ignoreIfaceCnt; i++) {
			strcat(tbuffer, " ");
			strcat(tbuffer, ignoreIfaceList[i]);
		}
	}

	sprintf(buffer + strlen(buffer), "ignoreIfaceList%s;", tbuffer);
	sprintf(buffer + strlen(buffer), "SolicitOnBcast %d;", solicitOnBcast);
	sprintf(buffer + strlen(buffer), "movementDetection %d;", movementDetection);
	sprintf(buffer + strlen(buffer), "HApriority %d;", HApriority);
	sprintf(buffer + strlen(buffer), "EagerSoliciting %d;", EagerSoliciting);
	sprintf(buffer + strlen(buffer), "LazySoliciting %d;", LazySoliciting);

	*tbuffer = '\0';
	for(i = 0; i < DEFAULT_MAX_ATTEMPTS; i++) {
		strcat(tbuffer, " ");
		sprintf(tbuffer + strlen(tbuffer), "%d", retransmissionPolicy[i]);
	}
	sprintf(buffer + strlen(buffer), "retransmissionPolicy%s;", tbuffer);

	*tbuffer = '\0';
	for(i = 0; i < DEFAULT_MAX_ATTEMPTS; i++) {
		strcat(tbuffer, " ");
		sprintf(tbuffer + strlen(tbuffer), "%3.02f", renewalPolicy[i]);
	}

	sprintf(buffer + strlen(buffer), "renewalPolicy%s;", tbuffer);
	sprintf(buffer + strlen(buffer), "mnHomeAddr %s;", ntoa(mnHomeAddr, chAddr1));
	sprintf(buffer + strlen(buffer), "mnHomeNetmask %s;", ntoa(mnHomeNetmask, chAddr2));
	sprintf(buffer + strlen(buffer), "mnHomeBcastAddr %s;", ntoa(mnHomeBcastAddr, chAddr3));
	sprintf(buffer + strlen(buffer), "mnSolicitAddr %s;",ntoa(mnSolicitAddr, chAddr4));

	if(homeRouterTableSize <= 0) {
		strcpy(tbuffer, " Null");
	} else {
		*tbuffer = '\0';
		for(i = 0; i < homeRouterTableSize; i++) {
			strcat(tbuffer, " ");
			sprintf(tbuffer + strlen(tbuffer), "%s", ntoa(homeRouterTable[i], chAddr1));
		}
	}
	sprintf(buffer + strlen(buffer), "homeRouterTable%s",tbuffer);

	return buffer;
}

/**
* Function to fill the interface state information to the
* sent buffer
*/
char *buildIfcStateInfo(char *buffer) {
	char tbuffer[256], chAddr1[24], chAddr2[24];
	int i, found;

	strcpy(buffer, "InterfaceState@");

	found = FALSE;
	for(i = 1; i < DEFAULT_MAX_INTERFACES; i++) {
		if(*(ISTable[i].ifaceName) == '\0')
			continue;

		found = TRUE;
		sprintf(buffer + strlen(buffer), "%s;%s;%d;%d;%s;%d;%d:",
				ISTable[i].ifaceName, ntoa(ISTable[i].COAddr, chAddr1),
				ISTable[i].currentState.state, ISTable[i].currentState.backoffIdx,
				ntoa(ISTable[i].currentState.routerAddr, chAddr2),
				ISTable[i].currentState.timeSpent,
				ISTable[i].currentState.nextSolicit);
	}

	if(!found) {
		sprintf(buffer + strlen(buffer), "%s;%s;%d;%d;%s;%d;%d:",
				ISTable[0].ifaceName, ntoa(ISTable[0].COAddr, chAddr1),
				ISTable[0].currentState.state, ISTable[0].currentState.backoffIdx,
				ntoa(ISTable[0].currentState.routerAddr, chAddr2),
				ISTable[0].currentState.timeSpent,
				ISTable[0].currentState.nextSolicit);
	}
	return buffer;
}

/**
* Function to fill the filter information to the
* sent buffer
*/
char *buildFilterInfo(char *buffer) {
	char tbuffer[256], chAddr1[24];
	int i, found, j, k, pick;
	BAFExtData  *bafedp;
	ProtoExtData *protoedp;
	SrcAddrExtData *srcaddredp;
	SrcNetExtData *srcnetedp;
	SrcPortExtData *srcportedp;
	SrcPortRangeExtData *srcportrangeedp;
	DstPortExtData *dstportedp;
	DstPortRangeExtData *dstportrangeedp;
	FreeFormExtData *freeformedp;
	InterfaceState *ifState;
	int boundIfcCount;
	u32 boundCoA;


	sprintf(buffer, "Filter@");

	found = FALSE;
	for(i = 0; i < MAX_FILTERS; i++) {
		if(mnFilters[i].filterModuleCount == 0)
			continue;

		// a filter is shown only if the interface to which it
		// belongs, is active and acquired a CoA (i.e. either FOREIGN or COLOCATED)
		pick = FALSE;
		for(j = 0; j < DEFAULT_MAX_INTERFACES; j++) {
			for(k = 0; k < DEFAULT_MAX_INTERFACES; k++) {
				if((strcmp(ISTable[j].ifaceName, mnFilters[i].controlModules[k].boundIfaceName) == 0)
					&& (ISTable[j].currentState.state == FOREIGN_STATE
						|| ISTable[j].currentState.state == COLOCATED_STATE)) {
					ifState = &ISTable[j];
					pick = TRUE;
					break;
				}
			}

		}
		if(!pick)
			continue;

		found = TRUE;
		sprintf(buffer + strlen(buffer), "%d;", mnFilters[i].filterModuleCount);

		// build filter information
		for(j = 0; j < mnFilters[i].filterModuleCount;j ++) {
			sprintf(buffer + strlen(buffer), "%d;",
					mnFilters[i].filterModules[j].filterType);
			switch(mnFilters[i].filterModules[j].filterType) {
				case FLT_BAF_EXT:
					bafedp = (BAFExtData *) mnFilters[i].filterModules[j].filterPtr;
					sprintf(tbuffer, "%d", bafedp->DSCP);
					break;
				case FLT_PROTO_EXT:
					protoedp = (ProtoExtData *) mnFilters[i].filterModules[j].filterPtr;
					sprintf(tbuffer, "%d", protoedp->protoNum);
					break;
				case FLT_SRCADR_EXT:
					srcaddredp = (SrcAddrExtData *) mnFilters[i].filterModules[j].filterPtr;
					sprintf(tbuffer, "%s", ntoa(srcaddredp->srcAddr, chAddr1));
					break;
				case FLT_SRCNET_EXT:
					srcnetedp = (SrcNetExtData *) mnFilters[i].filterModules[j].filterPtr;
					sprintf(tbuffer, "%s/%d", ntoa(srcnetedp->srcAddr, chAddr1),
								numBitsInMask(srcnetedp->srcNetMask));
					break;
				case FLT_SRCPORT_EXT:
					srcportedp = (SrcPortExtData *) mnFilters[i].filterModules[j].filterPtr;
					sprintf(tbuffer, "%d", srcportedp->portNum);
					break;
				case FLT_SRCPORTRNG_EXT:
					srcportrangeedp = (SrcPortRangeExtData *) mnFilters[i].filterModules[j].filterPtr;
					sprintf(tbuffer, "%d-%d", srcportrangeedp->minPortNum,
								srcportrangeedp->maxPortNum);
					break;
				case FLT_DSTPORT_EXT:
					dstportedp = (DstPortExtData *) mnFilters[i].filterModules[j].filterPtr;
					sprintf(tbuffer, "%d", dstportedp->portNum);
					break;
				case FLT_DSTPORTRNG_EXT:
					dstportrangeedp = (DstPortRangeExtData *) mnFilters[i].filterModules[j].filterPtr;
					sprintf(tbuffer, "%d-%d", dstportrangeedp->minPortNum,
								dstportrangeedp->maxPortNum);
					break;
				case FLT_FRFORM_EXT:
				default:
					sprintf(tbuffer, "not-supported");
					break;
			}
			sprintf(buffer + strlen(buffer), "%s;", tbuffer);
		}

		// build bound COA info
		boundIfcCount = 0;
		for(j = 0; mnFilters[i].controlModules[j].weight != (-1); j++) {
			boundIfcCount++;
		}

		sprintf(buffer + strlen(buffer), "%d;", boundIfcCount);

		for(j = 0; j < boundIfcCount; j++) {

			boundCoA = 0;
			for(k = 0; k < DEFAULT_MAX_INTERFACES; k++) {
				if(strcmp(ISTable[k].ifaceName,
					mnFilters[i].controlModules[j].boundIfaceName) == 0) {
					boundCoA = ISTable[k].currentState.COAddr;
					break;
				}
			}

			sprintf(buffer + strlen(buffer), "%s;%d;%s;",
				ntoa(boundCoA, chAddr1),
				mnFilters[i].controlModules[j].weight,
				mnFilters[i].controlModules[j].boundIfaceName);
		}

		sprintf(buffer + strlen(buffer), "%d:", mnFilters[i].idx);

	}
	if(!found)
		sprintf(buffer + strlen(buffer), "Null");

	return buffer;
}
/**
* Function to terminate info module
*/
int termInfoModule() {

	// if info module not activated
	if(!GUIRecvActive)
		return;

	// terminate thread
	pthread_kill(monitorGUIs, 0);

	// close sockets
	close(GUIRecvSocket);
	close(GUISendSocket);
}
