/*

  Sun Public License Notice

 The contents of this file are subject to the Sun Public License Version
 1.0 (the "License"). You may not use this file except in compliance with
 the License. A copy of the License is available at http://www.sun.com/

 The Original Code is the Sunlabs Mobile IP.
 The Initial Developer of the Original Code is Vipul Gupta.
 Portions created by Vipul Gupta are Copyright (C) 2003.
 All Rights Reserved.

 Contributor(s): Niko A. Fikouras (niko@comnets.uni-bremen.de)
		 Asanga D. Udugama (adu@comnets.uni-bremen.de)

*/

/*
 * An implementation of Mobile-IP as described in RFC 2002.
 * Author : Sunil Madhani (sunil.madhani@eng.sun.com)
 * mnInit.c : Various initialization routines for a mobile node.
 */

#pragma ident	"@(#)mnInit.c 1.3 98/08/19"

#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <signal.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>

#include "mip.h"
#include "mobile.h"

/* Other external declarations */
extern int  timerWentOff;
extern char msg[], currentTime[], relativeTime[];
extern char currentTime[], addr1[], addr2[];

extern int mipdebug;
extern int periodicInterval;
extern int agentSolicitThreshold;
extern int agentExpireThreshold;
extern int ifacePollThreshold;
extern int regLifetime;
extern int regRoundTrip;
extern int regFlags;
extern int haCooloffPeriod;
extern int faCooloffPeriod;
extern int IDfreshnessSlack;
extern int solicitOnBcast;
extern int retransmissionPolicy[];
extern float renewalPolicy[];
extern char ignoreIfaceList[][MAX_IFNAME_LEN];
extern int ignoreIfaceCnt;
extern int FilterStatus;

/* The mn's movement detection method (either LCS or CMS) */
/* modified by Uni. of Sheffield */
extern int movementDetection;

/* Determines whether the MN gives priority to HA's */
/* modified by Uni. of Sheffield */
extern int HApriority;

/* Determines whether the MN will spontaneously solicit when
/* detecting a new wireless segment (Eager Soliciting) and when
/* leaving a wireless segment -- loosing contact (Lazy Soliciting).
/* modified by Uni. of Sheffield */
extern int EagerSoliciting;
extern int LazySoliciting;

/* The mobile node's address, netmask and default routers at home */
extern u32 mnHomeAddr;
extern u32 mnHomeNetmask;
extern u32 mnHomeBcastAddr;
// currnetState is part of ISTable; changed to handle MSB
extern InterfaceState ISTable[DEFAULT_MAX_INTERFACES];
//extern mnState currentState;
extern int ISTableSize;

extern IfaceEntry ifaceTable[];
extern int ifaceTableSize;

extern HomeAgentEntry haTable[];
extern int haTableSize;

/* List of default routers on the home network */
extern u32 homeRouterTable[];
extern int homeRouterTableSize;

/* List of local Mobility Agents (considered active) */
extern LocalMaEntry localMaTable[];
extern int localMaTableSize;

/* List of known local Mobility Agents (considered active) */
/* Used by the ECS movement detection method - modified by */
/* Uni. Sheffield					   */
extern LocalMaEntry *knownLocalMaTable[];

/* List of registration requests (both accepted and pending) */
extern RegReqEntry regReqTable[];
extern int regReqTableSize;
extern u32 mnRegRequestsAcceptedCnt;

extern int  ICMPsolicitSock;
extern int  UDPregSock;

extern DomainInfo domainInfo;

/* to store the filters that have been defined for this mobile node*/
extern Filter mnFilters[MAX_FILTERS]; 				   // data for each filter
extern int mnFilterCount; 					   // the number of filters
extern FilterOp mnFilterOpTable[MAX_FILTERS];			   // the grid that holds the operations
								   // to be done for the filters
extern int mnFilterOpCount;					   // operations count

extern int GUIRecvActive;
extern int GUIRecvPortNum;

extern char *sprintTime(char *, int);
extern void undoState(mnState *);
extern void sendSolicitation(u32);
extern void printProtectedDomainInfo(DomainInfo);
extern void startHomeAction(mnState *, int);
extern int setFilterForHomeBinding(Filter [], int, FilterOp [], int *, char *);

/* OS-specific initialization for Mobile IP */
extern void OSinit();
extern void OScleanup();

/* Swap the xth and yth entry of haTable */
void swapHATableEntry(int x, int y)
{
    HomeAgentEntry tmpHAEntry;

    memcpy(&tmpHAEntry, &(haTable[x]), sizeof(HomeAgentEntry));
    memcpy(&(haTable[x]), &(haTable[y]), sizeof(HomeAgentEntry));
    memcpy(&(haTable[y]), &(tmpHAEntry), sizeof(HomeAgentEntry));
}

/* Get the idxth word from string str
 * if str = "123 245 dis" then getWord(str,2,word) will
 * return 0 and word will cotain "245"
 * If the idxth word is not found then it will return -1
 */
int getWord(char *str, int idx, char *word)
{
    int i;
    int wordIdx = 0;
    int len = strlen(str);
    char str1[30];

    for(i = 0; i < len; i++){
        if (str[i] == ' ' || str[i] == '\t')
	    continue;
        wordIdx++;
        if (sscanf((str+i), "%s", str1) != 1)
	    return(-1);
        if (wordIdx == idx) {
	    strcpy(word, str1);
	    return(0);
        }
        i += strlen(str1);
    }
    return(-1);
}

int getAttributeValue(char *line, int lineNum)
{
    char str1[30], word[30];
    int i, int1, previntvalue;
    float float1, prevfloatvalue;


    if (sscanf(line, "%s", str1) != 1) {
	mip0dbg(("Bad data on line %d.\n", lineNum));
        return(-1);
    }

    /* For some attributes, the value is non-numeric. Process them first.*/
    if (strcasecmp(str1, "ignoreIfaces") == 0) {
        i = 0;
	/*
	 * TODO: We currently do any validity checks on interface names.
	 * Shouldn't be a serious problem.
	 */
	while (((getWord(line, i+2, word) >= 0) &&
		(i < DEFAULT_MAX_INTERFACES))) {
	    strcpy(ignoreIfaceList[i], word);
	    i++;
	}
        ignoreIfaceCnt = i;
	return(0);
    }

    if ((sscanf(line, "%s %f", str1, &float1)) != 2) {
        mip0dbg(("Bad data on line %d.\n", lineNum));
	return(-1);
    }

    int1 = (int)float1;

    if (strcasecmp(str1, "debugLevel") == 0) {
        if (int1 < 0 || int1 > 5) {
	    mip0dbg(("Bad debugLevel %d on line %d.\n",
		    int1, lineNum));
	    return(-1);
	} else {
	    mipdebug = int1;
	}
	return(0);
    }

    if (strcasecmp(str1, "periodicInterval") == 0) {
        if (int1 <= 0) {
	    mip0dbg(("Bad periodicInterval %d on line %d.\n",
		    int1, lineNum));
	    return(-1);
	} else {
	    periodicInterval = int1;
	}
	return(0);
    }

    if (strcasecmp(str1, "agentSolicitThreshold") == 0) {
        if (int1 <= 0) {
	    mip0dbg(("Bad agentSolicitThreshold %d on line %d.\n",
		    int1, lineNum));
	    return(-1);
	} else {
	    agentSolicitThreshold = int1;
	}
	return(0);
    }

    if (strcasecmp(str1, "agentExpireThreshold") == 0) {
        if ((int1 <= 0) || (int1 <= agentSolicitThreshold)) {
	    mip0dbg(("Bad agentExpireThreshold %d on line %d.\n",
		    int1, lineNum));
	    return(-1);
	} else {
	    agentExpireThreshold = int1;
	}
	return(0);
    }

    if (strcasecmp(str1, "ifacePollThreshold") == 0) {
        if ((int1 <= 0) || (int1 < periodicInterval)) {
	    mip0dbg(("Bad ifacePollThreshold %d on line %d.\n",
		    int1, lineNum));
	    return(-1);
	} else {
	    ifacePollThreshold = int1;
	}
	return(0);
    }

    if (strcasecmp(str1, "regLifetime") == 0) {
        if (int1 <= 0){
	    mip0dbg(("Bad regLifetime %d on line %d.\n",
		    int1, lineNum));
	    return(-1);
	} else {
	    regLifetime = int1;
	}
	return(0);
    }

    if (strcasecmp(str1, "regRoundTrip") == 0) {
        if (int1 <= 0) {
	    mip0dbg(("Bad regRoundTrip %d on line %d.\n",
		    int1, lineNum));
	    return(-1);
	} else {
	    regRoundTrip = int1;
	}
	return(0);
    }

    if (strcasecmp(str1, "regFlags") == 0) {
//        if (((char) int1) && 0x23) {
//	    mip0dbg(("Bad regFlags %d on line %d.\n",
//		    int1, lineNum));
//	    return(-1);
//	} else {
	    regFlags = (char) int1;
//	}
	return(0);
    }

    if (strcasecmp(str1, "haCooloffPeriod") == 0) {
        if (int1 < 0) {
	    mip0dbg(("Bad haCooloffPeriod %d on line %d.\n",
		    int1, lineNum));
	    return(-1);
	} else {
	    haCooloffPeriod = int1;
	}
	return(0);
    }

    if (strcasecmp(str1, "faCooloffPeriod") == 0) {
        if (int1 < 0) {
	    mip0dbg(("Bad faCooloffPeriod %d on line %d.\n",
		    int1, lineNum));
	    return(-1);
	} else {
	    faCooloffPeriod = int1;
	}
	return(0);
    }

    if (strcasecmp(str1, "IDfreshnessSlack") == 0) {
        if (int1 < 0) {
	    mip0dbg(("Bad IDfreshnessSlack %d on line %d.\n",
		     int1, lineNum));
	    return(-1);
	} else {
	    IDfreshnessSlack = int1;
	}
	return(0);
    }

    if (strcasecmp(str1, "retransmissionPolicy") == 0) {
	i = 0;
        previntvalue = 0;
	while (((getWord(line, i+2, word) >= 0) && 
		(i < DEFAULT_MAX_ATTEMPTS))) {
            int1 = atoi(word);
	    if ((int1 <= 0) || (int1 < previntvalue)) {
	        mip0dbg(("Improper retransmission value %d on line %d.\n",
			 int1, lineNum));
	        return(-1);
            }
            else {
	        retransmissionPolicy[i] = int1;
		previntvalue = int1;
		i++;
	    }
	}
	retransmissionPolicy[i] = -1;
	return(0);
    }

    if (strcasecmp(str1, "renewalPolicy") == 0) {
	i = 0;
        prevfloatvalue = 200.0000; /* Something large */
	while (((getWord(line, i+2, word) >= 0) && 
		(i < DEFAULT_MAX_ATTEMPTS))) {
            float1 = atof(word);
	    if ((float1 > prevfloatvalue)){
	        mip0dbg(("Improper renewal value %f on line %d.\n",
		    float1, lineNum));
	        return(-1);
            }
            else {
	        renewalPolicy[i] = float1;
		prevfloatvalue = float1; 
		i++;
	    }
	}
	renewalPolicy[i] = 0;
	return(0);
    }

    /* added movement detection method option - modified by Uni. Sheffield */

    if (strcasecmp(str1, "movementDetection") == 0) {
       if (int1 < 0) {
          mip0dbg(("Bad movementDetection %d on line %d.\n", int1, lineNum));
          return(-1);
       }
       else {
            movementDetection = int1;
       }
       return(0);
    }

    if (strcasecmp(str1, "HApriority") == 0) {
       if ((int1 != 0) && (int1 != 1)) {
          mip0dbg(("Bad HApriority %d on line %d.\n", int1, lineNum));
          return(-1);
       }
       else {
            HApriority = int1;
       }
       return(0);
    }

    if (strcasecmp(str1, "EagerSoliciting") == 0) {
       if ((int1 != 0) && (int1 != 1)) {
          mip0dbg(("Bad EagerSoliciting %d on line %d.\n", int1, lineNum));
          return(-1);
       }
       else {
            EagerSoliciting = int1;
       }
       return(0);
    }

    if (strcasecmp(str1, "LazySoliciting") == 0) {
       if ((int1 != 0) && (int1 != 1)) {
          mip0dbg(("Bad LazySoliciting %d on line %d.\n", int1, lineNum));
          return(-1);
       }
       else {
            LazySoliciting = int1;
       }
       return(0);
    }

    if (strcasecmp(str1, "FilterStatus") == 0) {
       if ((int1 != 0) && (int1 != 1)) {
          mip0dbg(("Bad FilterStatus %d on line %d.\n", int1, lineNum));
          return(-1);
       }
       else {
            FilterStatus = int1;
       }
       return(0);
    }

    if (strcasecmp(str1, "GUIRecvActive") == 0) {
       if ((int1 != 0) && (int1 != 1)) {
          mip0dbg(("Bad GUIRecvActive %d on line %d.\n", int1, lineNum));
          return(-1);
       }
       else {
            GUIRecvActive = int1;
       }
       return(0);
    }

    if (strcasecmp(str1, "GUIRecvPortNum") == 0) {
       if (int1 <= 2048) {
          mip0dbg(("Bad GUIRecvPortNum %d on line %d.\n", int1, lineNum));
          return(-1);
       }
       else {
            GUIRecvPortNum = int1;
       }
       return(0);
    }

    return(-1);
}

int getProtectedDomainInfo(FILE *fd, int *lineNum)
{
    char line[256];
    int state = 0;
    char str1[30], str2[30], word[30];
    int int1, i;

    domainInfo.addrIntervalCnt = 0;
    domainInfo.firewallCnt = 0;
    while (fgets(line, 256, fd) != NULL) {
        /* we allow comments starting with # in column 1 */
	(*lineNum)++;
        if ((line[0] == '#') || isBlank(line, strlen(line)))
	    continue;

        switch (state) {
        case 0 : /* Get number of address intervals */
            if ((sscanf(line, "%d", &int1) != 1) ||
	       (int1 < 1) || (int1 > MAX_ADDR_INTERVALS)) {
	        mip0dbg(("Problem reading # of address intervals on line %d.\n",
	                *lineNum));
	        return(-1);
	    } else {
	        domainInfo.addrIntervalCnt = int1;
                i = 0;
	        state = 1;
            }
	    break;

        case 1 : /* Get the address and netmask */
	    if ((sscanf(line, "%s %s", str1, str2) != 2) ||
		(inet_addr(str1) == -1) ||
		(inet_addr(str2) == -1)) {
		mip0dbg(("Improper arguments on line %d.\n", *lineNum));
		return(-1);
	    } else {
		domainInfo.addr[i] = inet_addr(str1);
	        domainInfo.netmask[i] = inet_addr(str2);
		i++;
		if (i == domainInfo.addrIntervalCnt)
		    state = 2;
	    }
	    break;

        /*
	 * We don't need this part if we let a script handle all
         * IPSec related issues such as what IP Security parameters
         * to use, what firewall address to use for the tunnel
         * end point etc.
         */
      
	case 2 : 
            //return(0);
	    //  Get the list of firewalls 
	      i = 0;  
	      while (((getWord(line, i + 1, word) >= 0) &&
	             (i < MAX_FIREWALLS))) {
	         if (inet_addr(word) == -1) {
	             mip0dbg(("Improper firewall address value %s on " \
	                      "line %d.\n", word, *lineNum));
	             return(-1);
	         } else {
	             domainInfo.fwAddr[i] = inet_addr(word);
	             i++;
	         }
	      }
	      domainInfo.firewallCnt = i;
	      return(0);
	}
    }
    return(0);
}

int readConfigInfo(FILE *fd)
{

    int  i;
    int  state = 1;
    int  lineNum = 0;
    char line[256];
    char str1[30], str2[30], str3[30];
    char word[30];
    int  int1, int2;
    long SPI;
    char keyStr[2*MAX_KEY_LEN + 2];
    boolean HAdiscovery = FALSE;
    int discoveryAddrIdx;

    while ((fgets(line, 256, fd) != NULL) && (state != 7) && (state != 8)) {
        /* we allow comments starting with # in column 1 */
	lineNum++;
        if ((line[0] == '#') || isBlank(line, strlen(line)))
	    continue;

        switch (state) {
	case 1:  /* expect to see a version string */
	    if ((sscanf(line, "%s %d", str1, &int1) != 2) ||
		(strcasecmp(str1, "version") != 0) ||
		(int1 != 1)) {
	        mip0dbg(("Bad version in configuration file on line %d.\n",
			lineNum));
	        state = 8;
	    } else {
	        state = 2;
	    }

	    break;
	case 2:  /* expect (optional) attribute-value pairs */
	     if (sscanf(line, "%s", str1) != 1) {
		state = 8;
		break;
	     }
	     if (toupper(str1[0]) < 'A' || (toupper(str1[0]) > 'Z')) {
		state = 3;
             }
	     else {
		 if (getAttributeValue(line, lineNum) < 0) {
		    state = 8;
		 }
		 break;
	     }

	    /* Note that if we are unable to read a string-number(s)
	     * pair, we try to read the next part of the configuration
	     * file by falling through to case 3.
	     */

	case 3:  /* expect home address, home netmask & default routers */
	    /* TODO : Take care of duplicate addresses */
	    if ((sscanf(line, "%s %s %s", str1, str2, str3) != 3) ||
		(inet_addr(str1) == -1) ||
		(inet_addr(str2) == -1)) {
		mip0dbg(("Improper arguments on line %d.\n", lineNum));
		state = 8;
	    } else {
	        mnHomeNetmask = inet_addr(str2);
		mnHomeAddr = inet_addr(str1);
		state = 4;
		homeRouterTableSize = 0;
		if (strcasecmp(str3, "discover") == 0) {
		    break;
		} else {
		    i = 0;
		    while (((getWord(line, i + 3, word) >= 0) &&
		            (i < MAX_DEFAULT_ROUTERS))) {
		        if (inet_addr(word) == -1) {
		            mip0dbg(("Bad default router value on line %d.\n",
		                     lineNum));
			    state = 8;
			    break;
			} else {
			    homeRouterTable[homeRouterTableSize] =
				inet_addr(word);
			    homeRouterTableSize++;
			}
		        i++;
		    } /* while Loop */
		}
	    }
	    break;

	case 4:  /* expect to see number of home agents */
	    if ((sscanf(line, "%d", &int1) != 1) ||
		(int1 < 1) || (int1 > DEFAULT_MAX_HOME_AGENTS)) {
	        mip0dbg(("Problem reading # of home agents on line %d.\n",
			 lineNum));
	        state = 8;
	    } else {
	        haTableSize = int1;
	        state = 5;
		i = 0;
	    }

	    break;

	case 5:  /* expect one line for each home agent  */
	    mnHomeBcastAddr = ((mnHomeAddr & mnHomeNetmask) |
			      ~(mnHomeNetmask));
	    if ((sscanf(line, "%s %ld %d %d %s",
			str1, &SPI, &int1, &int2, keyStr) != 5) ||
		(int1 < 0) || (int1 > 2) || (int2 > MAX_KEY_LEN) ||
		(int2 < 0)) {
	        mip0dbg(("Problem reading home agent info on line %d.\n",
			 lineNum));
	        state = 8;
	    } else {
	        if ((strcasecmp(str1, "default") == 0) ||
		    (strcasecmp(str1, "discover") == 0)){
		    haTable[i].mnHaAddr = mnHomeBcastAddr;
	            if (HAdiscovery) {
		        mip0dbg(("Error reading HA address on line %d.\n",
			 	lineNum));
		        state = 8;
		        break;
		    }
		    HAdiscovery = TRUE;
		    discoveryAddrIdx = i;
		} else {
		    if((haTable[i].mnHaAddr = inet_addr(str1)) == -1){
		        mip0dbg(("Error reading HA address on line %d.\n",
				 lineNum));
			state = 8;
			break;
		    }
		}

		haTable[i].mnHaMSA.mipSecAddr = haTable[i].mnHaAddr ;
		haTable[i].mnHaMSA.mipSecSPI = SPI;
		haTable[i].mnHaMSA.mipSecReplayMethod = int1;
		haTable[i].mnHaMSA.mipSecAlgorithmType = MD5;
		haTable[i].mnHaMSA.mipSecAlgorithmMode = PREFIXSUFFIX;
		haTable[i].mnHaMSA.mipSecKeyLen = int2;
		if (hexConvert((char *)haTable[i].mnHaMSA.mipSecKey,
		       int2, keyStr) < 0) {
		    mip0dbg(("Problem reading key on line %d.\n",
			     lineNum));
		    state = 8;
		}
		/* Also update the remaining fields  haTable */
		haTable[i].mnHaUnusableTill = 0;
		haTable[i].mnHaWhyUnusable = 0;
	 	haTable[i].mnHALifetime = 0; /* DEFAULT_HA_LIFETIME; */
		haTable[i].mnHAflags = 0;    /* HA_IS_ALIVE; */

		haTable[i].mnHaLastIDhi = 0;
		haTable[i].mnHaTimeSkew = 0;
		i++;
		if (i == haTableSize) {
		   if ((HAdiscovery) && (discoveryAddrIdx !=0))
		       swapHATableEntry(discoveryAddrIdx, 0);
	           if (getProtectedDomainInfo(fd, &lineNum) < 0)
	               state = 8;
		   else
		       state = 7;
		}
	   }
	   break;
        }
    }
    fclose(fd);
    if (state == 7) {
        mip0dbg(("Configuration file read successfully.\n"));
	return(0);
    } else if (state == 8) {
        mip0dbg(("Configuration file had problems.\n"));
	return(-1);
    } else {
        mip0dbg(("Configuration file seems incomplete.\n"));
	return(-1);
    }
}

int readFilterInfo(FILE *ffd) {
    	int  state = 1, lineNum;
	int fltItmCnt;
	int i, j;
	char line[512], str1[25];
	int int1, int2, int3, int4;
	int k, l, bogusFilterCount, x, found;

	while ((fgets(line, 512, ffd) != NULL) && (state != 4) && (state != 5)) {
		/* we allow comments starting with # in column 1 */
		lineNum++;
        	if ((line[0] == '#') || isBlank(line, strlen(line)))
	    		continue;

        	switch (state) {
			case 1:  /* expect to see the number of filters */
	    			if ((sscanf(line, "%d", &int1) != 1) || (int1 < 0)) {
	        			mip0dbg(("Bad filter grouping count on line %d.\n",lineNum));
	        			state = 5;
	    			} else if(int1 == 0) {
	        			state = 4;
	    			} else {
					//mnFilterCount = int1;
					bogusFilterCount = int1;
					//mnFilterOpCount = int1;
					i = 0;
					x = 0;
					state = 2;
				}
				break;
			case 2:  /* expect to see the filter header dinterface & the number of filters */
				if((sscanf(line, "%d %s %d %d ", &int1, str1, &int2, &int3) != 4)) {
	        			mip0dbg(("Bad filter header on line %d.\n",lineNum));
	        			state = 5;
	    			} else {

					if(int3 > 0) {
	        				state = 3;

						mnFilters[i].filterModuleCount = int3;
						mnFilters[i].controlModules[0].boundCoA = 0;
						mnFilters[i].controlModules[0].weight = int2;
						strcpy(mnFilters[i].controlModules[0].boundIfaceName, str1);
						mnFilters[i].idx = int1;

					//mnFilterOpTable[i].action = int2;
					//mnFilterOpTable[i].seq = int1;
					//mnFilterOpTable[i].boundCoA = 0;
					//strcpy(mnFilterOpTable[i].boundIfaceName, str1);

						j = 0;
					} else {
						found = FALSE;
						for(k = 0; k < MAX_FILTERS; k++) {
							if(mnFilters[k].idx == int1) {
								for(l = 0; l < MAX_FILTER_MODULES_PER_FILTER; l++) {
									if(mnFilters[k].controlModules[l].weight == (-1)) {
										mnFilters[k].controlModules[l].boundCoA = 0;
										mnFilters[k].controlModules[l].weight = int2;
										strcpy(mnFilters[k].controlModules[l].boundIfaceName, str1);
										found = TRUE;
										break;
									}
								}
								if(found)
									break;
							}
						}
						x++;
						if(x >= bogusFilterCount) {
							state = 4;
						}
					}
				}
				break;
			case 3: /* expect to see the filter modules*/

				if(getFilterValues(&(mnFilters[i].filterModules[j]), line) == FALSE) {
        				mip0dbg(("Bad filter module on line %d.\n",lineNum));
	        			state = 5;
				} else {
					j++;
					if(j >= mnFilters[i].filterModuleCount) {
						state = 2;
						i++;
						x++;
						if(x >= bogusFilterCount) {
							state = 4;
						}
					}
				}
				break;
		}
	}
	fclose(ffd);
	mnFilterCount = i;
    	if (state == 4) {
        	mip0dbg(("Filter Definition file read successfully.\n"));
		return(0);
    	} else if (state == 5) {
        	mip0dbg(("Filter Definition file had problems.\n"));
		return(-1);
    	} else {
        	mip0dbg(("Filter Definition file seems incomplete.\n"));
		return(-1);
    	}
}

int getFilterValues(FilterModule *fltrmdl, char *ln) {
	int int1, int2, int3;
	char str1[256], str2[256];
	u32 ipaddr1, ipaddr2;

	BAFExtData  *bafedp;
	ProtoExtData *protoedp;
	SrcAddrExtData *srcaddredp;
	SrcNetExtData *srcnetedp;
	SrcPortExtData *srcportedp;
	SrcPortRangeExtData *srcportrangeedp;
	FreeFormExtData *freeformedp;
	DstAddrExtData *dstaddredp;
	DstNetExtData *dstnetedp;
	DstPortExtData *dstportedp;
	DstPortRangeExtData *dstportrangeedp;

	int1 = 0;
	if ((sscanf(ln, "%d", &int1) != 1) || (int1 <= 0))
		return FALSE;
	fltrmdl->filterType = int1;

	switch(fltrmdl->filterType) {
		case FLT_BAF_EXT: // get filtering info relating to the BAF Ext type filter
			if((fltrmdl->filterPtr = malloc(sizeof(BAFExtData))) == NULL)
				return FALSE;

			bafedp = (BAFExtData *) fltrmdl->filterPtr;

			if ((sscanf(ln, "%d %d", &int1, &int2) != 2) || (int2 <= 0))
				return FALSE;

			bafedp->DSCP = int2;

			break;

		case FLT_PROTO_EXT: // get filtering info relating to the Protocol Ext type filter
			if((fltrmdl->filterPtr = malloc(sizeof(ProtoExtData))) == NULL)
				return FALSE;

			protoedp = (ProtoExtData *) fltrmdl->filterPtr;

			if ((sscanf(ln, "%d %d", &int1, &int2) != 2) || (int2 <= 0))
				return FALSE;

			protoedp->protoNum = int2;

			break;

		case FLT_SRCADR_EXT: // get filtering info relating to the Source Address Ext type filter
			if((fltrmdl->filterPtr = malloc(sizeof(SrcAddrExtData))) == NULL)
				return FALSE;

			srcaddredp = (SrcAddrExtData *) fltrmdl->filterPtr;

			if ((sscanf(ln, "%d %s", &int1, str1) != 2) || ((ipaddr1 = inet_addr(str1)) == -1))
				return FALSE;

			srcaddredp->srcAddr = ipaddr1;

			break;

		case FLT_SRCNET_EXT: // get filtering info relating to the Source Network Ext type filter
			if((fltrmdl->filterPtr = malloc(sizeof(SrcNetExtData))) == NULL)
				return FALSE;

			srcnetedp = (SrcNetExtData *) fltrmdl->filterPtr;

			if ((sscanf(ln, "%d %s %s", &int1, str1, str2) != 3) || ((ipaddr1 = inet_addr(str1)) == -1)
								|| ((ipaddr2 = inet_addr(str2)) == -1))
				return FALSE;

			srcnetedp->srcAddr = ipaddr1;
			srcnetedp->srcNetMask = ipaddr2;

			break;

		case FLT_SRCPORT_EXT: // get filtering info relating to the Source Port Ext type filter
			if((fltrmdl->filterPtr = malloc(sizeof(SrcPortExtData))) == NULL)
				return FALSE;

			srcportedp = (SrcPortExtData *) fltrmdl->filterPtr;

			if ((sscanf(ln, "%d %d", &int1, &int2) != 2) || (int2 <= 0))
				return FALSE;

			srcportedp->portNum = int2;

			break;

		case FLT_SRCPORTRNG_EXT: // get filtering info relating to the Source Port Ranges Ext type filter
			if((fltrmdl->filterPtr = malloc(sizeof(SrcPortRangeExtData))) == NULL)
				return FALSE;

			srcportrangeedp = (SrcPortRangeExtData *) fltrmdl->filterPtr;

			if ((sscanf(ln, "%d %d %d", &int1, &int2, &int3) != 3) || (int2 <= 0) || (int3 <= 0))
				return FALSE;

			srcportrangeedp->minPortNum = int2;
			srcportrangeedp->maxPortNum = int3;

			break;

		case FLT_FRFORM_EXT: // get filtering info relating to the Free Form Ext type filter
			if((fltrmdl->filterPtr = malloc(sizeof(FreeFormExtData))) == NULL)
				return FALSE;

			freeformedp = (FreeFormExtData *) fltrmdl->filterPtr;

			if ((sscanf(ln, "%d %d %s %s", &int1, &int2, str1, str2) != 4) || (int2 <= 0))
				return FALSE;

			freeformedp->offset = int2;
			strcpy(freeformedp->value, str1);
			strcpy(freeformedp->mask, str2);
			freeformedp->length = strlen(str1);

			break;

		// new filters
		case FLT_DSTADR_EXT: // get filtering info relating to the Destination Address Ext type filter
			if((fltrmdl->filterPtr = malloc(sizeof(DstAddrExtData))) == NULL)
				return FALSE;

			dstaddredp = (DstAddrExtData *) fltrmdl->filterPtr;

			if ((sscanf(ln, "%d %s", &int1, str1) != 2) || ((ipaddr1 = inet_addr(str1)) == -1))
				return FALSE;

			dstaddredp->dstAddr = ipaddr1;

			break;

		case FLT_DSTNET_EXT: // get filtering info relating to the Destination Network Ext type filter
			if((fltrmdl->filterPtr = malloc(sizeof(DstNetExtData))) == NULL)
				return FALSE;

			dstnetedp = (DstNetExtData *) fltrmdl->filterPtr;

			if ((sscanf(ln, "%d %s %s", &int1, str1, str2) != 3) || ((ipaddr1 = inet_addr(str1)) == -1)
								|| ((ipaddr2 = inet_addr(str2)) == -1))
				return FALSE;

			dstnetedp->dstAddr = ipaddr1;
			dstnetedp->dstNetMask = ipaddr2;

			break;

		case FLT_DSTPORT_EXT: // get filtering info relating to the Destination Port Ext type filter
			if((fltrmdl->filterPtr = malloc(sizeof(DstPortExtData))) == NULL)
				return FALSE;

			dstportedp = (DstPortExtData *) fltrmdl->filterPtr;

			if ((sscanf(ln, "%d %d", &int1, &int2) != 2) || (int2 <= 0))
				return FALSE;

			dstportedp->portNum = int2;

			break;

		case FLT_DSTPORTRNG_EXT: // get filtering info relating to the Destination Port Ranges Ext type filter
			if((fltrmdl->filterPtr = malloc(sizeof(DstPortRangeExtData))) == NULL)
				return FALSE;

			dstportrangeedp = (DstPortRangeExtData *) fltrmdl->filterPtr;

			if ((sscanf(ln, "%d %d %d", &int1, &int2, &int3) != 3) || (int2 <= 0) || (int3 <= 0))
				return FALSE;

			dstportrangeedp->minPortNum = int2;
			dstportrangeedp->maxPortNum = int3;

			break;

		default:
			return FALSE;
	}

	return TRUE;
}

/* This function initializes the filter & operations tables
 */
void initFilterTable() {
	int i, j;

	for(i = 0; i < MAX_FILTERS; i++) {
		mnFilterOpTable[i].idx = (-1);
		mnFilterOpTable[i].action = (-1);

		mnFilters[i].idx = (-1);
		mnFilters[i].filterModuleCount = 0;
		mnFilters[i].markValue = (-1);

		for(j = 0; j < MAX_FILTER_MODULES_PER_FILTER; j++) {
			mnFilters[i].controlModules[j].boundCoA = 0;
			mnFilters[i].controlModules[j].weight = (-1);
			*(mnFilters[i].controlModules[j].boundIfaceName) = (char) NULL;
		}
	}
	mnFilterCount = 0;
	mnFilterOpCount = 0;
}

void initIfaceTable()
{
    int i;

    ifaceTableSize = 0;
    for (i = 0; i < DEFAULT_MAX_INTERFACES; i++) {
        memset(&ifaceTable[i], 0, sizeof(IfaceEntry));
    }

}
/* Set localMaTable to empty */
void initLocalMaTable()
{
    int i;

    localMaTableSize = 0;
    for (i = 0; i < DEFAULT_MAX_MOBILITY_AGENTS; i++) {
        memset(&localMaTable[i], 0, sizeof(LocalMaEntry));
    }
}
/* Set every pointer in knownLocalMaTable to point to its */
/* corresponding entry in localMaTable - modified by Uni. */
/* Sheffield.						  */
void initKnownLocalMaTable()
{
    int i;

    for (i = 0; i < DEFAULT_MAX_MOBILITY_AGENTS; i++) {
        knownLocalMaTable[i] = NULL;
    }
}
/* Set Reg. request table to empty */
void initRegReqTable()
{
    int i;

    regReqTableSize = 0;
    for (i = 0; i < DEFAULT_MAX_REG_REQ; i++) {
        memset(&regReqTable[i], 0, sizeof(RegReqEntry));
	regReqTable[i].mnRegStatus = REG_UNUSED;
    }
}

/* set state table to empty*/
void initISTable() {
    int i;

    for (i = 0; i < DEFAULT_MAX_INTERFACES; i++) {
	memset(ISTable[i].ifaceName, (char) NULL, MAX_IFNAME_LEN);
	ISTable[i].COAddr = 0;

	ISTable[i].currentState.state = UNKNOWN_STATE;
	ISTable[i].currentState.backoffIdx = 0;
	memset(ISTable[i].currentState.ifaceName, (char) NULL, MAX_IFNAME_LEN);
	ISTable[i].currentState.routerAddr = 0;
	ISTable[i].currentState.COAddr = 0;
	ISTable[i].currentState.timeSpent = 0;
	ISTable[i].currentState.nextSolicit = (retransmissionPolicy[0] * 1);
    }

    strcpy(ISTable[0].ifaceName, "NoIfc");
    ISTable[0].COAddr = 0;
    ISTable[0].currentState.state = UNKNOWN_STATE;
    ISTable[0].currentState.backoffIdx = 0;
    strcpy(ISTable[0].currentState.ifaceName, "NoIfc");
    ISTable[0].currentState.routerAddr = 0;
    ISTable[0].currentState.COAddr = 0;
    ISTable[0].currentState.timeSpent = 0;
    ISTable[0].currentState.nextSolicit = (retransmissionPolicy[0] * 1);

    ISTableSize = 1;
}

void initState()
{
    u32 currTime;
    int i;

    //currentState.state = UNKNOWN_STATE ;
    currTime = CurrentTimeNTPSec();
    for (i = 0; i < ifaceTableSize; i++) {
	sendSolicitation(ifaceTable[i].maIfaceAddr);
	ifaceTable[i].maIfaceSolicitationTime = currTime;
    }
    //currentState.backoffIdx = 0;
    //currentState.nextSolicit = (retransmissionPolicy[0] * 1);
}

void doAlarm(int junk)
{
    timerWentOff = 1;
    if (signal(SIGALRM, doAlarm) == SIG_ERR) {
        mip0dbg(("Error: Failed to set alarm in DoAlarm().\n"));
    }
    alarm(periodicInterval);
}

void Finalize(int junk)
{
    int i;

    mip2dbg(("\n---- %s (%s) ----\n", sprintTime(currentTime, 30),
             sprintRelativeTime(relativeTime, 30)));
    /*
     * TODO: Remove the assumption that the first interface is
     * what we ought to use to configure ourselves for "home".
     */

    for(i = 1;i < DEFAULT_MAX_INTERFACES; i++) {
        if(*(ISTable[i].ifaceName) == (char) NULL)
		continue;
    	undoState(&(ISTable[i].currentState));
    	if (ifaceTableSize > 0) {
//        	strcpy(currentState.ifaceName, ifaceTable[0].maIfaceName);
        	strcpy(ISTable[i].currentState.ifaceName, ISTable[i].ifaceName);

		// set filter removal with the new deregistration
		// when comming back to Home
		setFilterForHomeBinding(mnFilters, mnFilterCount, mnFilterOpTable,
							&mnFilterOpCount, ISTable[i].currentState.ifaceName);

		startHomeAction(&(ISTable[i].currentState), i);
		deleteAllFilters(mnFilters);

    	}
    }
    OScleanup();
    termInfoModule();
    mip2dbg(("Terminated by signal (SIGTERM or SIGINT)\n"));
    exit(-1);
}

/* Set up handlers for various signals */
void SetSigHandlers()
{
    if (signal(SIGINT, Finalize) == SIG_ERR) {
        mip0dbg(("Error: failed to set handler for SIGINT.\n"));
    }
 
    if (signal(SIGTERM, Finalize) == SIG_ERR) {
        mip0dbg(("Error: failed to set handler for SIGTERM.\n"));
    }
}

int InitSocket()
{
    struct sockaddr_in sa;

    /*
     * Open the ICMP sock for sending agent solicitation
     * Note that we don't need an ICMP socket to receive solicitations.
     * We use DLPI to grab the entire link level frame including
     * the sender's hardware address
     */

    mip2dbg(("Creating ICMP socket for sending agent solictations.\n"));
    if ((ICMPsolicitSock = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP)) < 0) {
        mip0dbg(("Error: Could not create ICMP socket in InitSockets().\n"));
	return (-1);
    }


    /*
     * Create UDPregSock for sending registration requests and
     * receiving registration replies.
     */

    mip2dbg(("Creating UDP socket for Mobile IP registrations & replies.\n"));
    if ((UDPregSock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
	mip0dbg(("Error: Could not create UDP socket in InitSockets().\n"));
	return (-1);
    }

    memset((char *)&sa, 0, sizeof(sa));
    sa.sin_family = AF_INET;
    sa.sin_port = htons(0);
    sa.sin_addr.s_addr = htonl(INADDR_ANY);

    if (bind(UDPregSock, (struct sockaddr *) &sa, sizeof(sa)) < 0) {
	 mip0dbg(("Error: Could not bind UDP socket.\n"));
	 return(-1);
    }
    return(0);
}

void showConfigInfo()
{
    int i;
    int j;

    mip3dbg(("Debuglevel                  : %d\n", mipdebug));
    mip3dbg(("PeriodicInterval            : %d\n", periodicInterval));
    mip3dbg(("AgentSolicitationThreshold  : %d\n", agentSolicitThreshold));
    mip3dbg(("AgentExpireThreshold        : %d\n", agentExpireThreshold));
    mip3dbg(("IfacePollthreshold          : %d\n", ifacePollThreshold));
    mip3dbg(("RegLifeTime                 : %d\n", regLifetime));
    mip3dbg(("RegRoundTrip                : %d\n", regRoundTrip));
    mip3dbg(("RegFlags                    : 0x%x\n",regFlags));
    mip3dbg(("HACoolOffPeriod             : %d\n", haCooloffPeriod));
    mip3dbg(("FACoolOffPeriod             : %d\n", faCooloffPeriod));
    mip3dbg(("IDFreshnessSlack            : %d\n", IDfreshnessSlack));
    mip3dbg(("Retrasmission Policy        : "));
    for (i = 0; retransmissionPolicy[i] != -1; i++)
        mip3dbg(("%d      ",retransmissionPolicy[i]));
    mip3dbg(("\n"));
    mip3dbg(("Renewal Policy              : "));
    for (i = 0; renewalPolicy[i] != 0; i++)
        mip3dbg(("%f      ",renewalPolicy[i]));
    mip3dbg(("\n"));
    /* Display movement detection method choice - modified by */
    /* the Uni. of Sheffield.				      */
    mip3dbg(("Movement Detection Method   : "));
    switch(movementDetection) {
    case 1: printf("LCS\n");
            break;
    case 2: printf("ECS\n");
            break;
    case 3: printf("HCS\n");
            break;
    case 4: printf("FHCS\n");
            break;
    default: printf("ERROR\n");
    }
    mip3dbg(("HApriority                  : %d\n", HApriority));
    mip3dbg(("EagerSoliciting             : %d\n", EagerSoliciting));
    mip3dbg(("LazySoliciting              : %d\n", LazySoliciting));
    mip3dbg(("Interfaces To Be Ignored    : "));
    for (i = 0; i < ignoreIfaceCnt; i++)
        mip3dbg(("%s      ", ignoreIfaceList[i]));
    mip3dbg(("\n"));
    mip3dbg(("HomeAddress                 : %s\n",
			ntoa(mnHomeAddr, addr1)));
    mip3dbg(("Netmask                     : %s\n",
			ntoa(mnHomeNetmask, addr1)));
    mip3dbg(("HomeBroadCastAddress        : %s\n",
			ntoa(mnHomeBcastAddr, addr1)));
    mip3dbg(("HomeRouters                 : "));
    if (homeRouterTableSize == 0) {
	mip3dbg(("DISCOVERY\n"));
    } else {
	for ( i = 0; i < homeRouterTableSize; i++)
	    mip3dbg(("%s\t", ntoa(homeRouterTable[i], addr1)));
	mip3dbg(("\n"));
    }
    printHATable(haTable, haTableSize);

    mip3dbg(("\n"));
    mip3dbg(("FilterStatus                : %d\n", FilterStatus));
    mip3dbg(("\n"));
    mip3dbg(("\n"));

    if (domainInfo.addrIntervalCnt > 0) {
        printProtectedDomainInfo(domainInfo);
	mip3dbg(("\n"));
    }
}

/*
* Display all data related to filters
*/
void showFilterInfo()
{
	int i, j;
	char addr1[25], addr2[25], fltType[256];
	BAFExtData  *bafedp;
	ProtoExtData *protoedp;
	SrcAddrExtData *srcaddredp;
	SrcNetExtData *srcnetedp;
	SrcPortExtData *srcportedp;
	SrcPortRangeExtData *srcportrangeedp;
	FreeFormExtData *freeformedp;

	DstAddrExtData *dstaddredp;
	DstNetExtData *dstnetedp;
	DstPortExtData *dstportedp;
	DstPortRangeExtData *dstportrangeedp;


	mip3dbg(("\n\n----------Filter Data------------ \n"));
	mip3dbg(("Number of Filters        : %d\n", mnFilterCount));

	for(i = 0; i < mnFilterCount; i++) {
		mip3dbg(("  Index Number           : %d\n", mnFilters[i].idx));
		mip3dbg(("  Filter Module Count    : %d\n", mnFilters[i].filterModuleCount));
		if(mnFilters[i].markValue == (-1)) mip3dbg(("  Unallocated\n"));
		else mip3dbg(("  Allocated\n"));

		for(j = 0; j < MAX_FILTER_MODULES_PER_FILTER; j++) {
			if(mnFilters[i].controlModules[j].weight == (-1))
				break;
			mip3dbg(("  Bound Interface        : %s\n", mnFilters[i].controlModules[j].boundIfaceName));
			//mip3dbg(("     Bound CoA              : %s\n", mnFilters[i].controlModules[j].boundCoA));
			mip3dbg(("     Weight              : %d\n", mnFilters[i].controlModules[j].weight));
		}

		for(j = 0; j < mnFilters[i].filterModuleCount; j++) {
			switch(mnFilters[i].filterModules[j].filterType) {
				case FLT_BAF_EXT:
					strcpy(fltType, "Behavior Aggregate Filter Extension");
					mip3dbg(("    Filter Type          : %s\n", fltType));

					bafedp = (BAFExtData *) mnFilters[i].filterModules[j].filterPtr;

					mip3dbg(("    DSCP Value           : %d\n\n", bafedp->DSCP));

					break;

				case FLT_PROTO_EXT:
					strcpy(fltType, "Protocol Extension");
					mip3dbg(("    Filter Type          : %s\n", fltType));

					protoedp = (ProtoExtData *) mnFilters[i].filterModules[j].filterPtr;
					mip3dbg(("    Protocol Number      : %d\n\n", protoedp->protoNum));

					break;

				case FLT_SRCADR_EXT:
					strcpy(fltType, "Source Address Extension");
					mip3dbg(("    Filter Type          : %s\n", fltType));

					srcaddredp = (SrcAddrExtData *) mnFilters[i].filterModules[j].filterPtr;
					mip3dbg(("    Source Address       : %s\n\n", ntoa(srcaddredp->srcAddr, addr1)));

					break;

				case FLT_SRCNET_EXT:
					strcpy(fltType, "Source Network Extension");
					mip3dbg(("    Filter Type          : %s\n", fltType));

					srcnetedp = (SrcNetExtData *) mnFilters[i].filterModules[j].filterPtr;
					mip3dbg(("    Source Address & Mask: %s/%s\n\n",
										ntoa(srcnetedp->srcAddr, addr1),
										ntoa(srcnetedp->srcNetMask, addr2)));

					break;

				case FLT_SRCPORT_EXT:
					strcpy(fltType, "Source Port Extension");
					mip3dbg(("    Filter Type          : %s\n", fltType));

					srcportedp = (SrcPortExtData *) mnFilters[i].filterModules[j].filterPtr;
					mip3dbg(("    Port Number          : %d\n\n", srcportedp->portNum));

					break;

				case FLT_SRCPORTRNG_EXT:
					strcpy(fltType, "Source Port Range Extension");
					mip3dbg(("    Filter Type          : %s\n", fltType));

					srcportrangeedp = (SrcPortRangeExtData *) mnFilters[i].filterModules[j].filterPtr;
					mip3dbg(("    Port Number Range    : %d-%d\n\n",
										srcportrangeedp->minPortNum,
										srcportrangeedp->maxPortNum));

					break;

				case FLT_FRFORM_EXT:
					strcpy(fltType, "Free-Form Extension");
					mip3dbg(("     Filter Type            : %s\n", fltType));
					freeformedp = (FreeFormExtData *) mnFilters[i].filterModules[j].filterPtr;
					mip3dbg(("     Offset, Value & Mask   : %d %s %s\n\n",
										freeformedp->offset,
										freeformedp->value,
										freeformedp->mask));

					break;

				// new filters

				case FLT_DSTADR_EXT:
					strcpy(fltType, "Destination Address Extension");
					mip3dbg(("    Filter Type          : %s\n", fltType));

					dstaddredp = (DstAddrExtData *) mnFilters[i].filterModules[j].filterPtr;
					mip3dbg(("    Destination Address       : %s\n\n", ntoa(dstaddredp->dstAddr, addr1)));

					break;

				case FLT_DSTNET_EXT:
					strcpy(fltType, "Destination Network Extension");
					mip3dbg(("    Filter Type          : %s\n", fltType));

					dstnetedp = (DstNetExtData *) mnFilters[i].filterModules[j].filterPtr;
					mip3dbg(("    Destination Address & Mask: %s/%s\n\n",
										ntoa(dstnetedp->dstAddr, addr1),
										ntoa(dstnetedp->dstNetMask, addr2)));

					break;

				case FLT_DSTPORT_EXT:
					strcpy(fltType, "Destination Port Extension");
					mip3dbg(("    Filter Type          : %s\n", fltType));

					dstportedp = (DstPortExtData *) mnFilters[i].filterModules[j].filterPtr;
					mip3dbg(("    Port Number          : %d\n\n", dstportedp->portNum));

					break;

				case FLT_DSTPORTRNG_EXT:
					strcpy(fltType, "Destination Port Range Extension");
					mip3dbg(("    Filter Type          : %s\n", fltType));

					dstportrangeedp = (DstPortRangeExtData *) mnFilters[i].filterModules[j].filterPtr;
					mip3dbg(("    Port Number Range    : %d-%d\n\n",
										dstportrangeedp->minPortNum,
										dstportrangeedp->maxPortNum));

					break;

				default:
					strcpy(fltType, "Cannot determine filter type");
					mip3dbg(("     Filter Type            : %s\n", fltType));

					break;
			}

		}
	}
}

void Initialize(FILE *fd)
{
    FILE *ffd;

    randomInit();

    if (readConfigInfo(fd) <0 ){
	mip0dbg(("Problems with configuration file\n"));
        exit(-1);
    }
    showConfigInfo();

    initFilterTable();

    /* get filter info, if flag set */
    if(FilterStatus) {
    	/* open the config file that has all the filter details */
    	if ((ffd = fopen(FILTER_FILE_NAME, "r")) == NULL) {
		fprintf(stderr, "Could not open %s\n", FILTER_FILE_NAME);
		exit(-1);
    	}

    	if(readFilterInfo(ffd) < 0) {
		mip0dbg(("Problems with filter definition file\n"));
        	exit(-1);
    	}
    }
    showFilterInfo();


    initIfaceTable();
    initLocalMaTable(); /* set to empty */
    initKnownLocalMaTable(); /* set to point to localMaTable */
			     /* modified by Uni. Sheffield.  */
    initRegReqTable();  /* set to empty */
    initISTable(); // set to empty
    /*
     * Do OS-specific initialization, eg tunneling interface, also
     * initialize interface list, sockets, kill in.rdisc etc.
     */
    InitSocket();
    initInfoModule();
    OSinit();

    initState();        /* set to UNKNOWN, solicit on each iface etc */
    doAlarm(0);
    SetSigHandlers();   /* Also maps ^C to Finalize() */
    mip2dbg(("\nInitialization completed.\n"));
}

