/*

  Sun Public License Notice

 The contents of this file are subject to the Sun Public License Version
 1.0 (the "License"). You may not use this file except in compliance with
 the License. A copy of the License is available at http://www.sun.com/

 The Original Code is the Sunlabs Mobile IP.
 The Initial Developer of the Original Code is Vipul Gupta.
 Portions created by Vipul Gupta are Copyright (C) 2003.
 All Rights Reserved.

 Contributor(s): Niko A. Fikouras (niko@comnets.uni-bremen.de)
		 Asanga D. Udugama (adu@comnets.uni-bremen.de)

*/

/*
 * An implementation of Mobile-IP as described in RFC 2002.
 * Authors  : Vipul Gupta (vipul.gupta@eng.sun.com)
 *            Sunil Madhani (sunil.madhani@eng.sun.com)
 * mnPeriodic.c : Collection of routines executed periodically by a mobile
 *                node.
 */

#pragma ident	"@(#)mnPeriodic.c 1.2 98/02/11"

#include <stdio.h>
#include "mip.h"
#include "mobile.h"


extern char currentTime[], relativeTime[], addr1[], addr2[], addr3[];

extern int mipdebug;
extern int periodicInterval;
extern int agentSolicitThreshold;
extern int agentExpireThreshold;
extern int ifacePollThreshold;
extern int regLifetime;
extern int regRoundTrip;
extern int regFlags;
extern int haCooloffPeriod;
extern int faCooloffPeriod;
extern int IDfreshnessSlack;
extern int solicitOnBcast;
extern int retransmissionPolicy[];
extern float renewalPolicy[];
/* defines the movement detection method - modified by Uni. Sheffield */
/* modified by Uni. of Sheffield */
extern int movementDetection;

/* Determines whether the MN gives priority to HA's */
/* modified by Uni. of Sheffield */
extern int HApriority;

/* The mobile node's address, netmask and default routers at home */
extern u32 mnHomeAddr;
extern u32 mnHomeNetmask;
extern u32 mnHomeBcastAddr;

// currnetState is part of ISTable; changed to handle MSB
extern InterfaceState ISTable[DEFAULT_MAX_INTERFACES];
//extern mnState currentState;
extern int ISTableSize;

extern IfaceEntry ifaceTable[];
extern int ifaceTableSize;

extern HomeAgentEntry haTable[];
extern int haTableSize;

/* List of default routers on the home network */
extern u32 homeRouterTable[];
extern int homeRouterTableSize;

/* List of local Mobility Agents (considered active) */
extern LocalMaEntry localMaTable[];
extern int localMaTableSize;

/* List of known local Mobility Agents (considered active) */
/* Used by the ECS movement detection method - modified by */
/* Uni. Sheffield					   */
extern LocalMaEntry *knownLocalMaTable[DEFAULT_MAX_MOBILITY_AGENTS];

/* List of registration requests (both accepted and pending) */
extern RegReqEntry regReqTable[];
extern int regReqTableSize;
extern u32 mnRegRequestsAcceptedCnt;

extern int  ICMPsolicitSock;
extern int  UDPregSock;

extern char *sprintTime(char *, int);
extern void delMaPtr(LocalMaEntry *, int *, LocalMaEntry *);
extern void delHaPtr(HomeAgentEntry *, int *, HomeAgentEntry *);
extern void delRegReqPtr(RegReqEntry *, int *, RegReqEntry *);
extern IfaceEntry * findIface(char *);
RegReqEntry *mkNewRegReq(mnState *, int);
boolean havePendingDiscovery(u32);
extern void sendSolicitation(u32);
extern void sendRegReq(RegReqEntry *);
extern u32 chooseHAaddr();

extern char *ntoa(u32, char *);

extern void getNewIfaceList(IfaceEntry *, int *);

extern int applyReverseFilters(Filter [], InterfaceState [], char *);

/* to store the filters that have been defined for this mobile node*/
extern Filter mnFilters[MAX_FILTERS]; 				   // data for each filter
extern int mnFilterCount; 					   // the number of filters
extern FilterOp mnFilterOpTable[MAX_FILTERS];			   // the grid that holds the operations
								   // to be done for the filters
extern int mnFilterOpCount;					   // operations count

/*
 * This function traverses the entire list of COA addresses. At other
 * places, we currently assume that only the first COA is ever used.
 * In that situation, this can be simplified.
 */
boolean isValidCOAinAdv(u32 addr,  LocalMaEntry * lmaep)
{
    int i;

    for (i = 0; i < lmaep->mnAdvCOACnt; i++) {
        if (lmaep->mnAdvCOAList[i] == addr)
	    return(TRUE);
    }
    return(FALSE);
}

/*
 * Update the list of available interfaces. For interfaces that disapper,
 * delete localMATable entries of MAs reachable through them. In turn,
 * pending registration requests that depended on these MAs are deleted
 * and accepted ones are marked non-renewable.
 */
void mnUpdateIfaces(int t)
{
    int i;
    static int pollTime = 0;
    IfaceEntry *ifep;

    pollTime += t;
    if (pollTime < ifacePollThreshold)
        return;                            /* Not time yet for the poll */

    /* Time to poll interfaces */
    pollTime = 0;

    /* Use an OS-specific routine to update the ifaceTable and its size */
    getNewIfaceList(ifaceTable, &ifaceTableSize);

    /*
     * Walk down the list of local mobility agents. Delete them if they
     * were reachable through an interface that disappeared.
     */

    /* Below for loop removed in order to simulate handoffs.
     * We got a MN attached to two subnets. Through the use
     * of a small home made program we can manipulate these
     * attachement points and perform handoffs through a
     * controled way. Mainly, we can determine the length of
     * the overlap are.
     * Modified by Uni. of Sheffield.
     */

/*    for (i = 0; i < DEFAULT_MAX_MOBILITY_AGENTS; i++) {
	if (localMaTable[i].mnAdvSourceAddr == 0) continue;
	ifep = findIface(localMaTable[i].mnAdvIncomingIface);
	if (ifep == NULL)
	    delMaPtr(localMaTable, &localMaTableSize, &(localMaTable[i]));
    }*/
}

/* Age the list of local mobility agents considered alive.  */
void mnAgeLocalMAs(int t)
{
    int i, j, chipped;
    u32 currTime, timeSince;
    LocalMaEntry *lmaep;
    IfaceEntry *ifep;

    for (i = 0; i < DEFAULT_MAX_MOBILITY_AGENTS; i++) {
	/* Is this a valid entry in the MA table? */
	if (localMaTable[i].mnAdvSourceAddr == 0) continue;

	lmaep = &(localMaTable[i]);
	/*
	 * Decrement COAunusableTill and AdvRemainingLifetime fields.
	 * TODO: If we always use the first COA then this can be simpler.
	 */
	for (j = 0; j < lmaep->mnAdvCOACnt; j++) {
	    if (lmaep->mnAdvCOAunusableTill[j] > t)
		lmaep->mnAdvCOAunusableTill[j] -= t;
	    else {
		lmaep->mnAdvCOAunusableTill[j] = 0;
		lmaep->mnAdvCOAwhyUnusable[j] = 0;
	    }
	}

	if (lmaep->mnAdvRemainingLifetime > t)
	    lmaep->mnAdvRemainingLifetime -= t;
	else
	    lmaep->mnAdvRemainingLifetime = 0;

	chipped = (lmaep->mnAdvMaxAdvLifetime -
		   lmaep->mnAdvRemainingLifetime);

	if ((lmaep->mnAdvRemainingLifetime == 0) ||
	    (chipped >= agentExpireThreshold)) {
	    /*
	     * Expire this agent, removing pending registrations through
	     * it and marking accepted ones as non-renewable.
	     */
	    delMaPtr(localMaTable, &localMaTableSize, lmaep);
	} else if (chipped >= agentSolicitThreshold) {
	    /* time to check if agent is alive by sending a solicitation */
	    ifep = findIface(lmaep->mnAdvIncomingIface);
	    if (ifep == NULL) {
		mip0dbg(("Found dead interface when soliciting MA.\n"));
		delMaPtr(localMaTable, &localMaTableSize, lmaep);
		continue;     /* done with this MA table entry */
	    }
	    currTime = CurrentTimeNTPSec();
	    timeSince = (currTime - ifep->maIfaceSolicitationTime);

	    /*
	     * Due to rounding with integer arithmetic, the timeSince
	     * variable is often less than the agentSolicitThreshold
	     * even when it shouldn't be. This causes some solicitations
	     * to be curbed unnecessarily. The (-1) adds a safety margin.
	     *
	     * TODO: How important is this, really?
	     */

	    if (timeSince < (agentSolicitThreshold - 1)) {
		mip3dbg(("Curbing agent solicitation.\n"));
	    } else {
		ifep->maIfaceSolicitationTime = currTime;
		sendSolicitation(ifep->maIfaceAddr);
	    }
	}
    }
}

/* Remove entries in the knownLocalMaTable of MAs that have been removed */
/* from the localMaTable- modified by Uni. of Sheffield			 */
void updateKnownLocalMaTable() {

   int i;
   for(i=0; i<DEFAULT_MAX_MOBILITY_AGENTS; i++) {
      if (knownLocalMaTable[i] == NULL) continue;
      if ((knownLocalMaTable[i]->mnAdvSourceAddr) == 0)
         knownLocalMaTable[i] = NULL;
   }
}


/*
 * AGE the list of HAs. We assume HA list stores all valid entries
 * contiguously from 0 through (HATableSize - 1).
 */
void mnAgeHAs(int t)
{
    int i;  
    HomeAgentEntry *haep;
    
    for (i = 0; i < haTableSize; i++) {
	haep = &(haTable[i]);
	
	/* Decrement HaUnusableTill and HaLifetime fields */
	if (haep->mnHaUnusableTill > t) 
	    haep->mnHaUnusableTill -= t;
	else {
	    haep->mnHaUnusableTill = 0;
	    haep->mnHaWhyUnusable = 0;
	}

	if (haep->mnHALifetime > t)
	    haep->mnHALifetime -= t;
	else {
	    /* HA is no longer alive */
	    haep->mnHALifetime = 0;
	    haep->mnHAflags &= ~(HA_IS_ALIVE);
	    if (haep->mnHAflags & HA_IS_DYNAMICALLY_DISCOVERED) {
		/* free the entry */
		delHaPtr(haTable, &haTableSize, haep);
	    }
	}
    }        
}

/*
 * Retransmit the pending registration request pointed to by rrep. We assume
 * the retransmissionPolicy vector contains one or more positive integers
 * in ascending order and the last entry is non-positive.
 */
void retransmitRegReq(RegReqEntry regReqTable[], int *regReqTableSize,
		      RegReqEntry *rrep)
{
    int backoffFactor;

    /* Increment mnRegAttemptsIdx if appropriate */
    if (retransmissionPolicy[rrep->mnRegAttemptIdx + 1] > 0)
	rrep->mnRegAttemptIdx++;

    backoffFactor = retransmissionPolicy[rrep->mnRegAttemptIdx];
    rrep->mnRegNextTime = (regRoundTrip * backoffFactor);
    sendRegReq(rrep);
}

/*
 * Renew the accepted, renewable registration request pointed to by rrep.
 * We assume the renewalPolicy vector contains one or more positive fractions
 * in descending order ending with a zero.
 */
void renewRegReq(RegReqEntry regReqTable[], int *regReqTableSize,
		      RegReqEntry *rrep)
{
    u32 renewalLevel;

    /* Increment mnRegAttemptsIdx if appropriate. Use an approximation for 0 */
    if (renewalPolicy[rrep->mnRegAttemptIdx] > 0.0001)
	rrep->mnRegAttemptIdx++;

    renewalLevel = (renewalPolicy[rrep->mnRegAttemptIdx] *
		    rrep->mnRegTimeRequested);

    if (rrep->mnRegTimeRemaining > renewalLevel) {
        rrep->mnRegNextTime = rrep->mnRegTimeRemaining - renewalLevel;
    } else {
        /* no more renewals */
        rrep->mnRegNextTime = rrep->mnRegTimeRemaining;
    }

    sendRegReq(rrep);
}

/*
 * Age registrations -- some may be deleted, some (pending) registrations
 * will result in retransmission attempts, some (renewable accepted)
 * registrations will result in renewals. The renewals will create new
 * pending registrations.
 */
void mnAgeRegistrations(int t)
{
    int i;
    RegReqEntry *rrep;
    char ifaceName[MAX_IFNAME_LEN];

    // check & delete expired bindings
    for (i = 0; i < DEFAULT_MAX_REG_REQ; i++) {
	if (regReqTable[i].mnRegStatus == REG_UNUSED) continue;

	rrep = &(regReqTable[i]);

	/* Decrement time remaining. */
	if (rrep->mnRegTimeRemaining > t)
	    rrep->mnRegTimeRemaining -= t;
	else {
	    /* Time to give up. */
	    rrep->mnRegTimeRemaining = 0;
		fprintf(stderr, "Here\n");
	    strcpy(ifaceName, rrep->mnRegIfaceName);
	    delRegReqPtr(regReqTable, &regReqTableSize, rrep);

	    // apply reverse filters when a binding is deleted
	    applyReverseFilters(mnFilters, ISTable, ifaceName);

	    // send info to all the connected GUIs
	    sendInfoToAllGUIs();

	    continue;       /* done with this regReq entry */
	}

	/* Time for a retransmission? renewal? */
	if (rrep->mnRegNextTime > t)
	    rrep->mnRegNextTime -= t;
	else {
	    rrep->mnRegNextTime = 0;
	    if (rrep->mnRegStatus == REG_PENDING)
		retransmitRegReq(regReqTable, &regReqTableSize, rrep);
	    else if (rrep->mnRegStatus == REG_ACCEPTED_RENEWABLE)
		renewRegReq(regReqTable, &regReqTableSize, rrep);
	}
    }

}


/*
 * Make sure we have either completely deregistered ourselves from all
 * HAs or have such deregistration or discovery requests pending. One
 * of the local mobility agents should belong to our home network.
 */
boolean okHomeState(mnState *mnsp, int ISSeq)
{
    int i;
    boolean found = FALSE;
    boolean check = TRUE;

    /* We should not see any ACCEPTED registration requests. They should
     * all have been converted to deregistration requests. When a
     * deregistration is successful, the request is removed.
     */
    for (i = 0; i < DEFAULT_MAX_REG_REQ; i++) {
	if ( ((regReqTable[i].mnRegStatus == REG_ACCEPTED_RENEWABLE) ||
	    (regReqTable[i].mnRegStatus == REG_ACCEPTED_NONRENEWABLE))
	    && (strcmp(regReqTable[i].mnRegIfaceName, ISTable[ISSeq].ifaceName) == 0)) {
	    mip3dbg(("Found accepted registration in Home state.\n"));
	    /* TODO: enable printing of all entries. */
	    /* if (mipdebug > 2)
	     *   printRegReq(&(regReqTable[i]));
	     */
	    check = FALSE;
	}
    }

    /* Ideally, one of the mobility agents must be from the home network */
    for (i = 0; i <  DEFAULT_MAX_MOBILITY_AGENTS; i++) {
	if (localMaTable[i].mnAdvSourceAddr == 0) continue;
	if (((localMaTable[i].mnAdvSourceAddr ^ mnHomeAddr) & mnHomeNetmask)
	    == 0) {
	    found = TRUE;
	    break;
	}
    }

    if (!found) {
	mip3dbg(("No active mobility agent from home network\n"));
	check = FALSE;
    }

    return(check);
}

/*
 * Make sure we are registered or have such a registration pending (or
 * maybe we have a HA discovery pending) and the FA is in our list of
 * local mobility agents.
 */
boolean okForeignState(mnState *mnsp, int ISSeq)
{
    int i;
    boolean found = FALSE;
    RegReqEntry *rrep;

    /* Is this FA still alive? */
    for (i = 0; i <  DEFAULT_MAX_MOBILITY_AGENTS; i++) {
	if (localMaTable[i].mnAdvSourceAddr == 0) continue;
	if ((localMaTable[i].mnAdvSourceAddr == mnsp->routerAddr) &&
	    isValidCOAinAdv(mnsp->COAddr, &(localMaTable[i]))) {
	    found = TRUE;
	    break;
	}
    }

    if (!found) {
	mip3dbg(("Foreign agent no longer usable in foreign state.\n"));
	return(FALSE);
    }

    /*
     * Must have a pending or accepted registration request with
     * destination as our routerAddr (FA) and a non-broadcast HA address.
     * If not, it is time to send a new request. If the new request will
     * be a discovery and we already have one pending then curb the request.
     */
    found = FALSE;
    for (i = 0; i < DEFAULT_MAX_REG_REQ; i++) {
	if (regReqTable[i].mnRegStatus == REG_UNUSED) continue;
	if (((regReqTable[i].mnRegStatus == REG_PENDING) ||
	     (regReqTable[i].mnRegStatus == REG_ACCEPTED_RENEWABLE)) &&
	    (regReqTable[i].mnRegAgentAddr == mnsp->routerAddr) &&
	    (regReqTable[i].mnRegHAaddr != mnHomeBcastAddr) &&
	    (strcmp(regReqTable[i].mnRegIfaceName, ISTable[ISSeq].ifaceName) == 0)) {
	    found = TRUE;
        }
    }

    if (!(found)) {
        if ((chooseHAaddr() != mnHomeBcastAddr) ||
	    (havePendingDiscovery(mnHomeBcastAddr) == FALSE)) {
	    if ((rrep = mkNewRegReq(mnsp, ISSeq)) == NULL) {
	        mip1dbg(("Could not make new registration request for " \
			 "FA's COA %s.\n", ntoa(mnsp->COAddr, addr1)));
	    } else {
	        sendRegReq(rrep);
	    }
	}
    }

    return(TRUE);
}

/*
 * Make sure we are registered or have such a registration pending
 * (or maybe we have a HA discovery pending). Also make sure our colocated
 * care-of address hasn't expired.
 */
boolean okColocatedState(mnState *mnsp, int ISSeq)
{
    int i;
    boolean found = FALSE;
    RegReqEntry *rrep;


    for (i = 0; i <  ifaceTableSize; i++) {
	if ((ifaceTable[i].maIfaceAddr == mnsp->COAddr)
	    && (strcmp(ifaceTable[i].maIfaceName, ISTable[ISSeq].ifaceName) == 0)) {
	    found = TRUE;
	    break;
	}
    }

    if (!found) {
	mip3dbg(("Did not find an interface with COA.\n"));
	return(FALSE);
    }

    /*
     * Must have a pending or accepted registration request with
     * care of address as the colocated addr and a non-broadcast HA address.
     * If not, it is time to send a new request. If the new request will
     * be a discovery and we already have one pending then curb the request.
     */
    found = FALSE;
    for (i = 0; i < DEFAULT_MAX_REG_REQ; i++) {
	if (regReqTable[i].mnRegStatus == REG_UNUSED) continue;
	if (((regReqTable[i].mnRegStatus == REG_PENDING) ||
	     (regReqTable[i].mnRegStatus == REG_ACCEPTED_RENEWABLE)) &&
	    (regReqTable[i].mnRegCOA == mnsp->COAddr) &&
	    (regReqTable[i].mnRegHAaddr != mnHomeBcastAddr) &&
	    (strcmp(regReqTable[i].mnRegIfaceName, ISTable[ISSeq].ifaceName) == 0)) {
	    found = TRUE;
        }
    }

    if (!(found)) {
        if ((chooseHAaddr() != mnHomeBcastAddr) ||
	    (havePendingDiscovery(mnHomeBcastAddr) == FALSE)) {
	    if ((rrep = mkNewRegReq(mnsp, ISSeq)) == NULL) {
	        mip1dbg(("Could not make new registration request for " \
			 "colocated COA %s.\n", ntoa(mnsp->COAddr, addr1)));
	    } else {
	        sendRegReq(rrep);
	    }
	}
    }

    return(TRUE);
}

/* Make sure we don't have any available interfaces or alive local MAs. */
boolean okIsolatedState(mnState *mnsp, int ISSeq)
{

    if ((ifaceTableSize == 0) && (localMaTableSize == 0))
        return(TRUE);
    else
        return(FALSE);
}

/* Some periodic chores are state specific. Those are done here. */
void mnStateTasks(mnState *mnsp, int t, int ISSeq)
{
    int i;
    u32 currTime;

    mnsp->timeSpent += t;

    switch (mnsp->state) {
    case HOME_STATE:
        if (mipdebug == 3) {
	    mip0dbg((ISTable[ISSeq].ifaceName));
	    mip0dbg(("=H "));
	}
	mip5dbg(("In HOME state for %d sec.\n", mnsp->timeSpent));
	if (!okHomeState(mnsp, ISSeq))
	    mip1dbg(("\nokHomeState failed sanity checks.\n"));
	break;

    case FOREIGN_STATE:
        if (mipdebug == 3) {
	    mip0dbg((ISTable[ISSeq].ifaceName));
	    mip0dbg(("=F "));
        }
	mip5dbg(("In FOREIGN state for %d sec [COA %s Iface %s].\n",
		 mnsp->timeSpent, ntoa(mnsp->COAddr, addr1),
		 mnsp->ifaceName));
	if (!okForeignState(mnsp, ISSeq))
	    mip1dbg(("\nokForeignState failed sanity checks.\n"));
	break;

    case COLOCATED_STATE:
        if (mipdebug == 3) {
	    mip0dbg((ISTable[ISSeq].ifaceName));
	    mip0dbg(("=C "));
	}
	mip5dbg(("In COLOCATED state for %d sec [COA %s Router %s Iface %s]\n",
		 mnsp->timeSpent, ntoa(mnsp->COAddr, addr1),
		 ntoa(mnsp->routerAddr, addr2), mnsp->ifaceName));
	if (!okColocatedState(mnsp, ISSeq))
	    mip1dbg(("\nokColocatedState failed sanity checks.\n"));
	break;

    case ISOLATED_STATE:
        if (mipdebug == 3) {
	    mip0dbg((ISTable[ISSeq].ifaceName));
	    mip0dbg(("=I "));
	}
	mip5dbg(("In ISOLATED state for %d sec.\n", mnsp->timeSpent));
	if (!okIsolatedState(mnsp, ISSeq))
	    mip1dbg(("\nokIsolatedState failed sanity checks.\n"));
	break;

    case UNKNOWN_STATE:
        if (mipdebug == 3) {
	    mip0dbg((ISTable[ISSeq].ifaceName));
	    mip0dbg(("=U "));
	}
	mip5dbg(("In UNKNOWN state for %d sec.\n", mnsp->timeSpent));
	if (mnsp->nextSolicit > t)
	    mnsp->nextSolicit -= t;
	else {
	    /* Time to solicit mobility agents on each interface */
	    currTime = CurrentTimeNTPSec();
	    for (i = 0; i < ifaceTableSize; i++) {
		ifaceTable[i].maIfaceSolicitationTime = currTime;
		sendSolicitation(ifaceTable[i].maIfaceAddr);
	    }
	    if (retransmissionPolicy[(mnsp->backoffIdx) + 1] > 0)
	        (mnsp->backoffIdx)++;
	    mnsp->nextSolicit = (retransmissionPolicy[mnsp->backoffIdx] * 1);
	}
    }
}

/* clear state variables for unavailable interfaces */
void mnAgeStateTable(int t) {
	int i, j, found;

	for(i = 1; i < DEFAULT_MAX_INTERFACES; i++) {
		if(*(ISTable[i].ifaceName) != (char) NULL) {
			found = FALSE;
			for(j = 0; j < ifaceTableSize; j++) {
				if(strcmp(ISTable[i].ifaceName, ifaceTable[j].maIfaceName) == 0) {
					found = TRUE;
					break;
				}
			}
			if(!found) {
				memset(ISTable[i].ifaceName, (char) NULL, MAX_IFNAME_LEN);
				ISTable[i].COAddr = 0;

				ISTable[i].currentState.state = UNKNOWN_STATE;
				ISTable[i].currentState.backoffIdx = 0;
				memset(ISTable[i].currentState.ifaceName, (char) NULL, MAX_IFNAME_LEN);
				ISTable[i].currentState.routerAddr = 0;
				ISTable[i].currentState.COAddr = 0;
				ISTable[i].currentState.timeSpent = 0;
				ISTable[i].currentState.nextSolicit = (retransmissionPolicy[0] * 1);
				ISTableSize--;
			}
		}
	}
}

void doPeriodicTask(int t)
{
    int i, IfcAvailable;

    mnUpdateIfaces(t);

    // clear state variables for unavailable interfaces (MSB)
    // mnAgeStateTable(t);

    mnAgeLocalMAs(t);
    /* modified by Uni. of Sheffield - moved into main */
    /*if (movementDetection == 2) updateKnownLocalMaTable();*/
    mnAgeHAs(t);
    mnAgeRegistrations(t);

    // perform state tasks for all interfaces (if no interfaces
    // available, first state element is taken)
    IfcAvailable = FALSE;
    for(i = 1; i < DEFAULT_MAX_INTERFACES; i++) {
    	if(*(ISTable[i].ifaceName) != (char) NULL) {
    		mnStateTasks(&(ISTable[i].currentState), t, i);
		IfcAvailable = TRUE;
	}
    }

    if(!IfcAvailable)
	mnStateTasks(&(ISTable[0].currentState), t, 0);



/* printRegReqTable(regReqTable); */
}

