/*

  Sun Public License Notice

 The contents of this file are subject to the Sun Public License Version
 1.0 (the "License"). You may not use this file except in compliance with
 the License. A copy of the License is available at http://www.sun.com/

 The Original Code is the Sunlabs Mobile IP.
 The Initial Developer of the Original Code is Vipul Gupta.
 Portions created by Vipul Gupta are Copyright (C) 2003.
 All Rights Reserved.

 Contributor(s): Niko A. Fikouras (niko@comnets.uni-bremen.de)
		 Asanga D. Udugama (adu@comnets.uni-bremen.de)

*/

/*
 * An implementation of Mobile-IP as described in RFC 2002.
 * Author: Vipul Gupta (vgupta@eng.sun.com)
 *         Sunil Madhani (smadhani@eng.sun.com)
 * mnState.c: Routines to manage state transitions and related actions.
 */

#pragma ident	"%Z%%M% %I% %E%"

#include <stdio.h>
#include "mip.h"
#include "mobile.h"

extern void delRegReqPtr(RegReqEntry *, int *, RegReqEntry *);
extern void sendRegReq(RegReqEntry *);

/* OS dependent function prototypes */
extern int routeadd(u32, u32, u32, int, char*);
extern int routedel(u32, u32, u32, int);
extern int routeflush(char *);
extern int arpflush();
extern int arpdel(u32);
extern int restoreArp();
extern int disableArp(u32, u32);
extern int configIface(char *, u32, u32);
extern int configCloneForIface(char *, u32, u32);
extern void unconfigCloneForIface(char *);
extern int changeSrcOnDefaultRoute(u32, u32);
extern int sendgarp(char *);
extern IfaceEntry *findIface(char *);

extern int haveColocatedAddr();
extern int getColocatedAddr(int);
extern boolean isValidCOAinAdv(u32, LocalMaEntry *);
extern u32 chooseHAaddr();
extern RegReqEntry *getEmptyRREntry();
extern LocalMaEntry *findMAIdx(u32);

/* Security related stuff */
extern boolean isOutsideProtectedDomain(u32 addr);
extern void enableIPSec(mnState *);
extern void disableIPSec(mnState *);
extern enableReverseTunneling(mnState *, u32);
extern disableReverseTunneling(mnState *, u32);

extern int setFilterForHomeBinding(Filter [], int, FilterOp [], int *, char *);
extern int setFilterForNewBinding(Filter [], int, FilterOp [], int *, char *, int, RegReqEntry []);
extern int ifaceInColocated(InterfaceState []);
extern int routeflushHomeNW();
extern int routeflushDefault();

extern u32  mnMoveFromFAToHACnt;
extern u32  mnMoveFromHAToFACnt;
extern u32  mnMoveFromFAToFACnt;

extern char currentTime[], addr1[], addr2[];

extern int mipdebug;
extern int periodicInterval;
extern int agentSolicitThreshold;
extern int agentExpireThreshold;
extern int ifacePollThreshold;
extern int regLifetime;
extern int regRoundTrip;
extern int regFlags;
extern int haCooloffPeriod;
extern int faCooloffPeriod;
extern int IDfreshnessSlack;
extern int solicitOnBcast;
extern int retransmissionPolicy[];
extern float renewalPolicy[];
/* the mn's movement detection method, either LCS or ECS */
/* modified by Uni. of Sheffield.			 */
extern int movementDetection;

/* Determines whether the MN gives priority to HA's */
/* modified by Uni. of Sheffield.			 */
extern int HApriority;

/* The mobile node's address, netmask and default routers at home */
extern u32 mnHomeAddr;
extern u32 mnHomeNetmask;
extern u32 mnHomeBcastAddr;

// currnetState is part of ISTable; changed to handle MSB
//extern mnState currentState;
extern InterfaceState ISTable[DEFAULT_MAX_INTERFACES];

extern IfaceEntry ifaceTable[];
extern int ifaceTableSize;

extern HomeAgentEntry haTable[];
extern int haTableSize;

/* List of default routers on the home network */
extern u32 homeRouterTable[];
extern int homeRouterTableSize;

/* List of local Mobility Agents (considered active) */
extern LocalMaEntry localMaTable[];
extern int localMaTableSize;

/* List of known local Mobility Agents (considered active) */
/* Used by the ECS movement detection method - modified by */
/* Uni. Sheffield					   */
extern LocalMaEntry *knownLocalMaTable[DEFAULT_MAX_MOBILITY_AGENTS];

/* List of registration requests (both accepted and pending) */
extern RegReqEntry regReqTable[];
extern int regReqTableSize;
extern u32 mnRegRequestsAcceptedCnt;

static int ifacepolltimer;

extern boolean isReverseTunnelSet;
static boolean completedColocatedAction = FALSE;
static u32 homeAgentAddr;  /* stores the target of a reverse tunnel */

/* to store the filters that have been defined for this mobile node*/
extern Filter mnFilters[MAX_FILTERS]; 				   // data for each filter
extern int mnFilterCount; 					   // the number of filters
extern FilterOp mnFilterOpTable[MAX_FILTERS];			   // the grid that holds the operations
								   // to be done for the filters
extern int mnFilterOpCount;					   // operations count

/*
 * Updates the corresponding transition couter
 * whenever there is a change of mobile node state
 */
void updateTransitionCounter(short currentState, short pastState)
{
    if (pastState == FOREIGN_STATE) {
	if ((currentState == FOREIGN_STATE)||
            (currentState == COLOCATED_STATE))
	    mnMoveFromFAToFACnt++;
	else if (currentState == HOME_STATE)
	    mnMoveFromFAToHACnt++;
    } else if (pastState == HOME_STATE) {
	if ((currentState == FOREIGN_STATE) ||
            (currentState == COLOCATED_STATE))
	    mnMoveFromHAToFACnt++;
    }
}

void undoState(mnState *mnsp)
{
    char tmp[30];
    char iface[MAX_IFNAME_LEN];
    int  errcode, i;

    /* Do actions based on state we are leaving */
    switch (mnsp->state) {
        case HOME_STATE :   
	    /* Delete the default route entries used on the home network 
	     *
	     * TODO: Need to be smarter if the default router was added
	     * automatically, i.e. homeRouterTableSize is zero and we use
	     * discovery to find default routers.
	     */
	    for (i = 0; i < homeRouterTableSize; i++) {		
	        if ((errcode = routedel(0, 0, homeRouterTable[i], 1)) < 0 ) {
		    mip2dbg(("Error in routedel default %s ... %s\n",
			     ntoa(homeRouterTable[i], addr1),
			     err2str(errcode)));
		}
	    }
	    break; 

        case FOREIGN_STATE :
             /** 
	     * According to Qualcomm researchers, Lorayne Wittie
	     * (lwitte@qualcomm.com) and Ramesh Ramaswamy  
	     * (rameshr@qualcomm.com): 
	     * "To support mobile reactivation when a call goes  
	     * dormant (this conserves over-the-air (OTA) traffic 
	     * channel) we commented out the removal of the default 
	     * route when leaving the FOREIGN state. For our envir
	     * this is typically a default route to ppp0."
	     *  
	     * If you operate in a similar environment, you may
	     * wish to do the same. 
	     *  
	     * Note, however, that for non-PPP situations
	     * commenting will almost certainly cause problems. 
	     **/
	  /* Delete default route which goes through FA */
            if ((errcode = routedel(0, 0, mnsp->routerAddr, 1)) < 0) {
	        mip2dbg(("Error in routedel 0.0.0.0 %s ... %s\n",
			 ntoa(mnsp->routerAddr, addr1), err2str(errcode)));
            }

            /* Delete the arp entry set for the FOREIGN address */
            if ((errcode = arpdel(mnsp->routerAddr)) < 0) {
		mip2dbg(("Error while deleting arp entry for %s ... %s\n",
			ntoa(mnsp->routerAddr, addr1), err2str(errcode)));
	    }

	    /* 
             * Delete the routing table entries set when it went 
	     * in FOREIGN state
	     */	
            if ((errcode = routedel(mnsp->routerAddr,
				    inet_addr("255.255.255.255"),
				    mnHomeAddr, 0)) < 0) {
	       mip2dbg(("Error in routedel %s %s ... %s\n",
			 ntoa(mnsp->routerAddr, addr1),
			 ntoa(mnHomeAddr, addr2), err2str(errcode)));
            }
	    
	    /* Restore Normal ARP Behaviour */	
	    restoreArp();

	    /*
	     * It's not a good idea to bring down the interface because
	     * the iface may now be in use for a colocated address.
	     */

	    break;

        case COLOCATED_STATE:
	    if (completedColocatedAction == TRUE) {
	        if (isOutsideProtectedDomain(mnsp->COAddr) &&
		    isReverseTunnelSet) {
		    errcode = disableReverseTunneling(mnsp, homeAgentAddr);
		    if (errcode < 0) {
		        mip1dbg(("Problem while disabling reverse " \
				 "tunnel ... %s.\n",
				 err2str(errcode)));
		    } else {
		        mip2dbg(("Disabled reverse tunnel to home agent " \
				 "%s\n", ntoa(homeAgentAddr, addr1)));
		    }
		    isReverseTunnelSet = FALSE;
		} else {	
		    /* Delete default route entry */
		    if ((errcode = routedel(0, 0, mnsp->routerAddr, 1))< 0) {
		        mip2dbg(("Error in routedel default %s  ... %s\n",
				 ntoa(mnsp->routerAddr, addr1), 
				 err2str(errcode)));
		    }

		    /*
		     * Delete route to the default router through the iface
		     * with our home address.
		     */	
		    if ((errcode = routedel(mnsp->routerAddr, 
					    inet_addr("255.255.255.255"),
					    mnHomeAddr, 0)) < 0) {
		      mip2dbg(("Error in routedel %s %s ... %s\n",
			       ntoa(mnsp->routerAddr, addr1),
			       ntoa(mnHomeAddr, addr2), err2str(errcode)));
		    }
           
		    /* Unconfigure the clone interface */ 
		    unconfigCloneForIface(mnsp->ifaceName); 
		}
		completedColocatedAction = FALSE;
	    }

            /*
             * Disable SKIP if the COA indicates that this mobile node
             * was outside protected network.
             */
            if (isOutsideProtectedDomain(mnsp->COAddr)) {
	        mip2dbg(("Disabling IPSec\n"));
		disableIPSec(mnsp);
	    }

	    mip3dbg(("Deleting proxy ARP entry for %s.\n",
		     ntoa(mnsp->COAddr, addr1)));
	    if ((errcode = arpdel(mnsp->COAddr)) < 0) {
	        mip1dbg(("arpdel failed ... %s\n", err2str(errcode)));
	    }

	    /* Restore regular ARP processing */
	    restoreArp();
	    break;

        case UNKNOWN_STATE :
	    break;

        case ISOLATED_STATE :
           /* can restore iface polling interval */
           ifacePollThreshold = ifacepolltimer;
           break;
    }
    return;
}

void updateState(mnState *mnsp, int newState, u32 MAidx, u32 ifaceIdx, int ISSeq)
{

    //undoState(mnsp);
    updateTransitionCounter(newState, mnsp->state);
    mnsp->state = newState;

    /* Do state-specific actions for the new state we are entering */
    switch (newState) {
        case HOME_STATE :
	    strcpy(mnsp->ifaceName, localMaTable[MAidx].mnAdvIncomingIface);
            /* modified by Uni. of Sheffield */
            mnsp->COAddr = 0;
            break;

	case FOREIGN_STATE :
	    strcpy(mnsp->ifaceName, localMaTable[MAidx].mnAdvIncomingIface);
	    mnsp->routerAddr = localMaTable[MAidx].mnAdvSourceAddr;
	    mnsp->COAddr = localMaTable[MAidx].mnAdvCOAList[0];
	    break;

	case COLOCATED_STATE :
	    strcpy(mnsp->ifaceName, ifaceTable[ifaceIdx].maIfaceName);
	    mnsp->COAddr = ifaceTable[ifaceIdx].maIfaceAddr;
	    mnsp->routerAddr = ifaceTable[ifaceIdx].maIfaceOtherAddr;
		fprintf(stderr, "routerAddr = %s\n", ntoa(mnsp->routerAddr, addr1));
	    break;

	case UNKNOWN_STATE :
	    mnsp->backoffIdx = 0;
	    mnsp->nextSolicit = 0;
	    break;

	case ISOLATED_STATE :
	    break;
    }
    mnsp->timeSpent = 0;
    ISTable[ISSeq].COAddr = mnsp->COAddr;
}

/*
 * Replace entry i and all other accepted registrations that use the same
 * HA with a "deregister-all" request to that HA. When finding other such
 * registrations, we only need to look past entry i. Amongst all entries
 * using the same HA, keep the highest mnRegTimeRemaining value in entry
 * i. Set the mnRegCOA field to mnHomeAddr, mnRegTimeRequested to zero,
 * mnRegAttempt to 0, mnRegNextTime to (retransmissionPolicy[0]*regRoundTrip),
 * mnRegStatus to Pending. Return a pointer to the entry thus formed. This
 * should get rid of all accepted entries using the same HA.
 */
RegReqEntry *mkAcceptedToPendingDereg(int idx, int ISSeq)
{
    int i;
    int timeleft;

    /* Get the max. mnRegTimeRemaining */
    timeleft = regReqTable[idx].mnRegTimeRemaining;
    for (i = (idx +1); i < DEFAULT_MAX_REG_REQ; i++){
	if ((regReqTable[i].mnRegStatus != REG_UNUSED)
	    && (regReqTable[i].mnRegHAaddr == regReqTable[idx].mnRegHAaddr)
	    && (strcmp(regReqTable[i].mnRegIfaceName, ISTable[ISSeq].ifaceName) == 0)){
	    if (timeleft < regReqTable[i].mnRegTimeRemaining)
	        timeleft = regReqTable[i].mnRegTimeRemaining;

	    delRegReqPtr(regReqTable, &regReqTableSize, &(regReqTable[i]));
	}
    }

    memset((char *)regReqTable[idx].mnRegAgentHWaddr, 0, MAX_HWADDR_LEN);
    regReqTable[idx].mnRegAgentAddr =  regReqTable[idx].mnRegHAaddr;
    regReqTable[idx].mnRegCOA = mnHomeAddr;
    regReqTable[idx].mnRegTimeRequested = 0;
    regReqTable[idx].mnRegTimeRemaining = timeleft;
    regReqTable[idx].mnRegAttemptIdx = 0;
    regReqTable[idx].mnRegNextTime = retransmissionPolicy[0]*regRoundTrip;
    regReqTable[idx].mnRegFlags = 0;
    regReqTable[idx].mnRegStatus = REG_PENDING;
    regReqTable[idx].mnRegCOAIsLocal = FALSE; /* ??? */

    return(&(regReqTable[idx]));
}

/* Things to do after a fresh transition to HOME_STATE */
void startHomeAction(mnState *mnsp, int ISSeq)
{
    int i;
    RegReqEntry *rrep;
    short errcode ;

    /* Flush ARP entries */
    if ((errcode = arpflush()) < 0) {
	mip1dbg(("Error in arpflush ... %s\n", err2str(errcode)));
    }

    /* Flush routes */
    if ((errcode = routeflush(ISTable[ISSeq].ifaceName)) < 0) {
	mip2dbg(("Error in routeflush ... %s\n", err2str(errcode)));
    }


    /* Flush route for home NW */
    if (routeflushHomeNW() < 0) {
	mip2dbg(("Error in routeflushHomeNW ... \n"));
    }

    /* Flush default only if no ifc in colocated*/
    if(routeflushDefault() < 0) {
		mip2dbg(("Error in routeFlushDefault ... \n"));
    }


    /* Assign home address to the interface stored in state structure */
    if ((errcode = configIface(mnsp->ifaceName, mnHomeAddr,
			       mnHomeNetmask)) < 0) {
	mip2dbg(("Error in configIface ... %s\n", err2str(errcode)));
    }

    /*
     * Add multicast route through the interface
     * TODO: For now our routeadd cannot deal with netmasks
     *       but we hope that the system knows the correct
     *       netmask to use for 224.0.0.0.
     *       Ideally we should do the same on foreign networks also.
     */
    if ((errcode = routeadd(inet_addr("224.0.0.0"),
			    inet_addr("240.0.0.0"), mnHomeAddr, 0, mnsp->ifaceName)) < 0) {
        mip2dbg(("Could not add routing entry for 224.0.0.0 ... %s\n",
		 err2str(errcode)));
    }

    /* Add default route */
    /*
     * TODO: if RouterTableSize is zero, initiate OS-specific
     * router discovery.
     */
    for (i = 0; i < homeRouterTableSize; i++) {
	if ((errcode = routeadd(0, 0, homeRouterTable[i], 1, mnsp->ifaceName)) < 0) {
	    mip2dbg(("Error in (routeAdd 0 %s) ... %s\n",
		     ntoa(homeRouterTable[i], addr1), err2str(errcode)));
        }
    }

    /* if (mipdebug > 4)
     *    system("netstat -nr");
     */

    /* Send a gratuitous ARP for itself through the above iface */
    if ((errcode = sendgarp(mnsp->ifaceName)) < 0) {
 	mip1dbg(("Error in sendgarp ... %s\n", err2str(errcode)));
    }

    /* Cancel all PENDING registrations */
    for (i = 0; i < DEFAULT_MAX_REG_REQ; i++) {
	if ((regReqTable[i].mnRegStatus == REG_PENDING)
		&& (strcmp(ISTable[ISSeq].ifaceName,regReqTable[i].mnRegIfaceName) == 0)) {
	    delRegReqPtr(regReqTable, &regReqTableSize, &(regReqTable[i]));
	}
    }

    /*
     * Send a deregistration for each accepted registration. Figure out
     * all HAs that have registered us and send each one a "deregister all"
     * message. New pending entries are created as a result. If we have
     * multiple registrations with an HA, we choose the largest remaining
     * registration lifetime to create the pending entry.
     */
    for (i = 0; i < DEFAULT_MAX_REG_REQ; i++) {
	if (regReqTable[i].mnRegStatus == REG_UNUSED) continue;
	/*
	 * This registration is ACCEPTED. Replace all accepted entries
	 * using the same HA as for entry i by a pending "deregister-all"
	 * request to that HA.
	 */
	rrep = mkAcceptedToPendingDereg(i, ISSeq);
	sendRegReq(rrep);
    }
}

/*
 * Pick an unused entry in the regReqTable. Create a new registration request
 * based on state information pointed to by mnsp. The state information is
 * also helpful in deciding whether the destination should be a foreign agent
 * (in FOREIGN state) or a homeagent (in COLOCATED state).
 *
 * Use mnHomeBcastAddr if no HA is alive. Return a pointer to this
 * newly created entry.
 */
RegReqEntry *mkNewRegReq(mnState *mnsp, int ISSeq)
{
    RegReqEntry *rrep;
    LocalMaEntry *lmaeIdx;
    HomeAgentEntry *haeIdx;
    IfaceEntry *ifaceIdx;

    if ((rrep = getEmptyRREntry()) == (RegReqEntry *) NULL) {
        mip0dbg(("Reg. Req. Table full\n"));
        return(NULL);
    }

    // set the interface for which this registration is valid
    strcpy(rrep->mnRegIfaceName, ISTable[ISSeq].ifaceName);

    rrep->mnRegHAaddr = chooseHAaddr();

    if (mnsp->state == FOREIGN_STATE) {
        if ((lmaeIdx = findMAIdx(mnsp->routerAddr)) == (LocalMaEntry *) NULL) {
            mip0dbg(("Mobility Agent Not Present\n"));
	    memset((char *)rrep, 0, sizeof(RegReqEntry));
	    rrep->mnRegStatus = REG_UNUSED;
	    return(NULL);
        }
	memcpy(rrep->mnRegAgentHWaddr,
	      lmaeIdx->mnAdvSrcHWaddr, MAX_HWADDR_LEN);
	rrep->mnRegAgentAddr = lmaeIdx->mnAdvSourceAddr;
        if (lmaeIdx->mnAdvMaxRegLifetime < regLifetime)
            rrep->mnRegTimeRequested = lmaeIdx->mnAdvMaxRegLifetime;
	else
	    rrep->mnRegTimeRequested = regLifetime;
	rrep->mnRegFlags = regFlags;
        rrep->mnRegCOAIsLocal = FALSE;

    } else {
        if (mnsp->state != COLOCATED_STATE)
	    mip0dbg(("Something is wrong in mkNewRegReq.\n"));
	memset((char *)rrep->mnRegAgentHWaddr, 0, MAX_HWADDR_LEN);
	rrep->mnRegAgentAddr = rrep->mnRegHAaddr;
        rrep->mnRegTimeRequested = regLifetime;
	rrep->mnRegFlags = regFlags | REG_DECAPSULATION_BY_MN;
        rrep->mnRegCOAIsLocal = TRUE;
    }

    rrep->mnRegCOA = mnsp->COAddr;
    rrep->mnRegAttemptIdx = 0;
    rrep->mnRegNextTime = retransmissionPolicy[0]*regRoundTrip;
    rrep->mnRegTimeRemaining = rrep->mnRegTimeRequested;
    rrep->mnRegStatus = REG_PENDING;
    regReqTableSize++;

    if (rrep->mnRegHAaddr == mnHomeBcastAddr) {
        rrep->mnRegTimeRemaining = DEFAULT_DISCOVERY_INTERVAL;
    }

    return(rrep);
}

/* Things to do after a fresh transition to a (new) foreign network. */
void startForeignAction(mnState *mnsp, LocalMaEntry *lmaep, int ISSeq)
{
    int i;
    int errcode;
    RegReqEntry *rrep;

    /*
     * Disable normal ARP behavior in favour of special mobile node
     * ARP processing.
     */
/*
    disableArp(mnHomeAddr, mnsp->routerAddr);
*/

    /* Flush the ARP table */
    if ((errcode = arpflush()) < 0) {
	mip2dbg(("Error in arpflush ... %s\n", err2str(errcode)));
    }

    /* Assign home address to the interface stored in state structure */

    if ((errcode = configIface(mnsp->ifaceName, mnHomeAddr,
			       mnHomeNetmask)) < 0) {
	mip2dbg(("Error in configIface ... %s\n", err2str(errcode)));
    }

    /* Flush the routing table */

    if ((errcode = routeflush(ISTable[ISSeq].ifaceName)) < 0) {
	mip2dbg(("Error in routeflush ... %s\n", err2str(errcode)));
    }





    /* Flush route for home NW */
    if (routeflushHomeNW() < 0) {
	mip2dbg(("Error in routeflushHomeNW ... \n"));
    }

    /* Flush default only if no ifc in colocated ADU*/
    //if(ifaceInColocated(ISTable) < 0) {
    //	if(routeflushDefault() < 0)
    //		mip2dbg(("Error in routeFlushDefault ... \n"));
    //}





    /*
     * Add host specific route to the foreign agent through above interface
     * and default route through the foreign agent.
     */
    if ((errcode = routeadd(mnsp->routerAddr, inet_addr("255.255.255.255"),
			    mnHomeAddr, 0, mnsp->ifaceName)) < 0) {
	mip2dbg(("Error in routeadd %s %s ... %s\n",
		ntoa(mnsp->routerAddr, addr1),
		ntoa(mnHomeAddr, addr2),
		err2str(errcode)));
    }

    //if(ifaceInColocated(ISTable) < 0) { ADU
	    if ((errcode = routeadd(0, 0, mnsp->routerAddr, 1, mnsp->ifaceName)) < 0) {
		mip2dbg(("Error in routeadd default %s ... %s\n",
			 ntoa(mnsp->routerAddr, addr1),
			 err2str(errcode)));
	    }
    //}

    /*
     * Add an ARP entry for the foreign agent. Note that for
     * advertisements received on a point to point link, we
     * wouldn't have the source hardware address. Arpadd is
     * written to return quietly in that case.
     */
    if ((errcode = arpadd(lmaep->mnAdvSourceAddr,
			  lmaep->mnAdvSrcHWaddr, "")) < 0 ) {
        mip2dbg(("Error in arpadd for %s ... %s\n",
		 ntoa(lmaep->mnAdvSourceAddr, addr1),
		 err2str(errcode)));
    }
    /*
    if (mipdebug > 4) {
       system("netstat -nr");
       system("netstat -np");
    }
    */
    /*
     * Cancel all PENDING registrations
     * TODO: Does it help to keep discovery requests? There may be a slight
     * benfit to keeping the discovery request and simply overwriting the
     * mnRegCOA field with the new COA value.
     */
    for (i = 0; i < DEFAULT_MAX_REG_REQ; i++) {
	if ((regReqTable[i].mnRegStatus == REG_PENDING)
		&& (strcmp(ISTable[ISSeq].ifaceName,regReqTable[i].mnRegIfaceName) == 0)) {
	    delRegReqPtr(regReqTable, &regReqTableSize, &(regReqTable[i]));
	}
    }

    /*
     * Mark any accepted registrations that use the same COA as renewable
     * and others as nonrenewable.
     */
    for (i = 0; i < DEFAULT_MAX_REG_REQ; i++) {
	if (regReqTable[i].mnRegStatus == REG_UNUSED) continue;
	/* This registration is ACCEPTED. */
	if( (regReqTable[i].mnRegCOA == mnsp->COAddr) &&
		(strcmp(regReqTable[i].mnRegIfaceName, ISTable[ISSeq].ifaceName) == 0) ) {
	    regReqTable[i].mnRegStatus = REG_ACCEPTED_RENEWABLE;
	}

	//if (regReqTable[i].mnRegCOA != mnsp->COAddr)
	//	regReqTable[i].mnRegStatus = REG_ACCEPTED_NONRENEWABLE;
	//else
	//    regReqTable[i].mnRegStatus = REG_ACCEPTED_RENEWABLE;
	//
    }

    /*
     * Also create a new pending registration request with the given COA
     * (fill out mnHomeBcastAddr if no HAs are alive, thus initiating HA
     * discovery). Later when we find an HA, we can clone the discovery
     * request and replace mnHomeBcastAddr with the HA's address.
     */
    if ((rrep = mkNewRegReq(mnsp, ISSeq)) == NULL) {
	mip1dbg(("Could not make new registration request for FA's COA %s.\n",
		 ntoa(mnsp->COAddr, addr1)));
    } else {
		sendRegReq(rrep);
    }
}

/*
 * Things to do immediately after a fresh transition to a (new)
 * colocated state.
 */
void startColocatedAction(mnState *mnsp, int ISSeq)
{
    int i;
    char cloneIface[MAX_IFNAME_LEN];
    short errcode;
    RegReqEntry *rrep;
    IfaceEntry *ifaceIdx;

    /*
     * Disable normal ARP behavior in favour of special mobile node
     * ARP processing. NOTE: Mobile node should still continue to answer
     * ARP requests for its colocated address.
     */
    disableArp(mnHomeAddr, mnsp->routerAddr);

    /*
     * Flush the ARP table.
     * TODO: But we now need a way to add an ARP
     * entry for the router on the foreign network.
     */
    if ((errcode = arpflush()) < 0) {
       mip2dbg(("Error in arpflush ... %s\n", err2str(errcode)));
    }

    /* Flush route for home NW */
    if (routeflushHomeNW() < 0) {
	mip2dbg(("Error in routeflushHomeNW ... \n"));
    }

    /*
     * We need to recreate the proxy ARP entry for the care-of address
     * (if appropriate).
     */

    if ((ifaceIdx = findIface(mnsp->ifaceName)) == NULL) {
        mip0dbg(("Something is wrong with configNetColocated()\n"));
    } else {
        if ((ifaceIdx->maIfaceFlags & IFACE_NOARP) == 0) {
	    mip2dbg(("Adding proxy ARP entry for %s (Iface %s)\n",
		     ntoa(mnsp->COAddr, addr1), mnsp->ifaceName));
	    if ((errcode = arpadd(mnsp->COAddr, ifaceIdx->maIfaceHWaddr,
				  "pub")) < 0 )
	        mip2dbg(("arpadd failed ... %s\n", err2str(errcode)));
	}
    }

    if ((errcode = routeadd(mnsp->routerAddr, inet_addr("255.255.255.255"),
			    mnHomeAddr, 0, mnsp->ifaceName)) < 0) {
	mip2dbg(("Error in routeadd %s %s ... %s\n",
		ntoa(mnsp->routerAddr, addr1),
		ntoa(mnHomeAddr, addr2),
		err2str(errcode)));
    }


    /* set host route to home ADU*/
    if ((errcode = routeadd(chooseHAaddr(), inet_addr("255.255.255.255"),
			    mnsp->routerAddr, 0, mnsp->ifaceName)) < 0) {
	mip2dbg(("Error in routeadd %s %s ... %s\n",
		ntoa(mnsp->routerAddr, addr1),
		ntoa(mnHomeAddr, addr2),
		err2str(errcode)));
    }



    if ((errcode = routeadd(0, 0, mnsp->routerAddr, 1, mnsp->ifaceName)) < 0) {
	mip2dbg(("Error in routeadd default %s ... %s\n",
		 ntoa(mnsp->routerAddr, addr1),
		 err2str(errcode)));
    }

    /*
     * Enable IPSec if the current COA is outside MN's protected
     * network. This will allow the registration requests to get
     * through intervening firewalls.
     */
    if (isOutsideProtectedDomain(mnsp->COAddr)) {
        mip2dbg(("COA %s is outside protected domain. Enabling IPSec.\n",
		 ntoa(mnsp->COAddr, addr1)));
	enableIPSec(mnsp);
    }

    /* Cancel pending registrations. */
    for (i = 0; i < DEFAULT_MAX_REG_REQ; i++) {
	if ((regReqTable[i].mnRegStatus == REG_PENDING)
	     && (strcmp(ISTable[ISSeq].ifaceName, regReqTable[i].mnRegIfaceName) == 0))  {
	    delRegReqPtr(regReqTable, &regReqTableSize, &(regReqTable[i]));
	}
    }

    /*
     * Mark any accepted registrations that use the same COA as renewable
     * and others as nonrenewable.
     */
    for (i = 0; i < DEFAULT_MAX_REG_REQ; i++) {
	if (regReqTable[i].mnRegStatus == REG_UNUSED) continue;
	/* This registration is ACCEPTED. */
	if( (regReqTable[i].mnRegCOA == mnsp->COAddr) &&
		(strcmp(regReqTable[i].mnRegIfaceName, ISTable[ISSeq].ifaceName) == 0) ) {
	    regReqTable[i].mnRegStatus = REG_ACCEPTED_RENEWABLE;
	}

	//if (regReqTable[i].mnRegCOA != mnsp->COAddr)
	//    regReqTable[i].mnRegStatus = REG_ACCEPTED_NONRENEWABLE;
	//else
	//    regReqTable[i].mnRegStatus = REG_ACCEPTED_RENEWABLE;
    }

    /*
     * Also create a new pending registration request with the given COA
     * (fill out mnHomeBcastAddr if no HAs are alive, thus initiating HA
     * discovery). Later when we find an HA, we can clone the discovery
     * request and replace mnHomeBcastAddr with the HA's address.
     * TODO: Hmm ... it seems like we could now have registration requests
     * going out because of the reneable entry and also because of
     * the new entry.
     */
    if ((rrep = mkNewRegReq(mnsp, ISSeq)) == NULL) {
	mip1dbg(("Could not make new registration request for colocated " \
                 "COA %s.\n", ntoa(mnsp->COAddr, addr1)));
    } else {
	sendRegReq(rrep);
    }
}

/* Called when a registration w/ colocated address is successful */
void completeColocatedAction(mnState *mnsp, RegReqEntry *rrep)
{
    char cloneInterface[MAX_IFNAME_LEN];
    short errcode;

    /*
     * Hopefully, ARP is already disabled and the ARP table was properly
     * flushed when we first entered colocated state.
     */

    /*
     * Add reverse tunneling if the current COA indicates that this
     * mobile node is outside protected network.
     */
    if (isOutsideProtectedDomain(mnsp->COAddr)) {
        if ((errcode = enableReverseTunneling(mnsp,
					      rrep->mnRegHAaddr)) < 0) {
	    mip1dbg(("Error while enabling reverse tunnel ... %s\n",
		     err2str(errcode)));
        } else {
	    homeAgentAddr = rrep->mnRegHAaddr;
	    isReverseTunnelSet = TRUE;
	    mip2dbg(("Established reverse tunnel to home agent %s.\n",
		     ntoa(homeAgentAddr, addr1)));
	}
    } else {
         /*
          * Assign home address to the logicalClone of interface stored in
          * state structure. The assumption is that configCloneIface
	  * is also responsible for deleting any spurious routing entries
	  * that may be added as part of bringing up the clone Iface.
	  */
         if ((errcode = configCloneForIface(mnsp->ifaceName, mnHomeAddr,
					    mnHomeNetmask)) < 0) {
	     mip2dbg(("configCloneForIface failed ... %s\n", err2str(errcode)));
         }

         if ((errcode = routeadd(mnsp->routerAddr,
				 inet_addr("255.255.255.255"),
				 mnHomeAddr, 0, mnsp->ifaceName))< 0){
             mip2dbg(("Error in routeadd %s %s ... %s\n",
		      ntoa(mnsp->routerAddr, addr1),
		      ntoa(mnHomeAddr, addr2), err2str(errcode)));
         }

         if ((errcode = changeSrcOnDefaultRoute(mnHomeAddr,
						mnsp->routerAddr)) < 0) {
             mip2dbg(("Error in changeSrcOnDefaultRoute through %s ... %s\n",
		      ntoa(mnsp->routerAddr, addr1),  err2str(errcode)));
         }
    }

    /*
      if (mipdebug > 4)
        system("netstat -nr");
     */
    completedColocatedAction = TRUE;
}

/*
 * We could try to configure ourselves for the home network. e.g. use
 * configIface and create routing table entries for the home network
 * default routes.
 */
void startUnknownAction()
{
;
}

/* We could try to decrease the iface polling interval here. */
void startIsolatedAction()
{
    ifacepolltimer = ifacePollThreshold;
    /* Leave this out for now.
     *   if ((ifacePollThreshold / 2) > 1)
     *       ifacePollThreshold /= 2;
     */
}

/*
 * Decide if a state transition is needed and if so, all the details.
 * Also maintain all mnMoveXXtoYYCnt counters.
 */
void decideStateTransition(mnState *mnsp, int ISSeq)
{
    int i,k;
    u32 tmpMnAdvLifetime = 0;
    int homeMAidx = (-1);
    int MAidx = (-1);
    int bestMAidx = (-1);
    int knownMAidx = (-1);
    int ifaceIdx = (-1);
    int UnknownMAidx = (-1);
    int choice;
    unsigned char state;
	char cmd[50];

    LocalMaEntry *lmaep;

    if(!(localMAAvailable(ISTable[ISSeq].ifaceName, localMaTable))) { // check if agents available for this interface
    //if (localMaTableSize <= 0) {
      if ((ifaceIdx = getColocatedAddr(ISSeq)) >= 0) {
         if ((mnsp->state != COLOCATED_STATE) ||
	    (mnsp->COAddr != ifaceTable[ifaceIdx].maIfaceAddr)) {
	   /* Update transition counter. Print message.*/
	    updateState(mnsp, COLOCATED_STATE, 0, ifaceIdx, ISSeq);
	    mip3dbg(("\nEntering COLOCATED_STATE on iface %s [COA %s].\n",
		    ifaceTable[ifaceIdx].maIfaceName,
		    ntoa(ifaceTable[ifaceIdx].maIfaceAddr, addr1)));

          /*
	   * Can complain about a colocated care-of address that belongs
	   * on the home network.
	   * TODO: We need to be careful if PPP gave us
	   * an address on the home network without really placing us on
	   * that net, in the sense of having direct link connectivity
	   * to that net.
	   */
           if ((ifaceTable[ifaceIdx].maIfaceAddr & mnHomeNetmask)
	         == mnHomeNetmask) {
	       mip1dbg(("\nColocated address %s belongs to Home Network\n",
			ntoa(ifaceTable[ifaceIdx].maIfaceAddr, addr1)));
	   }

           // set filter retransmission with the new binding
	   // as Colocated CoA
	   setFilterForNewBinding(mnFilters, mnFilterCount, mnFilterOpTable,
		&mnFilterOpCount, ISTable[ISSeq].ifaceName, ACT_REPLACE, regReqTable);

           startColocatedAction(mnsp, ISSeq);
	   applyReverseFilters(mnFilters, ISTable, ISTable[ISSeq].ifaceName);
	   sendInfoToAllGUIs();
         }
	 return;
      }

      if(isValidInterface(ISTable[ISSeq].ifaceName, ifaceTable, ifaceTableSize)) {
//    if (ifaceTableSize > 0) {
        if (mnsp->state != UNKNOWN_STATE) {
	   mip3dbg(("\nEntering Unknown state.\n"));
	   updateState(mnsp, UNKNOWN_STATE, 0, 0, ISSeq);
	   startUnknownAction();
	   applyReverseFilters(mnFilters, ISTable, ISTable[ISSeq].ifaceName);
	   sendInfoToAllGUIs();
        }
        return;
      }

      if (mnsp->state != ISOLATED_STATE) {
        mip3dbg(("\nEntering Isolated state.\n"));
        updateState(mnsp, ISOLATED_STATE, 0, 0, ISSeq);
        startIsolatedAction();
	applyReverseFilters(mnFilters, ISTable, ISTable[ISSeq].ifaceName);
	sendInfoToAllGUIs();
      }
      return;
    }

    /*
     * Make a pass over localMaTable and pick the two best mobility agents:
     * one from those on the home network and another from those not on the
     * home network and place their indices in homeMAidx and foreignMAidx
     * respectively.
     */

    for (i = 0; i < DEFAULT_MAX_MOBILITY_AGENTS; i++) {
	if ((localMaTable[i].mnAdvSourceAddr == 0)
			|| (strcmp(localMaTable[i].mnAdvIncomingIface,ISTable[ISSeq].ifaceName) != 0))
		continue;
	if (((localMaTable[i].mnAdvSourceAddr ^ mnHomeAddr) &
	    	 		mnHomeNetmask) == 0) {
            	if (homeMAidx < 0) {
              	 	homeMAidx = i;
               	 	if (mnsp->state == HOME_STATE) MAidx = i;
             	}
        }
	if (isValidCOAinAdv(mnsp->COAddr, &(localMaTable[i])))
		MAidx = i;
       	if((localMaTable[i].mnAdvRemainingLifetime)>tmpMnAdvLifetime) {
        	bestMAidx = i;
        	tmpMnAdvLifetime = localMaTable[i].mnAdvRemainingLifetime;
       	}
       	if((movementDetection==ECS)&&
       			(findKnownLocalMA(localMaTable[i].mnAdvSourceAddr)<0)) {
        	UnknownMAidx=i;
	}
    }

    if ((movementDetection == LCS) || (movementDetection >= HCS)) {
       if ((HApriority == 1) && (homeMAidx >= 0) && (homeMAidx != MAidx)) {
          MAidx = homeMAidx;
          goto handoff;
       }
       if (MAidx >= 0) return;
       MAidx = bestMAidx;
       goto handoff;
    }

    if (movementDetection == ECS) {
       if (UnknownMAidx >= 0) {
          if (localMaTable[UnknownMAidx].mnAdvPrefixList[0]!=0){

                /* Assumption : The iteration cycle is so small
                   that there will never be more than one new MAs
                   in the localMATable. */
                /* Modified by Uni. of Sheffield ;) */

             for (i = 0; i < DEFAULT_MAX_MOBILITY_AGENTS; i++) {
                 if(knownLocalMaTable[i] == NULL)
		 	continue;
                 if((knownLocalMaTable[i]->mnAdvRouterList[0] &
                      knownLocalMaTable[i]->mnAdvPrefixList[0])
                   		== (localMaTable[UnknownMAidx].mnAdvRouterList[0] &
                      			localMaTable[UnknownMAidx].mnAdvPrefixList[0]))
		      continue;
             }
          }
          knownMAidx = newKnownLocalMA();
          knownLocalMaTable[knownMAidx] = &(localMaTable[UnknownMAidx]);
          if (i != DEFAULT_MAX_MOBILITY_AGENTS) return;
          MAidx = UnknownMAidx;
       }
       else if (MAidx >= 0)
       	  return;
/* This should go before the search for an unknown MA */
       else
          MAidx = bestMAidx;
       goto handoff;
    }

    handoff:
//			fprintf(stderr, "\nlocalMaTableSize: %d\n",localMaTableSize);
//			fprintf(stderr, "UnknownMAidx: %d\n",UnknownMAidx);
//			fprintf(stderr, "bestMAidx: %d\n",bestMAidx);
//			fprintf(stderr, "homeMAidx: %d\n",homeMAidx);
//			fprintf(stderr, "MAidx: %d\n",MAidx);

    if (homeMAidx == MAidx)
    	state = HOME_STATE;
    else
    	state = FOREIGN_STATE;
    /* Update state & transition counter */
    updateState(mnsp, state, MAidx, 0, ISSeq);

    if (state == HOME_STATE) {
        mip3dbg(("\nEntering HOME_STATE on iface %s.\n",mnsp->ifaceName));

	// set filter removal with the new deregistration
	// when comming back to Home
	setFilterForHomeBinding(mnFilters, mnFilterCount, mnFilterOpTable,
							&mnFilterOpCount, ISTable[ISSeq].ifaceName);

        startHomeAction(mnsp, ISSeq);

	deleteAllFilters(mnFilters);
	sendInfoToAllGUIs();

    } else {
        mip3dbg(("\nEntering FOREIGN_STATE on iface %s [FA %s COA %s].\n",
		    mnsp->ifaceName,
		    ntoa(localMaTable[MAidx].mnAdvSourceAddr, addr1),
		    ntoa(localMaTable[MAidx].mnAdvCOAList[0], addr2)));

        // set filter retransmission with the new binding
	// as FA CoA
	setFilterForNewBinding(mnFilters, mnFilterCount, mnFilterOpTable,
		&mnFilterOpCount, ISTable[ISSeq].ifaceName, ACT_REPLACE, regReqTable);

        startForeignAction(mnsp, &(localMaTable[MAidx]), ISSeq);

	applyReverseFilters(mnFilters, ISTable, ISTable[ISSeq].ifaceName);
	sendInfoToAllGUIs();
    }

//    	  sprintf(cmd, "/usr/local/bin/findprinter LOCATION=%s > /etc/printcap &", ntoa(localMaTable[MAidx].mnAdvSourceAddr, addr1));
//          mip3dbg(("Searching for local printer ...\n"));
//          system(cmd);
    return;
}
