/*

  Sun Public License Notice

 The contents of this file are subject to the Sun Public License Version
 1.0 (the "License"). You may not use this file except in compliance with
 the License. A copy of the License is available at http://www.sun.com/

 The Original Code is the Sunlabs Mobile IP.
 The Initial Developer of the Original Code is Vipul Gupta.
 Portions created by Vipul Gupta are Copyright (C) 2003.
 All Rights Reserved.

 Contributor(s): Niko A. Fikouras (niko@comnets.uni-bremen.de)
		 Asanga D. Udugama (adu@comnets.uni-bremen.de)

*/

/*
 * An implementation of Mobile-IP as described in RFC 2002.
 * Author  : Sunil Madhani (sunil.madhani@eng.sun.com)
 * mnTable.c : Routines for manipulating table..
 */

#pragma ident	"@(#)mnTable.c 1.2 98/02/11"

#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>

#include "mip.h"
#include "mobile.h"


extern char currentTime[], addr1[], addr2[];

extern int mipdebug;
extern int periodicInterval;
extern int agentSolicitThreshold;
extern int agentExpireThreshold;
extern int ifacePollThreshold;
extern int regLifetime;
extern int regRoundTrip;
extern int regFlags;
extern int haCooloffPeriod;
extern int faCooloffPeriod;
extern int IDfreshnessSlack;
extern int solicitOnBcast;
extern int retransmissionPolicy[];
extern float renewalPolicy[];

/* The mobile node's address, netmask and default routers at home */
extern u32 mnHomeAddr;
extern u32 mnHomeNetmask;
extern u32 mnHomeBcastAddr;

// currnetState is part of ISTable; changed to handle MSB
extern InterfaceState ISTable[DEFAULT_MAX_INTERFACES];
//extern mnState currentState;

extern IfaceEntry ifaceTable[];
extern int ifaceTableSize;

extern HomeAgentEntry haTable[];
extern int haTableSize;

/* List of default routers on the home network */
extern u32 homeRouterTable[];
extern int homeRouterTableSize;

/* List of local Mobility Agents (considered active) */
extern LocalMaEntry localMaTable[];
extern int localMaTableSize;

/* List of known local Mobility Agents (considered active) */
/* Used by the ECS movement detection method - modified by */
/* Uni. Sheffield					   */
extern LocalMaEntry *knownLocalMaTable[DEFAULT_MAX_MOBILITY_AGENTS];

/* List of registration requests (both accepted and pending) */
extern RegReqEntry regReqTable[];
extern int regReqTableSize;
extern u32 mnRegRequestsAcceptedCnt;

extern int  ICMPsock;
extern int UDPregSock;

//extern mnState currentState;
extern DomainInfo domainInfo;

/*
 * Looks up the mipIfaceTable to check whether interface 
 * named "iface" is present
 * Returns pointer to that entry in ifaceTable if present 
 * else NULL.
 */
IfaceEntry *findIface(char *ifname)
{
    int i;

    for (i = 0; i < ifaceTableSize; i++) {
        if (strcmp(ifaceTable[i].maIfaceName, ifname) == 0)
	  return(&(ifaceTable[i]));
    }
    return(NULL);
}


int getColocatedAddr(int ISSeq)
{
   int i;

   for (i = 0; i < ifaceTableSize; i++) {
        if ((ifaceTable[i].maIfaceAddr != mnHomeAddr) &&
	    (ifaceTable[i].maIfaceAddr != 0)
	    && (strcmp(ifaceTable[i].maIfaceName, ISTable[ISSeq].ifaceName) == 0)) {
            return(i);
        }
    }    
    return(-1);
}

/* 
 * If there is a default HA specified, we store the home network broadcast
 * address at the 0th position. We start searching from the last entry
 * either till we find a LIVE entry or the one in 0th position. By doing
 * so, we automatically curb discovery if no default entry is specified.
 */
u32 chooseHAaddr()
{
   int i;

   for (i = (haTableSize - 1); i > 0; i--) {
       if (haTable[i].mnHAflags & HA_IS_ALIVE)
	   return(haTable[i].mnHaAddr);
   }

   return(haTable[0].mnHaAddr);
}


RegReqEntry *getEmptyRREntry()
{
    int i;

    for (i = 0 ; i < DEFAULT_MAX_REG_REQ; i++){
	if (regReqTable[i].mnRegStatus == REG_UNUSED)
	    return(&(regReqTable[i]));
    }
    return(NULL);
}

LocalMaEntry *findMAIdx(u32 addr) 
{
    int i;

    for (i = 0; i < DEFAULT_MAX_MOBILITY_AGENTS; i++){
	if (localMaTable[i].mnAdvSourceAddr == addr)
	    return(&(localMaTable[i]));
    }
    return(NULL);
}

boolean havePendingDiscovery(u32 addr)
{
    int i;

    for (i = 0 ; i < DEFAULT_MAX_REG_REQ; i++){
	if (regReqTable[i].mnRegStatus == REG_UNUSED) continue;
	if (regReqTable[i].mnRegHAaddr == addr)
	    return(TRUE);
    }
    return(FALSE);
}

/* Check each interval to see if the given addr is outside the protected domain */

boolean isInsideProtectedDomain(u32 addr)
{
    int i;

    if (domainInfo.addrIntervalCnt == 0)
        return(TRUE);

    for (i = 0; i < domainInfo.addrIntervalCnt; i++) {
        if (((addr ^ domainInfo.addr[i]) & domainInfo.netmask[i]) == 0)
            return(TRUE);
    }

    return(FALSE);
}

boolean isOutsideProtectedDomain(u32 addr)
{
    return (isInsideProtectedDomain(addr) ? FALSE: TRUE);
}

